/*
   File: flow.c
   Does dataflow analysis on the AST for global variables and effects
   as static semantic check and interprocedural dataflow analysis.

   Copyright (C) 2008-2010 Marc Seutter

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

   CVS ID: "$Id: flow.c,v 1.13 2012/04/21 14:00:00 marcs Exp $"
*/

/* standard includes */
#include <stdio.h>
#include <string.h>

/* support lib includes */
#include <dcg.h>
#include <dcg_alloc.h>
#include <dcg_error.h>
#include <dcg_string.h>

/* local includes */
#include "options.h"
#include "cdl_ds.h"
#include "ast_utils.h"
#include "globals.h"
#include "variables.h"
#include "modules.h"
#include "contsens.h"
#include "flow.h"

/*
   Two tests needed to build the flow graph
   Note that both tests must return a indeterminate for a Group
   Note 2: An operator can not fail implicitly: it either aborts or
           returns to the caller with status (and a possible effect)
*/
static int contains_global_variables (affix_term_list dst_vars)
{ int ix;
  for (ix = 0; ix < dst_vars -> size; ix++)
    if (is_a_global_variable (dst_vars -> array[ix]))
      return (1);
  return (0);
}

static int guard_has_an_effect (member mem)
{ guard_type typ = mem -> Res_guard.typ;
  switch (typ)
    { case arith_split:	return (0);
      case split:	return (contains_global_variables (mem -> Res_guard.expr));
      case arith_join:	/* fall thru */
      case join:	return (is_a_global_variable (mem -> Res_guard.var));
      default: dcg_bad_tag (typ, "guard_has_an_effect");
    };
  return (0);
}

static int call_has_an_effect (member mem)
{ proc called = mem -> Res_call.pdef;
  proc_type ptype = called -> pspec -> ptype;
  return ((ptype == action) || (ptype == predicate));
}

static int member_has_an_effect (member mem)
{ switch (mem -> tag)
    { case TAGOp:
      case TAGRes_equal:
      case TAGCast:	 
      case TAGGroup_fail:return (0);
      case TAGAssign:	 return (is_a_global_variable (mem -> Assign.dst));
      case TAGRes_guard: return (guard_has_an_effect (mem));
      case TAGRes_call:	 return (call_has_an_effect (mem));
      default: dcg_bad_tag (mem -> tag, "has_an_effect");
    };
  return (0);
}

static int call_can_fail (member mem)
{ proc called = mem -> Res_call.pdef;
  proc_type ptype = called -> pspec -> ptype;
  return ((ptype == predicate) || (ptype == test));
}

static int member_can_fail (member mem)
{ switch (mem -> tag)
    { case TAGOp:	 return (0);
      case TAGRes_equal: return (1);
      case TAGAssign:	 return (0);
      case TAGRes_guard: return (mem -> Res_guard.can_fail);
      case TAGRes_call:	 return (call_can_fail (mem));
      case TAGCast:	 return (0);
      case TAGGroup_fail:return (0);
      default: dcg_bad_tag (mem -> tag, "member_can_fail");
    };
  return (0);
}

/*
   Control flow may leave an enclosed clause to the next alternative,
   but in this case no local defects are allowed i.e. no variable
   may yet have been overwritten. 
*/
static void pick_ptype_properties (proc_type ptype, int *can_fail, int *has_effect)
{ switch (ptype)
    { case prelude: 
      case interlude:
      case postlude:
      case action:	*can_fail = 0; *has_effect = 1; break;
      case function:	*can_fail = 0; *has_effect = 0; break;
      case predicate:	*can_fail = 1; *has_effect = 1; break;
      case test:	*can_fail = 1; *has_effect = 0; break;
      default: dcg_bad_tag (ptype, "ptype_properties");
    };
}

static void add_successor (int tf, member mem, member successor)
{ /* Add current member to predecessor list of successor */
  if (successor -> pred == member_list_nil)
    successor -> pred = new_member_list ();
  app_member_list (successor -> pred, mem);

  /* Next add true or false successor */
  if (tf) mem -> true_succ = successor;
  else mem -> false_succ = successor;
}

static void check_group_exec_flow (module mod, group grp, alternative fail_alt, int fail_effect,
				   int init_effect, int *can_fail, int *has_effect);
static void check_member_exec_flow (module mod, alternative_list alts, int altnr,
				    alternative fail_alt, int fail_effect,
				    int init_effect, member_list mems, int memnr,
				    int *effect, int *can_fail, int *has_effect)
{ member mem = mems -> array[memnr];

  /* Groups have a recursive treatment */
  if (mem -> tag == TAGGroup)
    { group sub_group = mem -> Group.grp;
      alternative_list sub_alts = sub_group -> alts;
      member first_mem = sub_alts -> array[0] -> p1mems -> array[0];

      /* Check if there is an alternative to fail to, if group fails */
      alternative fail_alt2 = alternative_nil;
      if (altnr + 1 < alts -> size)
	fail_alt2 = alts -> array[altnr + 1];
      else if (fail_alt != alternative_nil)
	fail_alt2 = fail_alt;

      /* connect first member of group alternatives to the group member */
      add_successor (1, mem, first_mem);
      check_group_exec_flow (mod, sub_group, fail_alt2, init_effect, *effect, can_fail, has_effect);
      if (sub_alts -> array[sub_alts -> size - 1] -> p1mems -> array[0] -> tag == TAGGroup_fail)
        { /* If the groups fails, add a fail successor to the group */
	  if (fail_alt2 == alternative_nil)
	    dcg_internal_error ("check_member_exec_flow");
	  add_successor (0, mem, fail_alt2 -> p1mems -> array[0]);
	};
      return;
    };

  /* if member is a group fail, hint the user and return */
  if (mem -> tag == TAGGroup_fail)
    { /* Check if we are not ignoring an effect while failing out of the group */
      member nextmem = fail_alt -> p1mems -> array[0];
      /* Needed to add successor here? */
      add_successor (1, mem, nextmem);

      if (!fail_effect && init_effect)
	contsens_error (mod, mem -> line, mem -> col, "Defects are not allowed");
      else contsens_hint (mod, mem -> line, mem -> col,
		          "Enclosed group may fail to an (implicit) outer alternative");
      return;
    };

  /* if member is an operator, flow ends here: no successors */
  if (mem -> tag == TAGOp)
    { /* With an explicit fail (either added or in place), we can really fail */
      switch (mem -> Op.op)
	{ case f_failure:
	    { *can_fail = 1;
	      if (*effect)
	        contsens_error (mod, mem -> line, mem -> col, "Defects are not allowed");
	    }; break;
	  case f_success:
	    if (*effect) *has_effect = 1;
	    break;
	  default:
	    if (init_effect) *has_effect = 1;
	    break;
	};
      return;
    };

  /* Check if member can fail in flow */
  if (member_can_fail (mem))
    { /* We have a failing member: we need an alternative to fail to. Maybe create one */
      int act_fail_effect = init_effect;
      member_list fail_mems;

      /* If there is a next alternative to try, no problem */
      if (altnr < alts -> size - 1)
	fail_mems = alts -> array[altnr + 1] -> p1mems;
      else if ((fail_alt == alternative_nil) ||		/* Nothing to fail to */
	       ((fail_alt -> p1mems -> array[0] -> tag == TAGOp) &&
		(fail_alt -> p1mems -> array[0] -> Op.op == f_failure))) /* Failing to fail */
	{ /* No alternative to fail to, create an explicit fail as a last outer alternative */
	  member new_mem = new_Op (mem -> line, mem -> col, f_failure);
	  alternative new_alt;
	  fail_mems = new_member_list ();
	  new_alt = new_alternative (fail_mems, new_member_list ());
	  app_member_list (fail_mems, new_mem);
	  app_alternative_list (alts, new_alt);
	  if (fail_alt != alternative_nil)
	    contsens_hint (mod, mem -> line, mem -> col, "Added implicit fail to enclosed group");
	}
      else
	{ /*
	     There is an alternative outside this (enclosed) group to fail to.
	     To support intermediate code construction we add an extra group fail
	     alternative to the current group.
	  */
	  member new_mem = new_Group_fail (mem -> line, mem -> col);
	  alternative new_alt;
	  fail_mems = new_member_list ();
	  new_alt = new_alternative (fail_mems, new_member_list ());
	  app_member_list (fail_mems, new_mem);
	  app_alternative_list (alts, new_alt);
	  act_fail_effect = fail_effect;
	};

      /* first member of fail members is false successor */
      add_successor (0, mem, fail_mems -> array[0]);

      /* Since we are able to fail, check that we do not have a defect */
      if (!act_fail_effect && (*effect))
	contsens_error (mod, mem -> line, mem -> col, "Defects are not allowed");
    };

  /* Check if member has an effect */
  if (member_has_an_effect (mem))
    *effect = 1;

  /* We can succeed as well and we are not an operator */
  if (memnr == mems -> size - 1)
    { member new_mem = new_Op (mem -> line, mem -> col, f_success);
      app_member_list (mems, new_mem);
    };

  /* next member of alternative is true successor */
  add_successor (1, mem, mems -> array[memnr + 1]);
}

static void check_alternative_exec_flow (module mod, alternative_list alts, int altnr,
					 alternative fail_alt, int fail_effect,
					 int init_effect, int *can_fail, int *has_effect)
{ /* Note that we might extend the members by an explicit succeed */
  member_list mems = alts -> array[altnr] -> p1mems;
  int effect = init_effect;
  int ix = 0;
  while (ix < mems -> size)
    { check_member_exec_flow (mod, alts, altnr, fail_alt, fail_effect, init_effect, mems, ix,
			      &effect, can_fail, has_effect);
      ix++;
    };

  /* Check that the first member of each later alternative is reachable */
  if (altnr)
    { member firstmem = mems -> array[0];
      if (firstmem -> pred == NULL)
	contsens_error (mod, firstmem -> line, firstmem -> col, "Alternative is not reachable");
    };
}

static void check_group_exec_flow (module mod, group grp, alternative fail_alt, int fail_effect,
				   int init_effect, int *can_fail, int *has_effect)
{ /* Note that we might extend the alternatives by an explicit fail */
  alternative_list alts = grp -> alts;
  int ix = 0;
  while (ix < alts -> size)
    { check_alternative_exec_flow (mod, alts, ix, fail_alt, fail_effect,
				   init_effect, can_fail, has_effect);
      ix++;
    };
}

static void check_procedure_type (module mod, proc p)
{ proc_type ptype = p -> pspec -> real_ptype;
  int must_have_effect = 0;
  int has_effect = 0;
  int must_fail = 0;
  int can_fail = 0;

  if (p -> tag != TAGAlts) return;
  pick_ptype_properties (ptype, &must_fail, &must_have_effect);
  check_group_exec_flow (mod, p -> Alts.grp, alternative_nil, 0, 0, &can_fail, &has_effect);
  if (must_fail && !can_fail)
    contsens_error (mod, p -> pspec -> line, p -> pspec -> col, "'%s %s' must be able to fail",
		    string_from_proc_type (ptype), p -> pspec -> canonic_name);
  if (!must_fail && can_fail)
    contsens_error (mod, p -> pspec -> line, p -> pspec -> col, "'%s %s' is not allowed to fail",
		    string_from_proc_type (ptype), p -> pspec -> canonic_name);
  if (must_have_effect && !has_effect)
    contsens_error (mod, p -> pspec -> line, p -> pspec -> col,
		    "'%s %s' must have an effect (along at least one succeeding path)",
		    string_from_proc_type (ptype), p -> pspec -> canonic_name);
  if (!must_have_effect && has_effect)
    contsens_error (mod, p -> pspec -> line, p -> pspec -> col,
		    "'%s %s' is not allowed to have an effect",
		    string_from_proc_type (ptype), p -> pspec -> canonic_name);
}

static void check_procedure_types_in_module (module mod)
{ proc_list procs = mod -> procs;
  int ix;
  for (ix = 0; ix < procs -> size; ix++)
    check_procedure_type (mod, procs -> array[ix]);
}

static void check_procedure_types ()
{ int ix;
  dcg_hint ("      checking procedure types");
  for (ix = 0; ix < all_modules -> size; ix++)
    check_procedure_types_in_module (all_modules -> array[ix]);
}

/*
   The following code determines which local variables need to be
   recovered upon failing from an alternative. In a later stage
   we may extend this code by a global dataflow analysis.
*/
static state_list formals_visibility (fpar_list fpars)
{ state_list sl = init_state_list (fpars -> size);
  int ix;
  for (ix = 0; ix < fpars -> size; ix++)
    { fpar fp = fpars -> array[ix];
      switch (fp -> fdir)
	{ case d_inherited:
	  case d_transient: sl -> array[ix] = derived; break;
	  case d_derived:   sl -> array[ix] = undefined; break;
	  default: dcg_bad_tag (fp -> fdir, "formals_visibility");
	};
    };
  sl -> size = fpars -> size;
  return (sl);
}

/*
   The following routine determines how the locals are affected by a succeeding member
   We carry two sets around namely a state_list forw_sl, that keeps track of the history
   of the locals since the start of the procedure and a state_list upd_sl, which keeps
   track of those locals that have been modified in the current alternative since entering
   the most recent group. The first is needed to check for undefined local variables, the
   second for determining the set of locals that need to be recovered when failing from
   the current alternative.
*/
static void check_operator_locals_flow (module mod, proc p, member m, state_list forw_sl)
{ fpar_list fpars = p -> pspec -> p1fpars;
  int ix;
  if (m -> Op.op != f_success) return;
  for (ix = 0; ix < fpars -> size; ix++)
    { fpar fp = fpars -> array[ix];
      if ((fp -> fdir == d_derived) && (get_local_state (forw_sl, ix) == undefined))
	contsens_error (mod, m -> line, m -> col,
			"Derived affix %s has not received a value upon succeeding", fp -> fname);
    };
}

static void source_variable (module mod, member m, state_list sl, affix_term src)
{ switch (src -> tag)
    { case TAGVar:
	{ variable vdef = src-> Var.vdef;
	  if ((vdef -> tag == TAGLocal_var) && (get_local_state (sl, vdef -> nr) == undefined))
	    contsens_error (mod, m -> line, m -> col,
			    "Local affix %s is used with an undefined value", vdef -> vname);
	};
      case TAGSub: break;
      default: dcg_bad_tag (src -> tag, "source_variable");
    };
}

static void source_expr (module mod, member m, state_list sl, affix_term expr)
{ switch (expr -> tag)
    { case TAGDyop:
	source_expr (mod, m, sl, expr -> Dyop.arg1);
	source_expr (mod, m, sl, expr -> Dyop.arg2);
	break;
      case TAGMonop:
	source_expr (mod, m, sl, expr -> Monop.arg);
	break;
      case TAGConcat:
	{ affix_term_list terms = expr -> Concat.terms;
	  int ix;
	  for (ix = 0; ix < terms -> size; ix++)
	    source_expr (mod, m, sl, terms -> array[ix]);
	}; break;
      case TAGVar: source_variable (mod, m, sl, expr);
      case TAGSub:
      case TAGInt:
      case TAGReal:
      case TAGText: break;
      default: dcg_bad_tag (expr -> tag, "source_expr");
    };
}

static void update_variable (state_list nsl, affix_term dst)
{ switch (dst -> tag)
    { case TAGVar:
	{ variable vdef = dst -> Var.vdef;
	  if (vdef -> tag == TAGLocal_var)
	    set_local_state (nsl, vdef -> nr, attached);
	};
      case TAGSub: break;
      default: dcg_bad_tag (dst -> tag, "update_variable");
    };
}

static void check_assign_locals_flow (module mod, member m, state_list forw_sl, state_list upd_sl,
				      state_list *next_forw_sl, state_list *next_upd_sl)
{ *next_forw_sl = rdup_state_list (forw_sl);
  *next_upd_sl =  rdup_state_list (upd_sl);
  source_variable (mod, m, forw_sl, m -> Assign.src);
  update_variable (*next_forw_sl, m -> Assign.dst);
  update_variable (*next_upd_sl, m -> Assign.dst);
}

static void check_split_locals_flow (module mod, member m, state_list forw_sl, state_list upd_sl,
				     state_list *next_forw_sl, state_list *next_upd_sl)
{ affix_term_list dsts = m -> Res_guard.expr;
  affix_term src = m -> Res_guard.var;
  int ix;
  *next_forw_sl = (dsts -> size == 0)?attach_state_list (forw_sl):rdup_state_list (forw_sl);
  *next_upd_sl = (dsts -> size == 0)?attach_state_list (upd_sl):rdup_state_list (upd_sl);
  source_variable (mod, m, forw_sl, src);
  for (ix = 0; ix < dsts -> size; ix++)
    { update_variable (*next_forw_sl, dsts -> array[ix]);
      update_variable (*next_upd_sl, dsts -> array[ix]);
    };
}

static void check_join_locals_flow (module mod, member m, state_list forw_sl, state_list upd_sl,
				    state_list *next_forw_sl, state_list *next_upd_sl)
{ affix_term_list srcs = m -> Res_guard.expr;
  affix_term dst = m -> Res_guard.var;
  int ix;
  *next_forw_sl = rdup_state_list (forw_sl);
  *next_upd_sl = rdup_state_list (upd_sl);
  for (ix = 0; ix < srcs -> size; ix++)
    source_variable (mod, m, forw_sl, srcs -> array[ix]);
  update_variable (*next_forw_sl, dst);
  update_variable (*next_upd_sl, dst);
}

static void check_arith_split_locals_flow (module mod, member m, state_list forw_sl)
{ affix_term src = m -> Res_guard.expr -> array[0];
  affix_term src2 = m -> Res_guard.var;
  source_variable (mod, m, forw_sl, src2);
  source_expr (mod, m, forw_sl, src);
}

static void check_arith_join_locals_flow (module mod, member m,
					  state_list forw_sl, state_list upd_sl,
					  state_list *next_forw_sl, state_list *next_upd_sl)
{ affix_term src = m -> Res_guard.expr -> array[0];
  affix_term dst = m -> Res_guard.var;
  *next_forw_sl = rdup_state_list (forw_sl);
  *next_upd_sl = rdup_state_list (upd_sl);
  source_expr (mod, m, forw_sl, src);
  update_variable (*next_forw_sl, dst);
  update_variable (*next_upd_sl, dst);
}

static void check_guard_locals_flow (module mod, member m, state_list forw_sl, state_list upd_sl,
				     state_list *next_forw_sl, state_list *next_upd_sl)
{ switch (m -> Res_guard.typ)
    { case split:
	check_split_locals_flow (mod, m, forw_sl, upd_sl, next_forw_sl, next_upd_sl);
	return;
      case join: 
	check_join_locals_flow (mod, m, forw_sl, upd_sl, next_forw_sl, next_upd_sl);
	return;
      case arith_split: check_arith_split_locals_flow (mod, m, forw_sl); break;
      case arith_join: 
	check_arith_join_locals_flow (mod, m, forw_sl, upd_sl, next_forw_sl, next_upd_sl);
	return;
      default: dcg_bad_tag (m -> Res_guard.typ, "check_guard_locals_flow");
    };
  *next_forw_sl = attach_state_list (forw_sl);
  *next_upd_sl = attach_state_list (upd_sl);
}

static void check_equal_locals_flow (module mod, member m, state_list forw_sl)
{ source_variable (mod, m, forw_sl, m -> Res_equal.left);
  source_variable (mod, m, forw_sl, m -> Res_equal.right);
}

static void check_call_locals_flow (module mod, member m, state_list forw_sl, state_list upd_sl,
				    state_list *next_forw_sl, state_list *next_upd_sl)
{ affix_term_list act_args = m -> Res_call.p1args;
  fpar_list fpars = m -> Res_call.pdef -> pspec -> p1fpars;
  int need_update = 0;
  int ix;
  for (ix = 0; ix < fpars -> size; ix++)
    { fpar fp = fpars -> array[ix];
      affix_term act_arg = act_args -> array[ix];
      switch (fp -> fdir)
        { case d_inherited: source_variable (mod, m, forw_sl, act_arg); break;
	  case d_transient: source_variable (mod, m, forw_sl, act_arg); /* fall thru */
	  case d_derived: need_update = 1; break;
	  default: dcg_bad_tag (fp -> fdir, "check_call_locals_flow");
	};
    };

  if (!need_update)
    { *next_forw_sl = attach_state_list (forw_sl);
      *next_upd_sl = attach_state_list (upd_sl);
      return;
    };

  *next_forw_sl = rdup_state_list (forw_sl);
  *next_upd_sl = rdup_state_list (upd_sl);
  for (ix = 0; ix < fpars -> size; ix++)
    { fpar fp = fpars -> array[ix];
      affix_term act_arg = act_args -> array[ix];
      switch (fp -> fdir)
        { case d_transient:
	  case d_derived:
	    update_variable (*next_forw_sl, act_arg);
	    update_variable (*next_upd_sl, act_arg);
	  case d_inherited: break;
	  default: dcg_bad_tag (fp -> fdir, "check_call_locals_flow");
	};
    };
}

static void check_group_locals_flow (module mod, proc p, group grp, state_list init_sl);
static void check_member_locals_flow (module mod, proc p, member m,
				      state_list forw_sl, state_list upd_sl,
				      state_list *next_forw_sl, state_list *next_upd_sl)
{ switch (m -> tag)
    { case TAGGroup:
	/* When failing out of a group to a next alternative, we have recovered to forw_sl */
	check_group_locals_flow (mod, p, m -> Group.grp, forw_sl);
	break;
      case TAGAssign:
	check_assign_locals_flow (mod, m, forw_sl, upd_sl, next_forw_sl, next_upd_sl);
	return;
      case TAGRes_guard:  
	check_guard_locals_flow (mod, m, forw_sl, upd_sl, next_forw_sl, next_upd_sl);
	return;
      case TAGRes_equal:	
	check_equal_locals_flow (mod, m, forw_sl);
	break;
      case TAGRes_call:  
	check_call_locals_flow (mod, m, forw_sl, upd_sl, next_forw_sl, next_upd_sl);
	return;
      case TAGOp:
	check_operator_locals_flow (mod, p, m, forw_sl); /* fallthru */
      case TAGCast:
      case TAGGroup_fail: break;
      default: dcg_bad_tag (m -> tag, "check_member_locals_flow");
    };
  *next_forw_sl = attach_state_list (forw_sl);
  *next_upd_sl = attach_state_list (upd_sl);
}

/*
   In the following code we recursively analyse successor members,
   while symbolically executing. Thus we know which variables get
   overwritten. On the recursion rebound, we then add changed variables
   to the recovery list.
*/
static int try_add_recover_variable (variable pvar, proc p, assoc_pair_list *rec_pl)
{ assoc_pair_list rpl = *rec_pl;
  string auxvname;
  variable auxvar;
  int auxnr, ix;

  /* Create list if not yet allocated */
  if (rpl == assoc_pair_list_nil)
    { rpl = new_assoc_pair_list ();
      *rec_pl = rpl;
    };

  /* Already present? */
  for (ix = 0; ix < rpl -> size; ix++)
    if (rpl -> array[ix] -> lnr == pvar -> nr)
      return (0);
  
  /* Allocate */
  auxnr = p -> pvars -> size;
  auxvname = dcg_new_fmtd_string ("%s$auxf%d", pvar -> vname, auxnr);
  auxvar = new_Local_var (auxvname);
  auxvar -> adef = pvar -> adef;
  auxvar -> nr = auxnr;
  app_variable_list (p -> pvars, auxvar);
  app_assoc_pair_list (rpl, new_assoc_pair (pvar -> nr, auxnr));
  return (1);
}

static void try_update_recovery_list (module mod, proc p, member m, state_list init_sl,
				      state_list upd_sl, assoc_pair_list *rec_pl)
{ int ix;
  for (ix = 0; ix < upd_sl -> size; ix++)
    if ((get_local_state (upd_sl, ix) != undefined) && /* variable has been touched */
        (get_local_state (init_sl, ix) != undefined))  /* variable had a value before */
      { variable pvar = p -> pvars -> array[ix];
	if (try_add_recover_variable (pvar, p, rec_pl))
	  contsens_hint (mod, m -> line, m -> col, "Found recoverable variable %s", pvar -> vname);
      };
}

static void check_members_locals_flow (module mod, proc p, member m, state_list init_sl,
				       state_list forw_sl, state_list upd_sl,
				       assoc_pair_list *rec_pl, int need_rec)
{ state_list next_forw_sl, next_upd_sl;
  check_member_locals_flow (mod, p, m, forw_sl, upd_sl, &next_forw_sl, &next_upd_sl);
  if ((m -> true_succ != member_nil) && (m -> tag != TAGGroup_fail))
    check_members_locals_flow (mod, p, m -> true_succ, init_sl,
			       next_forw_sl, next_upd_sl, rec_pl, need_rec);
  if (need_rec && (m -> false_succ != member_nil))
    try_update_recovery_list (mod, p, m, init_sl, upd_sl, rec_pl);
  nonrec_detach_state_list (&next_forw_sl);
  nonrec_detach_state_list (&next_upd_sl);
}

static void check_alternative_locals_flow (module mod, proc p, alternative alt,
					   state_list init_sl, int need_rec)
{ assoc_pair_list rec_pl = assoc_pair_list_nil;
  state_list zero_sl = init_state_list (init_sl -> size);
  check_members_locals_flow (mod, p, alt -> p1mems -> array[0], init_sl, init_sl, zero_sl,
			     &rec_pl, need_rec);
  alt -> recovery = rec_pl;
}

static void check_group_locals_flow (module mod, proc p, group grp, state_list init_sl)
{ alternative_list alts = grp -> alts;
  int ix;
  grp -> visibility = attach_state_list (init_sl);
  for (ix = 0; ix < alts -> size; ix++)
    { int need_rec = 1;
      if (ix == alts -> size - 2)
	{ alternative last_alt = alts -> array[alts -> size - 1];
	  member fail_tgt = last_alt -> p1mems -> array[0];
	  /*
	     If the last alternative is a simple succeed, we must be able to recover
	     the values of the outgoing affixes at the start of the group.
	  */
	  if ((fail_tgt -> tag == TAGOp) && (fail_tgt -> Op.op != f_success))
	    need_rec = 0;
	};
      check_alternative_locals_flow (mod, p, alts -> array[ix], init_sl, need_rec);
    };
}

static void check_procedure_locals_in_module (module mod)
{ proc_list procs = mod -> procs;
  int ix;
  for (ix = 0; ix < procs -> size; ix++)
    { proc p = procs -> array[ix];
      state_list init_sl;
      if (p -> tag != TAGAlts) continue;

      init_sl = formals_visibility (p -> pspec -> p1fpars);
      check_group_locals_flow (mod, p, p -> Alts.grp, init_sl);
      nonrec_detach_state_list (&init_sl);
    };
}

static void check_procedure_locals ()
{ int ix;
  dcg_hint ("      checking locals flow");
  for (ix = 0; ix < all_modules -> size; ix++)
    check_procedure_locals_in_module (all_modules -> array[ix]);
}

/*
   Procedure call counting (Needed for inlining e.g.)

   The following code determines, starting from the root which includes
   the prelude and postlude calls, which procedures in the entire program
   are called.  When not called at all, a procedure will not be coded.
*/
static void count_procedure_calls (proc pdef);
static void count_procedure_calls_in_group (group grp);
static void count_procedure_calls_in_member (member mem)
{ switch (mem -> tag)
    { case TAGRes_call: count_procedure_calls (mem -> Res_call.pdef);  break;
      case TAGGroup: count_procedure_calls_in_group (mem -> Group.grp); break;
      case TAGOp:   
      case TAGRes_equal:
      case TAGAssign:
      case TAGRes_guard:
      case TAGCast:	
      case TAGGroup_fail:
	break;
      default: dcg_bad_tag (mem -> tag, "count_procedure_calls_in_member");
    };
}

static void count_procedure_calls_in_members (member_list mems)
{ int ix;
  for (ix = 0; ix < mems -> size; ix++)
    count_procedure_calls_in_member (mems -> array[ix]);
}

static void count_procedure_calls_in_group (group grp)
{ alternative_list alts = grp -> alts;
  int ix;
  for (ix = 0; ix < alts -> size; ix++)
    { alternative alt = alts -> array[ix];
      count_procedure_calls_in_members (alt -> p1mems);
    };
}

static void count_procedure_calls (proc pdef)
{ int first_time = (pdef -> called == 0);
  pdef -> called++;
  if (!first_time) return;
  switch (pdef -> tag)
    { case TAGAlts: count_procedure_calls_in_group (pdef -> Alts.grp); break;
      case TAGRoot: count_procedure_calls_in_members (pdef -> Root.mems); break;
      default: break;
    };
}

static void report_non_called_procedures ()
{ int ix;
  for (ix = 0; ix < all_procedures -> size; ix++)
    { proc pdef = all_procedures -> array[ix];
      if (!pdef -> called)
	{ module mod = all_modules -> array[pdef -> mnr];
	  spec pspec = pdef -> pspec;
	  if (!streq (mod -> mname, "predef"))
	    contsens_hint (mod, pspec -> line, pspec -> col, "'%s %s' is never called",
			   string_from_proc_type (pspec -> ptype), pspec -> canonic_name);
	};
    };
}

static void determine_called_procedures ()
{ proc_list root_procs = root_module -> procs;
  int ix;
  dcg_hint ("      counting calls");

  /* Mark root procs */
  count_procedure_calls (root_module -> Root_module.prelude);
  if (root_module -> Root_module.root != proc_nil)
    count_procedure_calls (root_module -> Root_module.root);
  count_procedure_calls (root_module -> Root_module.postlude);

  /* Mark exported procs */
  for (ix = 0; ix < root_procs -> size; ix++)
    { proc p = root_procs -> array[ix];
      if (p -> tag != TAGAlts) continue;
      if (p -> Alts.exported)
	count_procedure_calls (p);
    };

  report_non_called_procedures ();
}

void flow_analysis ()
{ dcg_warning (0, "   flow analysis...");
  check_procedure_types ();
  dcg_panic_if_errors ();
  check_procedure_locals ();
  dcg_panic_if_errors ();
  determine_called_procedures ();
  if (dump_flow)
    { dump_modules ();
      dcg_warning (0, "PRELUDE: ");
      pp_proc (dcg_error_file (), root_module -> Root_module.prelude);
      dcg_warning (0, "\nROOT: ");
      pp_proc (dcg_error_file (), root_module -> Root_module.root);
      dcg_warning (0, "\nPOSTLUDE: ");
      pp_proc (dcg_error_file (), root_module -> Root_module.postlude);
      dcg_warning (0, "");
    };
}
