/*
   File: imc_utils.c
   Defines support routines to handle the intermediate code

   Copyright (C) 2008-2010 Marc Seutter

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

   CVS ID: "$Id: imc_utils.c,v 1.8 2011/08/13 15:57:30 marcs Exp $"
*/

/* System includes */
#include <stdio.h>
#include <stdlib.h>

/* libdcg includes */
#include <dcg.h>
#include <dcg_error.h>

/* Local includes */
#include "cdl_ds.h"
#include "ast_utils.h"
#include "imc_utils.h"
#include "globals.h"

/*
   The basic structure of a procedure in the intermediate code is the graph
   consisting of the basic blocks of the procedure. Each basic block is
   constructed as a non circular double linked list. This allows easy code
   restructuring to implement an optimization pass at the intermediate code.
   All new basic blocks must be registered at the proc level to ease coding
   to the target language.
*/

/*
   Manipulation of block head and tail

   Append instr appends an instruction at the end of a block
   Prepend instr inserts an instruction at the start of a block, usually a label
*/
void append_instr (block bb, instr ins)
{ instr tail = bb -> tail;
  if (tail == instr_nil)
    { bb -> head = ins;
      bb -> tail = ins;
    }
  else
    { ins -> prev = tail;
      tail -> next = ins;
      bb -> tail = ins;
    };
}

void prepend_instr (block bb, instr ins)
{ instr head = bb -> head;
  if (head == instr_nil)
    { bb -> head = ins;
      bb -> tail = ins;
    }
  else
    { ins -> next = head;
      head -> prev = ins;
      bb -> head = ins;
    };
}

void remove_block_from_list (block_list bl, block bb)
{ int ix = 0;
  while (ix < bl -> size)
    if (bb == bl -> array[ix])
      delete_block_list (bl, ix);
    else ix++;
}

void replace_last_instruction (block bb, instr newlast)
{ instr old_last = bb -> tail;
  if (bb -> head == old_last)
    { bb -> head = newlast;
      bb -> tail = newlast;
    }
  else
    { newlast -> prev = old_last -> prev;
      newlast -> prev -> next = newlast;
      bb -> tail = newlast;
    };
  det_instr (&old_last);
}

/*
   Label administration
*/
static block_list label_table;
static void create_label_and_block (int *label, block *bb)
{ instr hdr;
  *label = label_table -> size;
  hdr = new_Label (*label);
  *bb = new_block ();
  append_instr (*bb, hdr);
  app_block_list (label_table, *bb);
}

int pick_or_create_new_block_label (block bb)
{ int lnr;
  if (bb -> head == instr_nil)
    dcg_internal_error ("pick_or_create_new_block_label");
  if (bb -> head -> tag == TAGLabel)
    return (bb -> head -> Label.lnr);
  lnr = label_table -> size;
  prepend_instr (bb, new_Label (lnr));
  app_block_list (label_table, bb);
  return (lnr);
}

block get_block_from_label (int label)
{ if ((0 <= label) && (label < label_table -> size))
    return (label_table -> array[label]);
  dcg_internal_error ("get_block_from_label: inconsistent label");
  return (block_nil);
}

int instr_has_fail_label (instr ins)
{ switch (ins -> tag)
    { case TAGEqual_altnr:
      case TAGEqual:
      case TAGEqual_int:
      case TAGEqual_real:
      case TAGCallf:
      case TAGSwitch_altnr:
      case TAGSwitch_int: return (1);
      default: return (0);
    };
}

void set_fail_label (instr ins, int label)
{ switch (ins -> tag)
    { case TAGEqual_altnr:	ins -> Equal_altnr.fail = label; break;
      case TAGEqual:		ins -> Equal.fail = label; break;
      case TAGEqual_int:	ins -> Equal_int.fail = label; break;
      case TAGEqual_real:	ins -> Equal_real.fail = label; break;
      case TAGCallf:		ins -> Callf.fail = label; break;
      case TAGSwitch_altnr:	ins -> Switch_altnr.fail = label; break;
      case TAGSwitch_int:	ins -> Switch_int.fail = label; break;
      default: dcg_bad_tag (ins -> tag, "set_fail_label");
    };
}

int pick_fail_label (instr ins)
{ switch (ins -> tag)
    { case TAGEqual_altnr:	return (ins -> Equal_altnr.fail);
      case TAGEqual:		return (ins -> Equal.fail);
      case TAGEqual_int:	return (ins -> Equal_int.fail);
      case TAGEqual_real:	return (ins -> Equal_real.fail);
      case TAGCallf:		return (ins -> Callf.fail);
      case TAGSwitch_altnr:	return (ins -> Switch_altnr.fail);
      case TAGSwitch_int:	return (ins -> Switch_int.fail);
      default: dcg_bad_tag (ins -> tag, "pick_fail_label");
    };
  return (-1);
}

int is_a_local_address (address adr, int *lnr)
{ switch (adr -> tag)
    { case TAGPart: return (is_a_local_address (adr -> Part.adr, lnr));
      case TAGLocal: *lnr = adr -> Local.lnr; break;
      default: return (0);
    };
  return (1);
}

int is_a_reference (address adr, address *deref)
{ if (adr -> tag != TAGReference) return (0);
  *deref = adr -> Reference.adr;
  return (1);
}

int is_a_local_detach (instr ins, int *lnr)
{ if (ins -> dont_code) return (0);
  if (ins -> tag != TAGDetach) return (0);
  return (is_a_local_address (ins -> Detach.loc, lnr));
}

int is_a_local_attach (instr ins, int *lnr)
{ if (ins -> dont_code) return (0);
  if (ins -> tag != TAGAttach) return (0);
  return (is_a_local_address (ins -> Attach.loc, lnr));
}

int is_a_call (instr ins, address_list *args)
{ if (ins -> dont_code) return (0);
  switch (ins -> tag)
    { case TAGCallp: *args = ins -> Callp.args; break;
      case TAGCallf: *args = ins -> Callf.args; break;
      default: return (0);
    };
  return (1);
}

int is_a_local_update (instr ins, int *lnr)
{ if (ins -> dont_code) return (0);
  switch (ins -> tag)
    { case TAGCreate:    return (is_a_local_address (ins -> Create.dst, lnr));
      case TAGCopy:      return (is_a_local_address (ins -> Copy.dst, lnr));
      case TAGCopy_int:  return (is_a_local_address (ins -> Copy_int.dst, lnr));
      case TAGCopy_real: return (is_a_local_address (ins -> Copy_real.dst, lnr));
      case TAGConc_text: return (is_a_local_address (ins -> Conc_text.dst, lnr));
      case TAGRepl_text: return (is_a_local_address (ins -> Repl_text.dst, lnr));
      default: return (0);
    };
}

static int is_this_local (address adr, int lnr)
{ int lnr2;
  if (!is_a_local_address (adr, &lnr2)) return (0);
  return (lnr == lnr2);
}

static int iexpr_contains_this_local (iexpr exp, int lnr)
{ switch (exp -> tag)
    { case TAGIvref:  return (is_this_local (exp -> Ivref.adr, lnr));
      case TAGIdyop:
	if (iexpr_contains_this_local (exp -> Idyop.src1, lnr)) return (1);
	  return (iexpr_contains_this_local (exp -> Idyop.src2, lnr));
      case TAGImonop: return (iexpr_contains_this_local (exp -> Imonop.src, lnr));
      case TAGInum:   break;
      default: dcg_bad_tag (exp -> tag, "iexpr_contains_this_local");
    };
  return (0);
}

static int rexpr_contains_this_local (rexpr exp, int lnr)
{ switch (exp -> tag)
    { case TAGRvref:  return (is_this_local (exp -> Rvref.adr, lnr));
      case TAGRdyop:
	if (rexpr_contains_this_local (exp -> Rdyop.src1, lnr)) return (1);
	  return (rexpr_contains_this_local (exp -> Rdyop.src2, lnr));
      case TAGRmonop: return (rexpr_contains_this_local (exp -> Rmonop.src, lnr));
      case TAGRnum:   break;
      default: dcg_bad_tag (exp -> tag, "iexpr_contains_this_local");
    };
  return (0);
}

static int args_contain_this_local (address_list adrs, int lnr)
{ int ix;
  for (ix = 0; ix < adrs -> size; ix++)
    { address arg = adrs -> array[ix];
      (void) is_a_reference (arg, &arg);
      if (is_this_local (arg, lnr)) return (1);
    };
  return (0);
}

int is_a_local_use (instr ins, int lnr)
{ if (ins -> dont_code) return (0);
  switch (ins -> tag)
    { case TAGNop:
      case TAGProc_entry:
      case TAGLabel:
      case TAGComment:
      case TAGSucceed:
      case TAGFail:
      case TAGExit:
      case TAGAbort:  break;
      case TAGAttach: return (is_this_local (ins -> Attach.loc, lnr));
      case TAGDetach: return (is_this_local (ins -> Detach.loc, lnr));
      case TAGCreate: break;
      case TAGEqual_altnr: return (is_this_local (ins -> Equal_altnr.src, lnr));
      case TAGEqual:       return (is_this_local (ins -> Equal.src1, lnr) ||	
				   is_this_local (ins -> Equal.src2, lnr));
      case TAGEqual_int:   return (is_this_local (ins -> Equal_int.src, lnr) ||
				   iexpr_contains_this_local (ins -> Equal_int.cmp, lnr));
      case TAGEqual_real:  return (is_this_local (ins -> Equal_real.src, lnr) ||
				   rexpr_contains_this_local (ins -> Equal_real.cmp, lnr));
      case TAGCopy:      return (is_this_local (ins -> Copy.src, lnr));
      case TAGCopy_int:  return (iexpr_contains_this_local (ins -> Copy_int.src, lnr));
      case TAGCopy_real: return (rexpr_contains_this_local (ins -> Copy_real.src, lnr));
      case TAGCallp:     return (args_contain_this_local (ins -> Callp.args, lnr));
      case TAGCallf:     return (args_contain_this_local (ins -> Callf.args, lnr));
      case TAGIndex:     return (iexpr_contains_this_local (ins -> Index.idx, lnr));
      case TAGConc_text: return (args_contain_this_local (ins -> Conc_text.srcs, lnr));
      case TAGRepl_text: return (is_this_local (ins -> Repl_text.src, lnr) ||
				 iexpr_contains_this_local (ins -> Repl_text.fac, lnr));
      case TAGSwitch_altnr: return (is_this_local (ins -> Switch_altnr.src, lnr));
      case TAGSwitch_int:   return (is_this_local (ins -> Switch_int.src, lnr));
      default: dcg_bad_tag (ins -> tag, "is_a_local_use");
    };
  return (0);
}

/*
   Per instruction gen/use calculation
*/
static void calc_local_use (address adr, char_list use)
{ int lnr;
  if (!is_a_local_address (adr, &lnr)) return;
  use -> array[lnr] = 1;
}

/*
   When we assign to a local address, at to gen and delete from use
   Remember that an assignment to a part of a local is a use of that local
   We allow gen to be the null ptr, necessary for pass2 of the live dead analysis
*/
static void calc_local_gen (address adr, char_list gen, char_list use)
{ int lnr;
  if (!is_a_local_address (adr, &lnr)) return;
  if ((adr -> tag == TAGPart) && is_a_local_address (adr -> Part.adr, &lnr))
    { use -> array[lnr] = 1;
      return;
    };

  if (gen != char_list_nil) gen -> array[lnr] = 1;
  use -> array[lnr] = 0;
}

static void calc_local_int_uses (iexpr exp, char_list use)
{ switch (exp -> tag)
    { case TAGIvref: calc_local_use (exp -> Ivref.adr, use); break;
      case TAGIdyop:
	calc_local_int_uses (exp -> Idyop.src1, use);
	calc_local_int_uses (exp -> Idyop.src2, use);
	break;
      case TAGImonop:
	calc_local_int_uses (exp -> Imonop.src, use);
      case TAGInum: break;
      default: dcg_bad_tag (exp -> tag, "calc_local_int_uses");
    };
}

static void calc_local_real_uses (rexpr exp, char_list use)
{ switch (exp -> tag)
    { case TAGRvref: calc_local_use (exp -> Rvref.adr, use); break;
      case TAGRdyop:
	calc_local_real_uses (exp -> Rdyop.src1, use);
	calc_local_real_uses (exp -> Rdyop.src2, use);
	break;
      case TAGRmonop:
	calc_local_real_uses (exp -> Rmonop.src, use);
      case TAGRnum: break;
      default: dcg_bad_tag (exp -> tag, "calc_local_real_uses");
    };
}

static void calc_local_uses (address_list adrs, char_list use)
{ int ix;
  for (ix = 0; ix < adrs -> size; ix++)
    calc_local_use (adrs -> array[ix], use);
}

static void calc_local_proc_gen_use (fpar fp, address act, char_list gen, char_list use)
{ address deref;
  switch (fp -> fdir)
    { case d_inherited:
	calc_local_use (act, use);
	break;
      case d_derived:
	if (is_a_reference (act, &deref))
	  calc_local_gen (deref, gen, use);
	break;
      case d_transient:
	if (is_a_reference (act, &deref))
	  calc_local_gen (deref, gen, use);
	calc_local_use (deref, use);
	break;
      default: dcg_bad_tag (fp -> fdir, "calc_local_proc_gen_use");
    };
}

static void calc_local_proc_gens_uses (int pnr, address_list adrs, char_list gen, char_list use)
{ proc called = all_procedures -> array[pnr];
  fpar_list p1fpars = called -> pspec -> p1fpars;
  int ix;
  for (ix = 0; ix < p1fpars -> size; ix++)
    calc_local_proc_gen_use (p1fpars -> array[ix], adrs -> array[ix], gen, use);
}

void calc_local_gen_use (instr ins, char_list gen, char_list use)
{ if (ins -> dont_code) return;
  switch (ins -> tag)
    { case TAGNop:
      case TAGProc_entry:
      case TAGLabel:
      case TAGComment:
      case TAGSucceed:
      case TAGFail:
      case TAGExit:
      case TAGAbort: break;
      case TAGAttach: calc_local_use (ins -> Attach.loc, use); break;
      case TAGDetach: calc_local_use (ins -> Detach.loc, use); break;
      case TAGCreate: calc_local_gen (ins -> Create.dst, gen, use); break;
      case TAGEqual_altnr: calc_local_use (ins -> Equal_altnr.src, use); break;
      case TAGEqual:
	calc_local_use (ins -> Equal.src1, use);
	calc_local_use (ins -> Equal.src2, use);
	break;
      case TAGEqual_int:  calc_local_use (ins -> Equal_int.src, use); break;
      case TAGEqual_real: calc_local_use (ins -> Equal_real.src, use); break;
      case TAGCopy:
	calc_local_gen (ins -> Copy.dst, gen, use);
	calc_local_use (ins -> Copy.src, use);
	break;
      case TAGCopy_int:
	calc_local_gen (ins -> Copy_int.dst, gen, use);
	calc_local_int_uses (ins -> Copy_int.src, use);
	break;
      case TAGCopy_real:
	calc_local_gen (ins -> Copy_real.dst, gen, use);
	calc_local_real_uses (ins -> Copy_real.src, use);
	break;
      case TAGCallp:
	calc_local_proc_gens_uses (ins -> Callp.pnr, ins -> Callp.args, gen, use);
	break;
      case TAGCallf:
	calc_local_proc_gens_uses (ins -> Callf.pnr, ins -> Callf.args, gen, use);
	break;
      case TAGIndex: calc_local_int_uses (ins -> Index.idx, use); break;
      case TAGConc_text:
	calc_local_gen (ins -> Conc_text.dst, gen, use);
	calc_local_uses (ins -> Conc_text.srcs, use);
	break;
      case TAGRepl_text:
	calc_local_gen (ins -> Repl_text.dst, gen, use);
	calc_local_int_uses (ins -> Repl_text.fac, use);
	calc_local_use (ins -> Repl_text.src, use);
	break;
      case TAGSwitch_altnr: calc_local_use (ins -> Switch_altnr.src, use); break;
      case TAGSwitch_int: calc_local_use (ins -> Switch_int.src, use); break;
      default: dcg_bad_tag (ins -> tag, "calc_local_gen_use");
    };
}

/*
   Basic block graph construction
*/
void connect_true_successor (block *current, proc pdef)
{ block follower = new_block ();
  block old = *current;

  /* Copy state information */
  app_block_list (pdef -> bbs, follower);
  follower -> visibility = rdup_state_list (old -> visibility);

  /* Connect old and follower block */
  old -> true_succ = new_block_list ();
  app_block_list (old -> true_succ, follower);
  follower -> pred = new_block_list ();
  app_block_list (follower -> pred, old);
  *current = follower;
}

void connect_new_fail_successor (block failing_bb, block prev_rec_bb, proc pdef,
				 int *fail_label, block *current)
{ block follower;
  create_label_and_block (fail_label, &follower);

  /* Copy state information */
  app_block_list (pdef -> bbs, follower);
  follower -> visibility = rdup_state_list (failing_bb -> visibility);

  /* Connect */
  follower -> pred = new_block_list ();
  if (prev_rec_bb != block_nil)
    { prev_rec_bb -> true_succ = new_block_list ();
      app_block_list (prev_rec_bb -> true_succ, follower);
      app_block_list (follower -> pred, prev_rec_bb);
    };

  if (failing_bb == block_nil)
    dcg_internal_error ("connect_new_fail_successor");
  
  failing_bb -> false_succ = follower;
  app_block_list (follower -> pred, failing_bb);
  *current = follower;
}

void connect_fail_successor (block failing_bb, block follower)
{ failing_bb -> false_succ = follower;
  app_block_list (follower -> pred, failing_bb);
}

state get_variable_state (block bb, address var_adr)
{ int nr;
  if (var_adr -> tag != TAGLocal) return (undefined);
  nr = var_adr -> Local.lnr;
  return (get_local_state (bb -> visibility, nr));
}

void update_variable (block bb, address dst_adr, instr loc, state ns)
{ state os;
  int nr;
  if (dst_adr -> tag != TAGLocal) return;
  nr = dst_adr -> Local.lnr;
  os = get_local_state (bb -> visibility, nr);
  set_local_state (bb -> visibility, nr, ns);
}

/*
   Manipulation of the double linked lists that keep the instructions
*/
void insert_instr_before (instr new, instr anchor)
{ instr prev = anchor -> prev;
  new -> next = anchor;
  new -> prev = prev;
  anchor -> prev = new;
  if (prev != instr_nil)
    prev -> next = new;
}

void insert_instr_after (instr new, instr *anchor)
{ instr next = (*anchor) -> next;
  new -> prev = *anchor;
  new -> next = next;
  (*anchor) -> next = new;
  if (next != instr_nil)
    next -> prev = new;
  *anchor = new;
}

/*
   Dumping of the intermediate code
*/

/*
   Dumping of a procedure name
*/
static void dump_fpars_names (fpar_list fpars)
{ int ix;
  for (ix = 0; ix < fpars -> size; ix++)
    dcg_eprint ("_%s", fpars -> array[ix] -> fdef -> adef -> aname);
}

void dump_proc_name (proc pdef)
{ switch (pdef -> tag)
    { case TAGExt_proc: dcg_eprint (pdef -> Ext_proc.ename); break;
      case TAGAlts:
        { string mname = all_modules -> array[pdef -> mnr] -> mname;
          dcg_eprint ("M%s_P%s", mname, pdef -> pspec -> concat_name);
        }; break;
      case TAGRoot:
        { string mname = all_modules -> array[pdef -> mnr] -> mname;
          dcg_eprint ("M%s_%s", mname, pdef -> pspec -> concat_name);
        }; break;
      default: dcg_bad_tag (pdef -> tag, "dump_proc_name");
    };
  dump_fpars_names (pdef -> pspec -> p1fpars);
}

/*
   Dumping of a formal procedure header
*/
static void dump_proc_args (proc pdef)
{ fpar_list fpars = pdef -> pspec -> p1fpars;
  int ix;
  dcg_eprint (" (");
  for (ix = 0; ix < fpars -> size; ix++)
    { fpar fp = fpars -> array[ix];
      string aname = fp -> fdef -> adef -> aname;
      if (ix) dcg_eprint (", ");
      switch (fp -> fdir)
	{ case d_inherited: dcg_eprint (">P%d: %s", ix, aname); break;
	  case d_derived:   dcg_eprint ("*P%d: %s>", ix, aname); break;
	  case d_transient: dcg_eprint (">*P%d: %s>", ix, aname); break;
	  default: dcg_bad_tag (fp -> fdir, "dump_proc_args");
	};
    };
  dcg_eprint (")");
}

/*
   Dumping of proc locals
*/
static void dump_proc_locals (variable_list pvars)
{ int ix;
  for (ix = 0; ix < pvars -> size; ix++)
    { variable pvar = pvars -> array[ix];
      if (pvar -> dont_code)
	dcg_wlog (" deleted local %d: %s (snr = %d)", ix, pvar -> vname, pvar -> nr);
      dcg_wlog (" local %d: %s (snr = %d)", ix, pvar -> vname, pvar -> nr);
    };
}

/*
   Dumping of addresses
*/
static void dump_address (address addr)
{ switch (addr -> tag)
    { case TAGParameter:	dcg_eprint ("P%d", addr -> Parameter.pnr); break;
      case TAGGlobal:		dcg_eprint ("G%d", addr -> Global.gnr);  break;
      case TAGLocal:		dcg_eprint ("L%d", addr -> Local.lnr);   break;
      case TAGIlocal:		dcg_eprint ("I%d", addr -> Ilocal.inr);  break;
      case TAGTconst:		dcg_eprint ("TC%d", addr -> Tconst.cnr); break;
      case TAGIndirect:		dcg_eprint ("*"); dump_address (addr -> Indirect.adr); break;
      case TAGReference:	dcg_eprint ("&"); dump_address (addr -> Reference.adr); break;
      case TAGPart:		dump_address (addr -> Part.adr);
				dcg_eprint (".vpart[%d]", addr -> Part.idx); break;
      default: dcg_bad_tag (addr -> tag, "dump_address");
    };
}

/*
   Dumping of local expressions
*/
static void dump_operator (operator op)
{ switch (op)
    { case times:		dcg_eprint ("*"); break;
      case modulo:		dcg_eprint ("%"); break;
      case divides:		dcg_eprint ("/"); break;
      case plus:		dcg_eprint ("+"); break;
      case minus:		dcg_eprint ("-"); break;
      case shift_left:		dcg_eprint ("<<"); break;
      case shift_right:		dcg_eprint (">>"); break;
      case bitwise_and:		dcg_eprint ("&"); break;
      case bitwise_or:		dcg_eprint ("|"); break;
      case bitwise_xor:		dcg_eprint ("^"); break;
      case bitwise_not:		dcg_eprint ("~"); break;
      case int_times_int:	dcg_eprint ("*"); break;
      case real_times_real:	dcg_eprint ("*"); break;
      case int_times_text:	dcg_eprint ("*"); break;
      case int_plus_int:	dcg_eprint ("+"); break;
      case real_plus_real:	dcg_eprint ("+"); break;
      case text_plus_text:	dcg_eprint ("+"); break;
      case int_minus_int:	dcg_eprint ("-"); break;
      case real_minus_real:	dcg_eprint ("-"); break;
      default: dcg_bad_tag (op, "dump_operator");
    };
}

static void dump_iexpr (iexpr exp)
{ switch (exp -> tag)
    { case TAGIvref:
	dump_address (exp -> Ivref.adr);
	dcg_eprint (".int");
	break;
      case TAGIdyop:
	dcg_eprint ("(");
	dump_iexpr (exp -> Idyop.src1);
	dcg_eprint (" ");
	dump_operator (exp -> Idyop.dop);
	dump_iexpr (exp -> Idyop.src2);
	dcg_eprint (")");
	break;
      case TAGImonop:		
	dump_operator (exp -> Imonop.mop);
	dump_iexpr (exp -> Imonop.src);
	break;
      case TAGInum:
	dcg_eprint ("#%d", exp -> Inum.num);
	break;
      default: dcg_bad_tag (exp -> tag, "dump_iexpr");
    };
}

static void dump_rexpr (rexpr exp)
{ switch (exp -> tag)
    { case TAGRvref:
	dump_address (exp -> Rvref.adr);
	dcg_eprint (".real");
	break;
      case TAGRdyop:
	dcg_eprint ("(");
	dump_rexpr (exp -> Rdyop.src1);
	dcg_eprint (" ");
	dump_operator (exp -> Rdyop.dop);
	dump_rexpr (exp -> Rdyop.src2);
	dcg_eprint (")");
	break;
      case TAGRmonop:		
	dump_operator (exp -> Rmonop.mop);
	dump_rexpr (exp -> Rmonop.src);
	break;
      case TAGRnum:
	dcg_eprint ("#%g", exp -> Rnum.num);
	break;
      default: dcg_bad_tag (exp -> tag, "dump_rexpr");
    };
}

/*
   Dumping of instructions
*/
static void dump_address_list (address_list args)
{ int ix;
  dcg_eprint (" (");
  for (ix = 0; ix < args -> size; ix++)
    { if (ix) dcg_eprint (", ");
      dump_address (args -> array[ix]);
    };
  dcg_eprint (")");
}

static void dump_call (int pnr, address_list args)
{ proc pdef = all_procedures -> array[pnr];
  dump_proc_name (pdef);
  dump_address_list (args);
}

static void dump_label (int lnr)
{ dcg_eprint ("LAB%d", lnr);
}

static void dump_case (altnr_label_pair cl)
{ dcg_eprint ("(%d, LAB%d)", cl -> altnr, cl -> lnr);
}

static void dump_cases_list (altnr_label_pair_list cl)
{ int ix;
  dcg_eprint ("[");
  for (ix = 0; ix < cl -> size; ix++)
    { if (ix) dcg_eprint (", ");
      dump_case (cl -> array[ix]);
    };
  dcg_eprint ("]");
}

static void dump_instruction (instr ins)
{ if (ins -> dont_code) dcg_eprint ("/* NC ");

  if (ins -> tag == TAGLabel) dump_label (ins -> Label.lnr);
  else dcg_eprint ("\t");

  switch (ins -> tag)
    { case TAGNop:		dcg_eprint ("nop");		break;
      case TAGProc_entry:	dcg_eprint ("entry");	break;
      case TAGLabel:		dcg_eprint (":");		break;
      case TAGComment:		dcg_eprint ("/* %s */", ins -> Comment.text); break;
      case TAGSucceed:		dcg_eprint ("succeed");	break;
      case TAGFail:		dcg_eprint ("fail");	break;
      case TAGExit:		dcg_eprint ("exit");	break;
      case TAGAbort:		dcg_eprint ("abort, module %d, line %d",
					ins -> Abort.mnr, ins -> Abort.lnr);
				break;
      case TAGAttach:
	dcg_eprint ("attach ");
	dump_address (ins -> Attach.loc);
	dcg_eprint (", tnr=%d", ins -> Attach.tnr);
	break;
      case TAGDetach:
	dcg_eprint ("detach ");
	dump_address (ins -> Detach.loc);
	dcg_eprint (", tnr=%d", ins -> Detach.tnr);
	break;
      case TAGCreate:
	dcg_eprint ("create ");
	dump_address (ins -> Create.dst);
	dcg_eprint (", altnr=%d, size=%d", ins -> Create.altnr, ins -> Create.size);
	break;
      case TAGEqual_altnr:
	dcg_eprint ("equal_altnr ");
	dump_address (ins -> Equal_altnr.src);
	dcg_eprint (", altnr=%d failing to ", ins -> Equal_altnr.altnr);
	dump_label (ins -> Equal_altnr.fail);
	break;
      case TAGEqual:
	dcg_eprint ("equal ");
	dump_address (ins -> Equal.src1);
	dcg_eprint (", ");
	dump_address (ins -> Equal.src2);
	dcg_eprint (", tnr=%d failing to ", ins -> Equal.tnr);
	dump_label (ins -> Equal.fail);
	break;
      case TAGEqual_int:
	dcg_eprint ("equal_int ");
	dump_address (ins -> Equal_int.src);
	dcg_eprint (", ");
	dump_iexpr (ins -> Equal_int.cmp);
	dcg_eprint (" failing to ");
	dump_label (ins -> Equal_int.fail);
	break;
      case TAGEqual_real:
	dcg_eprint ("equal_real ");
	dump_address (ins -> Equal_real.src);
	dcg_eprint (", ");
	dump_rexpr (ins -> Equal_real.cmp);
	dcg_eprint (" failing to ");
	dump_label (ins -> Equal_real.fail);
	break;
      case TAGCopy:	
        dcg_eprint ("copy ");
	dump_address (ins -> Copy.src);
	dcg_eprint (" to ");
	dump_address (ins -> Copy.dst);
	break;
      case TAGCopy_int:
	dcg_eprint ("copy int ");
	dump_iexpr (ins -> Copy_int.src);
	dcg_eprint (" to ");
	dump_address (ins -> Copy_int.dst);
	break;
      case TAGCopy_real:
	dcg_eprint ("copy real ");
	dump_iexpr (ins -> Copy_int.src);
	dcg_eprint (" to ");
	dump_address (ins -> Copy_int.dst);
	break;
      case TAGCallp:	
	dump_call (ins -> Callp.pnr, ins -> Callp.args);
	break;
      case TAGCallf:
	dump_call (ins -> Callf.pnr, ins -> Callf.args);
	dcg_eprint (" failing to ");
	dump_label (ins -> Callf.fail);
	break;
      case TAGIndex:
	dcg_eprint ("copy array ref ");
	dump_address (ins -> Index.glb);
	dcg_eprint ("{");
	dump_iexpr (ins -> Index.idx);
	dcg_eprint ("} to ");
	dump_address (ins -> Index.iloc);
	break;
      case TAGConc_text:
	dcg_eprint ("concat texts ");
	dump_address_list (ins -> Conc_text.srcs);
	dcg_eprint (" to ");
	dump_address (ins -> Conc_text.dst);
	break;
      case TAGSwitch_altnr:
	dcg_eprint ("switch_altnr ");
	dump_address (ins -> Switch_altnr.src);
	dcg_eprint (", ");
	dump_cases_list (ins -> Switch_altnr.cases);
	dcg_eprint (" failing to ");
	dump_label (ins -> Switch_altnr.fail);
	break;
      case TAGSwitch_int:
	dcg_eprint ("switch_int ");
	dump_address (ins -> Switch_int.src);
	dcg_eprint (", ");
	dump_cases_list (ins -> Switch_int.cases);
	dcg_eprint (" failing to ");
	dump_label (ins -> Switch_altnr.fail);
	break;
      default: dcg_bad_tag (ins -> tag, "dump_instruction");
    };
  if (ins -> dont_code) dcg_eprint (" */");
  dcg_eprint ("\n");
}

static int index_from_block (block_list bbs, block bb)
{ int ix;
  for (ix = 0; ix < bbs -> size; ix++)
    if (bb == bbs -> array[ix]) return (ix);
  return (-1);
}

static void dump_basic_block_nr (block_list bbs, block bb)
{ int bnr = index_from_block (bbs, bb);
  dcg_eprint ("%d", bnr);
}

static void dump_basic_block_list (block_list bbs, block_list dfl)
{ int ix;
  if (dfl == block_list_nil)
    { dcg_eprint ("<null>");
      return;
    };
  for (ix = 0; ix < dfl -> size; ix++)
    { if (ix) dcg_eprint (", ");
      dump_basic_block_nr (bbs, dfl -> array[ix]);
    };
}

static void dump_live_list (char_list locs)
{ int ix;
  for (ix = 0; ix < locs -> size; ix++)
    dcg_eprint (" %d", locs -> array[ix]);
}

static void dump_basic_block (block_list bbs, int ix)
{ block bb = bbs -> array[ix];
  instr ins;
  dcg_eprint ("\n\t/* Basic block %d", ix);
  if (bb -> dont_code)
    { dcg_eprint (", deleted by optimization */\n");
      return;
    };
  
  dcg_eprint ("\n\t   Pred: ");
  dump_basic_block_list (bbs, bb -> pred);
  dcg_eprint (", TrueSucc: ");
  dump_basic_block_list (bbs, bb -> true_succ);
  dcg_eprint (", FalseSucc: ");
  dump_basic_block_nr (bbs, bb -> false_succ);
  if (bb -> gen != char_list_nil)
    { dcg_eprint ("\n\t   Gen:");
      dump_live_list (bb -> gen);
    };
  if (bb -> use != char_list_nil)
    { dcg_eprint ("\n\t   Use:");
      dump_live_list (bb -> use);
    };
  if (bb -> callf_gen != char_list_nil)
    { dcg_eprint ("\n\t   Callf Gen:");
      dump_live_list (bb -> callf_gen);
    };
  if (bb -> callf_use != char_list_nil)
    { dcg_eprint ("\n\t   Callf Use:");
      dump_live_list (bb -> callf_use);
    };
  if (bb -> in != char_list_nil)
    { dcg_eprint ("\n\t   In: ");
      dump_live_list (bb -> in);
    }
  if (bb -> out != char_list_nil)
    { dcg_eprint ("\n\t   Out:");
      dump_live_list (bb -> out);
    }
  dcg_eprint ("\n\t*/\n");
  for (ins = bb -> head; ins != instr_nil; ins = ins -> next)
    dump_instruction (ins);
}

void dump_proc (proc pdef)
{ int ix;
  dcg_wlog ("/* Code for module %s, %s %s */", all_modules -> array[pdef -> mnr] -> mname,
	    string_from_proc_type (pdef -> pspec -> ptype), pdef -> pspec -> canonic_name);
  dump_proc_name (pdef);
  dump_proc_args (pdef);
  dcg_wlog (":");
  dump_proc_locals (pdef -> pvars);
  for (ix = 0; ix < pdef -> bbs -> size; ix++)
    dump_basic_block (pdef -> bbs, ix);
  dcg_wlog ("");
}

void dump_intermediate_code ()
{ int ix;
  for (ix = 0; ix < all_procedures -> size; ix++)
    { proc pdef = all_procedures -> array[ix];
      if (!pdef -> called) continue;
      if (pdef -> tag != TAGExt_proc)
	dump_proc (pdef);
    };
}

#define INIT_LABEL_SIZE 1024
void init_imc_utils ()
{ label_table = init_block_list (INIT_LABEL_SIZE);
}
