/*
   File: intermed.c
   Does intermediate code generation

   Copyright (C) 2008 Marc Seutter

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

   CVS ID: "$Id: intermed.c,v 1.18 2011/09/02 10:11:23 marcs Exp $"
*/

/* global includes */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

/* libdcg includes */
#include <dcg.h>
#include <dcg_error.h>
#include <dcg_string.h>

/* local includes */
#include "cdl_ds.h"
#include "modules.h"
#include "contsens.h"
#include "ast_utils.h"
#include "imc_utils.h"
#include "globals.h"
#include "affix_rules.h"
#include "variables.h"
#include "intermed.h"
#include "options.h"
#include "coder.h"

/*
   Variables are either global or local.

   Local variables can be in one of 4 states:
   a) undefined: No value at all.
   b) constant:  A constant value has been copied to the local. No attach needed, no detach needed
   c) derived:	 The variable has a temporary shortlived value which is a copy from another.
		 No attach done, no detach needed (This would be rather desastrous).
   d) attached:  The variable has a long living value, either by attaching or creation.
		 Upon relinquishing the value, it must be detached (else memory leak)

   During intermediate code generation, we keep track of the state of all local variables
*/

/*
   While translating, symbolic addresses will be generated. These can be one
   of the following:

      Parameter pnr:	A parameter passed by stack to a call
      Global gnr:	A global variable
      Local lnr:	A local variable
      Ilocal inr:	An indirect short lived variable used in array references
      Tconst cnr:	A text constant
      Indirect adr:	An indirection (*adr)
      Reference adr:	A reference to an adr (&adr)
      Part (adr,idx):	A part of a composed value (*adr.idx)
*/

/*
   The three different kinds of parameters are treated in the following way:

   a) inherited: The caller must assure that a defined attached value is passed
		 to the callee. The callee may use this value as a derived value.

   b) derived:   Upon success, the callee copies a defined value back to the caller.
		 Hence, the callee must ensure that the value is attached.

		 If the call cannot fail, the caller may detach the original value
		 before the call, if necessary (mostly not)

		 If it can fail, and the original value is an attached value, it
		 must copy the variable to a transfer variable, mark the current
		 one as a derived value and then call. Afterward, the recovery
		 algorithm will insert detaches on the safe variable if necessary

   c) transient: The caller ensures that the value is attached before the call.
		 The callee ensures that the value which is copied back upon success
		 is attached. The callee also detaches the original value before
		 copying back
*/
static void copy_inherited_formal_parameters (proc pdef)
{ state_list visibility = init_state_list (pdef -> pvars -> size);
  fpar_list fpars = pdef -> pspec -> p1fpars;
  block current = pdef -> bbs -> array[0];
  int ix;
  for (ix = 0; ix < fpars -> size; ix++)
    { /* Note: Parameter ix gets copied to local ix */
      dir fdir = fpars -> array[ix] -> fdir;
      address src = address_nil;
      instr copy;
      switch (fdir)
	{ case d_derived:  
	    set_local_state (visibility, ix, undefined);
	    continue;
	  case d_transient:
            set_local_state (visibility, ix, derived);
	    src = new_Indirect (new_Parameter (ix));
	    break;
	  case d_inherited:
            set_local_state (visibility, ix, derived);
	    src = new_Parameter (ix);
	    break;
	  default: dcg_bad_tag (fdir, "copy_inherited_formal_parameters");
	};
      copy = new_Copy (src, new_Local (ix));
      append_instr (current, copy);
    };

  /* Register the current sets: they are not yet in their final state */
  current -> visibility = visibility;
}

static void copy_back_derived_parameters (proc pdef, block current, member m)
{ /* state_list visibility = init_state_list (pdef -> pvars -> size); */
  fpar_list fpars = pdef -> pspec -> p1fpars;
  int ix;
  for (ix = 0; ix < fpars -> size; ix++)
    { /* Note: Parameter ix gets copied to local ix */
      fpar arg = fpars -> array[ix];
      affix_rule arule = get_repr_affix_rule (arg -> fdef -> adef);
      address src, dst;
      state visibility;
      instr copy;

      /* Inherited arguments should not be copied back */
      if (arg -> fdir == d_inherited) continue;

      /* Note: ensure that we have an attached value */
      src = new_Local (ix);
      dst = new_Indirect (new_Parameter (ix));
      visibility = get_local_state (current -> visibility, ix);
      switch (visibility)
	{ case attached:
	    set_local_state (current -> visibility, ix, derived);
	    break;
	  case derived:
	    if (arule -> needs_gc)
	      { instr attach = new_Attach (src, arule -> anr);
      		append_instr (current, attach);
	      };
	  case constant: break;
	  case undefined:
	    contsens_error_by_mnr (pdef -> mnr, m -> line, m -> col,
				   "Derived affix %s has not received a value upon succeeding",
				   arg -> fname);
	    break;
	  default: dcg_bad_tag (visibility, "copy_back_derived_parameters");
	};

      /*
	 MS: This fixes the bug reported by Albert.
	 When the argument is a transient, it has been attached before
	 the call, so it must be detached before overwriting with a
	 newly attached value
      */
      if (arule -> needs_gc && (arg -> fdir == d_transient))
	{ instr detach = new_Detach (dst, arule -> anr);
	  append_instr (current, detach);
	};
      copy = new_Copy (src, dst);
      append_instr (current, copy);
    };
}

/*
   Try to attach and detach
*/
void try_attach_variable (block bb, address var_adr, affix_rule arule)
{ affix_rule adef = get_repr_affix_rule (arule);
  if (!adef -> needs_gc) return;
  if (var_adr -> tag == TAGLocal)
    { int nr = var_adr -> Local.lnr;
      state os = get_local_state (bb -> visibility, nr);
      switch (os)
        { case constant: return;
	  case derived:
	  case attached: break;
	  default: dcg_bad_tag (os, "try_attach_variable");
	};
    };
  append_instr (bb, new_Attach (var_adr, adef -> anr));
}

void try_detach_variable (block bb, address var_adr, affix_rule arule)
{ affix_rule adef = get_repr_affix_rule (arule);
  if (!adef -> needs_gc) return;
  if (var_adr -> tag == TAGLocal)
    { int nr = var_adr -> Local.lnr;
      state os = get_local_state (bb -> visibility, nr);
      if (os != attached) return;
      set_local_state (bb -> visibility, nr, undefined);
    };
  append_instr (bb, new_Detach (var_adr, adef -> anr));
}

/*
   translate_int_expr translates an affix expression of INT type into an iexpr
   (which is directly codable in C as it does not need garbage collection)
*/
static address translate_variable (affix_term term, int dst, block *current, proc pdef,
				   int *is_global);
static iexpr translate_int_expr (affix_term term, block *current, proc pdef)
{ int is_global;
  switch (term -> tag)
    { case TAGDyop:
	{ iexpr src1 = translate_int_expr (term -> Dyop.arg1, current, pdef);
	  iexpr src2 = translate_int_expr (term -> Dyop.arg2, current, pdef);
	  return (new_Idyop (term -> Dyop.dop, src1, src2));
	};
      case TAGMonop:
	{ iexpr src = translate_int_expr (term -> Monop.arg, current, pdef);
	  return (new_Imonop (term -> Monop.mop, src));
	};
      case TAGInt: return (new_Inum (term -> Int.num));
      case TAGVar:
      case TAGSub: return (new_Ivref (translate_variable (term, 0, current, pdef, &is_global)));
      default: dcg_bad_tag (term -> tag, "translate_int_expr");
    };
  return (iexpr_nil);
}

/*
   translate_real_expr translates an affix expression of REAL type into a rexpr
   (which is directly codable in C as it does not need garbage collection)
*/
static rexpr translate_real_expr (affix_term term, block *current, proc pdef)
{ int is_global;
  switch (term -> tag)
    { case TAGDyop:
	{ rexpr src1 = translate_real_expr (term -> Dyop.arg1, current, pdef);
	  rexpr src2 = translate_real_expr (term -> Dyop.arg2, current, pdef);
	  return (new_Rdyop (term -> Dyop.dop, src1, src2));
	};
      case TAGMonop:
	{ rexpr src = translate_real_expr (term -> Monop.arg, current, pdef);
	  return (new_Rmonop (term -> Monop.mop, src));
	};
      case TAGReal: return (new_Rnum (term -> Real.num));
      case TAGVar:
      case TAGSub: return (new_Rvref (translate_variable (term, 0, current, pdef, &is_global)));
      default: dcg_bad_tag (term -> tag, "translate_real_expr");
    };
  return (rexpr_nil);
}

/*
   Two support routines to add new anonymous affix variables
*/
static address new_text_local (proc pdef)
{ affix_rule text_type = get_primitive_type (prim_text);
  variable new_local = new_local_var (text_type);
  int lnr = pdef -> pvars -> size;
  app_variable_list (pdef -> pvars, new_local);
  return (new_Local (lnr));
}

/*
   translate_text_expression translates an affix term of TEXT type in such
   a way that the resulting value can be found in the location *dst_adr. 
   This location is either determined by the context or a new anonymous
   location is generated and returned as a result parameter.

   Intermediate variables are marked as such: replication and concatenation
   will always yield new attached) values. A special marking is reserved for
   constants, which never need to be detached.
*/
static void translate_text_expr (affix_term term, block *current, proc pdef,
				 address *dst_adr, int *is_global_dst)
{ switch (term -> tag)
    { case TAGDyop:
	{ iexpr fact = translate_int_expr (term -> Dyop.arg1, current, pdef);
	  address src_adr = address_nil;
          int is_global_src;
	  instr repl;
	  if (*dst_adr == address_nil)
	    { *dst_adr = new_text_local (pdef);
	      *is_global_dst = 0;
	    };
	  translate_text_expr (term -> Dyop.arg2, current, pdef, &src_adr, &is_global_src);
	  repl = new_Repl_text (fact, src_adr, *dst_adr);
	  append_instr (*current, repl);
	  update_variable (*current, *dst_adr, repl, attached);
	}; break;
      case TAGConcat:
	{ int adrs_size = term -> Concat.terms -> size;
	  address_list adrs = init_address_list (adrs_size);
	  instr concat;
	  int ix;

	  if (*dst_adr == address_nil)
	    { *dst_adr = new_text_local (pdef);
	      *is_global_dst = 0;
	    };
	  for (ix = 0; ix < adrs_size; ix++)
	    { affix_term cterm = term -> Concat.terms -> array[ix];
	      address cterm_adr = address_nil;
	      int cterm_global;
	      translate_text_expr (cterm, current, pdef, &cterm_adr, &cterm_global);
	      app_address_list (adrs, cterm_adr);
	    };
	  concat = new_Conc_text (adrs, *dst_adr);
	  append_instr (*current, concat);
	  update_variable (*current, *dst_adr, concat, attached);
	}; break;
      case TAGText:
	{ int cnr = enter_constant (-1, term -> Text.txt);
	  address const_adr = new_Tconst (cnr);
	  if (*dst_adr == address_nil)
	    { *dst_adr = const_adr;
	      *is_global_dst = 0;
	    }
	  else 
	    { instr copy = new_Copy (const_adr, *dst_adr);
	      append_instr (*current, copy);
	      update_variable (*current, *dst_adr, copy, constant);
	    };
	}; break;
      case TAGVar:
      case TAGSub:
	{ int var_is_global;
	  address var_adr = translate_variable (term, 0, current, pdef, &var_is_global);
	  if (*dst_adr == address_nil)
	    { *dst_adr = var_adr;
	      *is_global_dst = var_is_global;
	    }
	  else 
	    { instr copy = new_Copy (var_adr, *dst_adr);
	      append_instr (*current, copy);
	      update_variable (*current, *dst_adr, copy, derived);
	    };
	}; break;
      default: dcg_bad_tag (term -> tag, "translate_text_expr");
    };
};

static address translate_index (affix_term term, block *current, proc pdef)
{ address glb = new_Global (term -> Sub.vdef -> nr);
  iexpr t_index = translate_int_expr (term -> Sub.index, current, pdef);
  address iloc = new_Ilocal (pdef -> ivars);
  instr access = new_Index (glb, t_index, iloc);
  append_instr (*current, access);
  pdef -> ivars++;
  return (iloc);
};

/*
   Translate_variable is responsible for translating an affix variable in
   such a way that it ends up in a global or local affix. This may either
   be the affix variable self or an indirect reference to an indexed array.
*/
static address translate_variable (affix_term term, int dst, block *current, proc pdef,
				   int *is_global)
{ *is_global = 0;
  switch (term -> tag)
    { case TAGVar:
	{ variable vdef = term -> Var.vdef;
	  switch (vdef -> tag)
	    { case TAGLocal_var:
		{ int nr = vdef -> nr;
		  if (!dst && (get_local_state ((*current) -> visibility, nr) == undefined))
		    contsens_error_by_mnr (pdef -> mnr, term -> line, term -> col,
				           "Variable %s is used with an undefined value",
				           vdef -> vname);
		  return (new_Local (nr));
		};
	      case TAGModule_var:
		*is_global = 1;
		return (new_Global (vdef -> nr));
	      default: dcg_bad_tag (vdef -> tag, "translate_variable");
	    };
	};
      case TAGSub:
	*is_global = 1;
	return (new_Indirect (translate_index (term, current, pdef)));
      default: dcg_bad_tag (term -> tag, "translate_variable");
    };
  return (address_nil);
}

/*
   Operator translation
*/
static void detach_attached_variables (block current, proc pdef)
{ state_list sl = current -> visibility;
  int ix;
  for (ix = 0; ix < sl -> size; ix++)
    if (sl -> array[ix] == attached)
      { variable pvar = pdef -> pvars -> array[ix];
	affix_rule adef = get_repr_affix_rule (pvar -> adef);
	instr detach;
	if (!adef -> needs_gc) continue;
	detach = new_Detach (new_Local (ix), adef -> anr);
	append_instr (current, detach);
      };
}

static void translate_operator (member m, block *current, proc pdef)
{ instr op = instr_nil;
  switch (m -> Op.op)
    { case f_success:
	copy_back_derived_parameters (pdef, *current, m);
	detach_attached_variables (*current, pdef);
	op = new_Succeed ();
	break;
      case f_failure:
	detach_attached_variables (*current, pdef);
	op = new_Fail ();
	break;
      case f_exit:    op = new_Exit (); break;
      case f_abort:   op = new_Abort (pdef -> mnr, m -> line); break;
      default: dcg_bad_tag (m -> Op.op, "translate_operator");
    };
  append_instr (*current, op);
}

/*
   The translation of the equal is simple. Determine locally derived copies
   of the operands (These may be actual local variables themselves) and test
   their equality by calling the appropriate type equal routine

   T ([ VARA = VARB ]):
     Lleft := Tdc (VARA);
     Lright := Tdc (VARB);
     EQUAL (Lleft, Lright, typeof (VARA), fail_label)

   We do not try to attach the locally derived copies.
*/

static void translate_equal_guard (member m, block *current, proc pdef)
{ affix_term src1 = m -> Res_equal.left;
  affix_term src2 = m -> Res_equal.right;
  int is_global1;
  int is_global2;
  address adr1 = translate_variable (src1, 0, current, pdef, &is_global1);
  address adr2 = translate_variable (src2, 0, current, pdef, &is_global2);
  int tnr = get_repr_affix_rule (src1 -> adef) -> anr;
  instr equal = new_Equal (adr1, adr2, tnr, -1);
  append_instr (*current, equal);
  connect_true_successor (current, pdef);
}

/*
   For the translation of the assign guard, we always attach the source
   and then detach the destination (if necessary). After the copy the
   destination is marked as attached.

   T ([ VARA -> VARB ]):
     Lleft := Tdc (VARA);
     Lright := Tdc (VARB);
     if (typeof (VARA) needs gc)
       { Note: we do not attach constants/undefineds
	 if (VARA is derived or attached or global)
	   ATTACH (Lleft, typeof (VARA))
	 if (VARB is attached or global)
	   DETACH (Lright, typeof (VARA))
       }
     COPY (Lleft, Lright)

*/
static void translate_assign (member m, block *current, proc pdef)
{ int is_global_src, is_global_dst;
  address src_adr = translate_variable (m -> Assign.src, 0, current, pdef, &is_global_src);
  address dst_adr = translate_variable (m -> Assign.dst, 1, current, pdef, &is_global_dst);
  affix_rule arule = get_repr_affix_rule (m -> Assign.src -> adef);
  instr copy = new_Copy (src_adr, dst_adr);

  if (arule -> needs_gc)
    { try_attach_variable (*current, src_adr, arule);
      try_detach_variable (*current, dst_adr, arule);
    };
  append_instr (*current, copy);
  update_variable (*current, dst_adr, copy, attached);
}

static void translate_split (member m, block *current, proc pdef)
{ int is_global_src, ix;
  affix_term src = m -> Res_guard.var;
  affix_term_list vars = m -> Res_guard.expr;
  affix_rule arule = get_repr_affix_rule (src -> adef);
  address src_adr = translate_variable (src, 0, current, pdef, &is_global_src);

  /* Check for an enumerable split */
  if (arule -> tag != TAGAffix_alts) dcg_internal_error ("translate_split");
  if (arule -> Affix_alts.enumerable)
    { instr equal;
      if (!m -> Res_guard.can_fail) return;
      equal = new_Equal_int (src_adr, new_Inum (m -> Res_guard.altnr), -1);
      append_instr (*current, equal);
      connect_true_successor (current, pdef);
      return;
    };

  /*
     Non enumerable split, check the alternative nr (if necessary)
  */
  if (m -> Res_guard.can_fail)
    { instr equal = new_Equal_altnr (src_adr, m -> Res_guard.altnr, -1);
      append_instr (*current, equal);
      connect_true_successor (current, pdef);
    };

  /*
     Copy the subparts and attach them initially
     When the second pass is implemented some of the attaches will be deleted.
     We now attach objects of which it is not certain that they will be necessary.
  */
  for (ix = 0; ix < vars -> size; ix++)
    { affix_rule dst_arule = vars -> array[ix] -> adef;
      int is_global_dst; 
      address dst_adr = translate_variable (vars -> array[ix], 1, current, pdef, &is_global_dst);
      instr copy_part;
      if (dst_arule -> needs_gc)
	try_detach_variable (*current, dst_adr, dst_arule);
      copy_part = new_Copy (new_Part (src_adr, ix + 1), dst_adr);
      append_instr (*current, copy_part);
      if (dst_arule -> needs_gc)
        { instr attach = new_Attach (dst_adr, dst_arule -> anr);
          append_instr (*current, attach);
          update_variable (*current, dst_adr, copy_part, attached);
	}
      else update_variable (*current, dst_adr, copy_part, derived);
    };
}

static void translate_join (member m, block *current, proc pdef)
{ int is_global_dst;
  affix_term dst = m -> Res_guard.var;
  affix_term_list vars = m -> Res_guard.expr;
  affix_rule arule = get_repr_affix_rule (dst -> adef);
  address dst_adr = translate_variable (dst, 1, current, pdef, &is_global_dst);
  instr create;
  int ix;
  
  /* Check for an enumerable join */
  if (arule -> tag != TAGAffix_alts) dcg_internal_error ("translate_join");
  if (arule -> Affix_alts.enumerable)
    { /* Update visibility and assignment set */
      instr copy = new_Copy_int (new_Inum (m -> Res_guard.altnr), dst_adr);
      update_variable (*current, dst_adr, copy, derived);
      append_instr (*current, copy);
      return;
    };

  /*
     Non enumerable join, create new value of appropriate size. Before
     creation, the destination is detached. The check_procedures3 code
     must have checked beforehand, that for a guard of the form
     [ M v1...vn -> V ] that none of the vi is equal to V. 

     When a join is made with 0 members, an appropriate constant is
     allocated in the constant table and used as the source of the join.
  */
  try_detach_variable (*current, dst_adr, arule);
  if (vars -> size == 0)
    { int cnr = enter_constant (m -> Res_guard.altnr, string_nil);
      address const_adr = new_Tconst (cnr);
      instr copy = new_Copy (const_adr, dst_adr);
      append_instr (*current, copy);
      update_variable (*current, dst_adr, copy, constant);
      return;
    };

  create = new_Create (m -> Res_guard.altnr, vars -> size + 1, dst_adr);
  append_instr (*current, create);
  update_variable (*current, dst_adr, create, attached);
  for (ix = 0; ix < vars -> size; ix++)
    { affix_rule src_arule = vars -> array[ix] -> adef;
      int is_global_src; 
      address src_adr = translate_variable (vars -> array[ix], 0, current, pdef, &is_global_src);
      instr copy_part;
      if (src_arule -> needs_gc && (get_variable_state (*current, src_adr) != constant))
        { instr attach = new_Attach (src_adr, src_arule -> anr);
          append_instr (*current, attach);
	};
      copy_part = new_Copy (src_adr, new_Part (dst_adr, ix + 1));
      append_instr (*current, copy_part);
    };
}

static void translate_arith_split (member m, block *current, proc pdef)
{ affix_term var = m -> Res_guard.var;
  affix_term src = m -> Res_guard.expr -> array[0];
  affix_rule arule = get_repr_affix_rule (var -> adef);
  int is_global_var;
  address var_adr = translate_variable (var, 0, current, pdef, &is_global_var);
  instr equal = instr_nil;
  if (is_integer_type (arule))
    { iexpr t_src = translate_int_expr (src, current, pdef);
      equal = new_Equal_int (var_adr, t_src, -1);
    } 
  else if (is_real_type (arule))
    { rexpr t_src = translate_real_expr (src, current, pdef);
      equal = new_Equal_real (var_adr, t_src, -1);
    }
  else if (is_text_type (arule))
    { address src_adr = address_nil;
      int is_global_src;
      translate_text_expr (src, current, pdef, &src_adr, &is_global_src);
      equal = new_Equal (var_adr, src_adr, arule -> anr, -1);
    }
  else
    { contsens_error_by_mnr (pdef -> mnr, m -> line, m -> col, "This should not happen");
      dcg_panic ("Illegal primitive expression in arithmetic split");
    };
  append_instr (*current, equal);
  connect_true_successor (current, pdef);
}

static void translate_arith_join (member m, block *current, proc pdef)
{ affix_term dst = m -> Res_guard.var;
  affix_term src = m -> Res_guard.expr -> array[0];
  affix_rule arule = get_repr_affix_rule (dst -> adef);
  int is_global_dst;
  address dst_adr = translate_variable (dst, 1, current, pdef, &is_global_dst);
  instr copy = instr_nil;
  if (is_integer_type (arule))
    { iexpr t_src = translate_int_expr (src, current, pdef);
      copy = new_Copy_int (t_src, dst_adr);
    }
  else if (is_real_type (arule))
    { rexpr t_src = translate_real_expr (src, current, pdef);
      copy = new_Copy_real (t_src, dst_adr);
    }
  else if (is_text_type (arule))
    { translate_text_expr (src, current, pdef, &dst_adr, &is_global_dst);
      return;
    }
  else
    { contsens_error_by_mnr (pdef -> mnr, m -> line, m -> col, "This should not happen");
      dcg_panic ("Illegal primitive expression in arithmetic join");
    };
  append_instr (*current, copy);
  update_variable (*current, dst_adr, copy, derived);
}

static void translate_guard (member m, block *current, proc pdef)
{ switch (m -> Res_guard.typ)
    { case split:	translate_split (m, current, pdef); break;
      case join:	translate_join (m, current, pdef); break;
      case arith_split:	translate_arith_split (m, current, pdef); break;
      case arith_join:	translate_arith_join (m, current, pdef); break;
      default: dcg_bad_tag (m -> Res_guard.typ, "translate_guard");
    };
}

/*
   Call translation
   The translation of the call is easy, because it only involves
   local affixes. Upon failure, none of the derived or transient
   values is overwritten. However in the second pass, detaches
   may be inserted on the fail path towards the next alternative
*/
static void prepare_inherited_call_parameters (proc pdef, member m, block bb,
					       fpar_list fpars, address_list args)
{ int ix;
  for (ix = 0; ix < fpars -> size; ix++)
    { address arg = args -> array[ix];
      fpar fp = fpars -> array[ix];
      variable var = variable_nil;
      affix_rule arule = get_repr_affix_rule (fp -> fdef -> adef);
      switch (arg -> tag)
	{ case TAGReference: var = pdef -> pvars -> array[arg -> Reference.adr -> Local.lnr]; break;
	  case TAGLocal:     var = pdef -> pvars -> array[arg -> Local.lnr]; break;
	  default: dcg_bad_tag (arg -> tag, "prepare_inherited_call_parameters");
        };

      switch (fp -> fdir)
	{ case d_inherited:
	    { switch (get_variable_state (bb, arg))
	        { case derived:
		    if (arule -> needs_gc)
		      { instr attach = new_Attach (arg, arule -> anr);
      			append_instr (bb, attach);
			update_variable (bb, arg, attach, attached);
		      }; break;
		  case undefined:
	      	    contsens_error_by_mnr (pdef -> mnr, m -> line, m -> col,
			             "Actual parameter %s has not received a value before call",
			             var -> vname);
		  default: break;
		};
	    }; break;
	  case d_transient:
	    { switch (get_variable_state (bb, arg -> Reference.adr))
		{ case undefined:
	            contsens_error_by_mnr (pdef -> mnr, m -> line, m -> col,
				     "Actual parameter %s has not received a value before call",
				     var -> vname);
		  case derived:
		    if (arule -> needs_gc)
		      { instr attach = new_Attach (arg -> Reference.adr, arule -> anr);
      			append_instr (bb, attach);
			update_variable (bb, arg -> Reference.adr, attach, attached);
		      };
		  default: break;
		};
	    };
	  case d_derived: break;
	  default: dcg_bad_tag (fp -> fdir, "prepare_inherited_call_parameters");
	};
    };
}

static void replace_attached_derived_parameters (proc pdef, member m, block bb, fpar_list fpars,
						 address_list args, assoc_pair_list *al)
{ int ix;
  for (ix = 0; ix < fpars -> size; ix++)
    { address arg = args -> array[ix];
      fpar fp = fpars -> array[ix];
      affix_rule arule = get_repr_affix_rule (fp -> fdef -> adef);
      variable auxvar, pvar;
      string auxvname;
      int vnr, auxnr;

      /* Check if this actual parameter is a derived parameter with an attach value */
      if (fp -> fdir != d_derived) continue;
      if (arg -> tag != TAGReference)
	dcg_bad_tag (arg -> tag, "replace_attached_derived_parameters");
      arg = arg -> Reference.adr;
      vnr = arg -> Local.lnr;
      if (!arule -> needs_gc) continue;
      if (get_variable_state (bb, arg) != attached) continue;

      /*
	 Derived parameter had an attached value, which must be kept over the call
         either when failing or to be detached after the call.
	 Create an auxilary variable which replaces the original derived parameter
	 Upon success, we copy its value to the former derived value while detaching
	 the old value
      */
      pvar = pdef -> pvars -> array[vnr];
      if (*al == assoc_pair_list_nil) *al = new_assoc_pair_list ();
      auxnr = pdef -> pvars -> size;
      auxvname = dcg_new_fmtd_string ("%s$auxi%d", pvar -> vname, auxnr);
      auxvar = new_Local_var (auxvname);
      auxvar -> adef = pvar -> adef;
      auxvar -> nr = auxnr;
      app_variable_list (pdef -> pvars, auxvar);
      app_assoc_pair_list (*al, new_assoc_pair (pvar -> nr, auxnr));
      arg -> Local.lnr = auxnr;
    };
  if (*al != assoc_pair_list_nil)
    contsens_hint_by_mnr (pdef -> mnr, m -> line, m -> col, "Restructured call");
}

static int is_a_derived_copy (assoc_pair_list copies, int arg_nr, int *orig_nr)
{ int ix;
  if (copies == assoc_pair_list_nil) return (0);
  for (ix = 0; ix < copies -> size; ix++)
    { assoc_pair ap = copies -> array[ix];
      if (ap -> auxnr != arg_nr) continue;
      *orig_nr = ap -> lnr;
      return (1);
    };
  return (0);
}
      
static void recover_or_mark_derived_parameters (proc pdef, block bb, instr call, fpar_list fpars,
						address_list args, assoc_pair_list copies)
{ int ix;
  for (ix = 0; ix < fpars -> size; ix++)
    { fpar fp = fpars -> array[ix];
      address arg = args -> array[ix];
      int lnr, orignr;
      if (fp -> fdir != d_derived) continue;
      lnr = arg -> Reference.adr -> Local.lnr;
      if (is_a_derived_copy (copies, lnr, &orignr))
        { /* Copy auxilary to original derived actual parameter */
	  variable dvar = pdef -> pvars -> array[orignr];
	  affix_rule arule = get_repr_affix_rule (dvar -> adef);
	  instr detach = new_Detach (new_Local (orignr), arule -> anr);
	  instr copy = new_Copy (new_Local (lnr), new_Local (orignr));
          append_instr (bb, detach);
          append_instr (bb, copy);
	  /* Aux was never marked, derived actual was marked attached, stays attached */
	}
      else update_variable (bb, arg -> Reference.adr, call, attached);
    };
}

static void translate_call (member m, block *current, proc pdef)
{ int call_always_succeeds = (m -> false_succ == member_nil);
  affix_term_list acts = m -> Res_call.p1args;
  proc called_pdef = m -> Res_call.pdef;
  fpar_list fpars = called_pdef -> pspec -> p1fpars;
  int pnr = called_pdef -> pnr;
  address_list args = init_address_list (fpars -> size);
  assoc_pair_list derived_copies = assoc_pair_list_nil;
  instr call;
  int ix;

  /*
     Make sure that args exist or are detached/attached at right places
  */
  for (ix = 0; ix < fpars -> size; ix++)
    { affix_term act = acts -> array[ix];
      fpar fp = fpars -> array[ix];
      dir fdir = fp -> fdir;
      address arg;

      if (act -> tag != TAGVar) dcg_internal_error ("translate_call");
      arg = new_Local (act -> Var.vdef -> nr);
      if (fdir != d_inherited) arg = new_Reference (arg);
      app_address_list (args, arg);
    };

  /* Generate right instruction for call */
  prepare_inherited_call_parameters (pdef, m, *current, fpars, args);
  replace_attached_derived_parameters (pdef, m, *current, fpars, args, &derived_copies);
  if (call_always_succeeds)
    { call = new_Callp (pnr, args);
      append_instr (*current, call);
    }
  else
    { call = new_Callf (pnr, args, -1);
      append_instr (*current, call);
      connect_true_successor (current, pdef);
    };
  recover_or_mark_derived_parameters (pdef, *current, call, fpars, args, derived_copies);
}

static void translate_group (group grp, block *current, proc pdef);
static void translate_member (member m, block *current, proc pdef)
{ switch (m -> tag)
    { case TAGOp:		translate_operator (m, current, pdef); break;
      case TAGGroup:		translate_group (m -> Group.grp, current, pdef); break;
      case TAGRes_equal:	translate_equal_guard (m, current, pdef); break;
      case TAGAssign:		translate_assign (m, current, pdef); break;
      case TAGRes_guard:	translate_guard (m, current, pdef); break;
      case TAGRes_call:		translate_call (m, current, pdef);
      case TAGGroup_fail:
      case TAGCast:		break;
      default: dcg_bad_tag (m -> tag, "translate_member");
    };
}

static void translate_members (member_list mems, block *current, proc pdef)
{ int ix;
  for (ix = 0; ix < mems -> size; ix++)
    translate_member (mems -> array[ix], current, pdef);
}

/*
   The following code is responsible for copying some local variables
   to auxilaries to enable their recovery upon failing
*/
static void insert_variable_saving (block current, assoc_pair_list recovery, proc pdef)
{ int ix;
  if (recovery == assoc_pair_list_nil) return;
  for (ix = 0; ix < recovery -> size; ix++)
    { assoc_pair p = recovery -> array[ix];
      variable src = pdef -> pvars -> array[p -> lnr];
      affix_rule arule = get_repr_affix_rule (src -> adef);
      address src_adr = new_Local (p -> lnr);
      address dst_adr = new_Local (p -> auxnr);
      instr copy = new_Copy (src_adr, dst_adr);
      append_instr (current, copy);
      update_variable (current, dst_adr, copy, derived);
      if (arule -> needs_gc)
        { try_attach_variable (current, dst_adr, arule);
	  update_variable (current, dst_adr, copy, attached);
	};
    };
}

/*
   Generation of recover code
*/
static int is_a_recovery_variable (assoc_pair_list rec_list, int auxnr)
{ int ix;
  if (rec_list == assoc_pair_list_nil) return (0);
  for (ix = 0; ix < rec_list -> size; ix++)
    if (rec_list -> array[ix] -> auxnr == auxnr) return (1);
  return (0);
}

static void generate_block_recover_code (block alt_start, block last, state_list grp_visibility,
					 int *fail_label, block *fail_tgt_bb, proc pdef, int rec,
					 assoc_pair_list rec_list)
{ /*
     Note all blocks between alt_start and current must have failed,
     current being the last block of the alternative.
  */
  block current = last;
  int iy;
  while (current != alt_start)
    { state_list prev_visibility, curr_visibility;
      instr last_ins;
      int ix, size;

      /*
	 One block back, pick up its end visibility and determine its predecessor visibility
	 In the current, variables may have become attached that were not before.
	 These then are candidates for detaches.
      */
      current = current -> pred -> array[0];
      curr_visibility = current -> visibility;
      if (current == alt_start) prev_visibility = grp_visibility;
      else prev_visibility = current -> pred -> array[0] -> visibility;
      size = curr_visibility -> size;
      if (prev_visibility -> size > size) size = prev_visibility -> size;

      /*
	 All predecessor blocks must have failed in some way
	 If not already present, create a block and fail label, to fail to
      */
      last_ins = current -> tail;
      if (!instr_has_fail_label (last_ins))
        dcg_internal_error ("generate_block_recover_code");

      /*
	 If we have not generated a fail label or if the basic block
	 to which we have been failing already contains recover code
	 we definitely need a new basic block
      */
      if ((*fail_label == -1) || ((*fail_tgt_bb) -> tail -> tag != TAGLabel))
	/* Definitely a new block needed */
	connect_new_fail_successor (current, *fail_tgt_bb, pdef, fail_label, fail_tgt_bb);
      else /* Else we can reuse the fail label */
	connect_fail_successor (current, *fail_tgt_bb);

      /* Fill in the fail label in the last instruction of the failing basic block */
      set_fail_label (last_ins, *fail_label);

      /* Now fill fail_tgt_bb with the differences in visibility, if required */
      if (!rec) continue;
      for (ix = 0; ix < size; ix++)
	{ state os = get_local_state (prev_visibility, ix);
	  state ns = get_local_state (curr_visibility, ix);
	  variable pvar = pdef -> pvars -> array[ix];
	  affix_rule arule = get_repr_affix_rule (pvar -> adef);
	  if (is_a_recovery_variable (rec_list, ix)) continue;
	  if ((ns == attached) && (os != attached))
	    try_detach_variable (*fail_tgt_bb, new_Local (ix), arule);
	  if (ns != os)
	    set_local_state ((*fail_tgt_bb) -> visibility, ix, os);
	};
    };

  /* Copy back the recovery variables, but only in the final failing block */
  if (rec_list == assoc_pair_list_nil) return;
  for (iy = 0; iy < rec_list -> size; iy++)
    { assoc_pair rp = rec_list -> array[iy];
      variable orgvar = pdef -> pvars -> array[rp -> lnr];
      affix_rule arule = get_repr_affix_rule (orgvar -> adef);
      instr copy_back = new_Copy (new_Local (rp -> auxnr), new_Local (rp -> lnr));
      append_instr (*fail_tgt_bb, copy_back);
      set_local_state ((*fail_tgt_bb) -> visibility, rp -> auxnr, derived);
      if (arule -> needs_gc)
	set_local_state ((*fail_tgt_bb) -> visibility, rp -> lnr, attached);
    };
}

/*
   Generation of the alternative recover code
   Note we do not need to recover when the next member is an explicit fail or
   group fail, but we still need the fail basic block.
*/
static void generate_alt_recover_code (alternative alt, block *current, block alt_start,
				       state_list grp_visibility, proc pdef, int recover)
{ assoc_pair_list rec_list = alt -> recovery;
  member_list mems = alt -> p1mems;
  block fail_tgt_bb = block_nil;
  block fw_curr = *current;
  int fail_label = -1;

  /* Recovery for last member if it is a group */
  member last_mem = mems -> array[mems -> size - 1];
  if (last_mem -> tag == TAGGroup)
    { group grp = last_mem -> Group.grp;
      alternative_list group_alts = grp -> alts;
      alternative last_alt = group_alts -> array[group_alts -> size - 1];
      member last_group_mem = last_alt -> p1mems -> array[0];
      fw_curr = grp -> start;
      if (last_group_mem -> tag == TAGGroup_fail)
	{ fail_label = (*current) -> head -> Label.lnr;
	  fail_tgt_bb = *current;
        };
    };

  /* We must restore the state as it was at the start of the alternative */
  generate_block_recover_code (alt_start, fw_curr, grp_visibility,
			       &fail_label, &fail_tgt_bb, pdef, recover, rec_list);

  /* After possible value recovery, we must have a basic block to fail to */
  if (fail_tgt_bb == block_nil)
    dcg_internal_error ("generate_alt_recover_code:3");
  *current = fail_tgt_bb;
}

/*
   The translation of an alternative results in a set of basic blocks
   who are connected tail to head through the true successors. Between
   two consecutive alternatives, recover code is inserted to restore
   the local variables to the state they had before entering the first
   alternative. For this purpose the assignment set and visibility set
   is saved at the beginning of the alternatives.
*/
static void translate_group (group grp, block *alts_current, proc pdef)
{ alternative_list alts = grp -> alts;
  block alt_start = *alts_current;
  int ix;

  grp -> start = alt_start;
  for (ix = 0; ix < alts -> size; ix++)
    { alternative alt = alts -> array[ix];
      block current = alt_start;
      int last_fail = 0;
      if (ix < alts -> size - 1)
	{ member fail_mem = alts -> array[ix + 1] -> p1mems -> array[0];
	  if ((fail_mem -> tag == TAGOp) && (fail_mem -> Op.op == f_failure))
	    last_fail = 1;

	  /* Check the next condition */
	  if (fail_mem -> tag == TAGGroup_fail)
	    last_fail = 1;
	};
      if (!last_fail)
	insert_variable_saving (current, alt -> recovery, pdef);
      translate_members (alt -> p1mems, &current, pdef);

      if (ix < alts -> size - 1)
        { generate_alt_recover_code (alt, &current, alt_start, grp -> visibility, pdef, 1);
	  alt_start = current;
	};
    };
  *alts_current = alt_start;
}

#define INIT_BBS 16
static void translate_proc (proc pdef)
{ block current = new_block ();
  pdef -> bbs = init_block_list (INIT_BBS);
  app_block_list (pdef -> bbs, current);
  append_instr (current, new_Proc_entry ());
  copy_inherited_formal_parameters (pdef);
  if (pdef -> tag == TAGAlts) translate_group (pdef -> Alts.grp, &current, pdef);
  else translate_members (pdef -> Root.mems, &current, pdef);
}

static void translate_procedures ()
{ int ix;
  for (ix = 0; ix < all_procedures -> size; ix++)
    { proc pdef = all_procedures -> array[ix];
      if (!pdef -> called) continue;
      if (pdef -> tag != TAGExt_proc)
	translate_proc (pdef);
    };
}

void intermediate_code_generation ()
{ dcg_warning (0, "   generating intermediate code...");
  init_imc_utils ();
  init_coder ();
  translate_procedures ();
  if (dump_intermed) dump_intermediate_code ();
  dcg_panic_if_errors ();
};
