/*
   File: lexer.c
   Defines a lexer for CDL3

   Copyright (C) 2008 Marc Seutter

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

   CVS ID: "$Id: lexer.c,v 1.9 2013/03/13 12:56:10 marcs Exp $"
*/

/* global includes */
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>

/* libdcg includes */
#include <dcg.h>
#include <dcg_error.h>
#include <dcg_string.h>
#include <dcg_plist.h>
#include <dcg_plist_ops.h>

/* local includes */
#include "options.h"
#include "cdl_ds.h"
#include "lexer.h"
#include "incr_error.h"

/* Lexer administration */
#define MAXBUFLEN 1024
static char line_buffer[MAXBUFLEN];
static FILE *lex_fd;
static string lex_fname;
static int lex_line;
static int lex_col;

static void lexer_error (char *format, ...)
{ char buf[MAXSTRLEN];
  va_list arg_ptr;
  va_start (arg_ptr, format);
  vsprintf (buf, format, arg_ptr);
  va_end (arg_ptr);
  dcg_error (0, "File %s, line %d, col %d: %s", lex_fname, lex_line, lex_col + 1, buf);
  maybe_enter_editor (lex_fname, lex_line, lex_col + 1);
};

static void lexer_error_at (int line, int col, char *format, ...)
{ char buf[MAXSTRLEN];
  va_list arg_ptr;
  va_start (arg_ptr, format);
  vsprintf (buf, format, arg_ptr);
  va_end (arg_ptr);
  dcg_error (0, "File %s, line %d, col %d: %s", lex_fname, line, col + 1, buf);
  maybe_enter_editor (lex_fname, line, col + 1);
};

static void lexer_warning_at (int line, int col, char *format, ...)
{ char buf[MAXSTRLEN];
  va_list arg_ptr;
  va_start (arg_ptr, format);
  vsprintf (buf, format, arg_ptr);
  va_end (arg_ptr);
  dcg_warning (0, "File %s, line %d, col %d: %s", lex_fname, line, col + 1, buf);
};

/*
   Position interface
   The column returned is one higher than our position in the line
*/
void get_current_filename (string *rfname)
{ *rfname = lex_fname;
}

void get_current_position (int *rline, int *rcol)
{ *rline = lex_line;
  *rcol = lex_col + 1;
}

void get_erroneous_name (string *rname)
{ *rname = new_string ("<ERROR>");
}

/*
   We read on a per line basis
*/
static void read_next_line ()
{ char *rbuf;
  if (line_buffer[0] == '\0') return;
  lex_line++;
  lex_col = 0;
  rbuf = fgets (line_buffer, MAXBUFLEN - 1, lex_fd);
  if (rbuf == NULL) line_buffer[0] = '\0';	/* Mark EOF */
}

int end_of_file ()
{ return (line_buffer[0] == '\0');
}

static int end_of_line ()
{ return (line_buffer[lex_col] == '\n');
}

/*
   Recognition of character classes
*/
static int is_char (char ch)
{ if (line_buffer [lex_col] != ch) return (0);
  lex_col++;
  return (1);
}

static char get_next_char ()
{ if (end_of_line ()) return ('\0');
  return (line_buffer [lex_col++]);
}

static int is_white_space ()
{ if (end_of_line ())
    { read_next_line ();
      return (1);
    }
  if (is_char ('\r')) return (1);
  if (is_char ('\t')) return (1);
  return (is_char (' '));
}

static int is_embedded_white_space ()
{ if (is_white_space ()) return (1);
  return (is_char ('_'));
}

static int is_digit (char *rch)
{ char ch = line_buffer [lex_col];
  if (('0' <= ch) && (ch <= '9'))
    { lex_col++;
      *rch = ch;
      return (1);
    };
  return (0);
}

static int is_octal_digit (int *rdig)
{ char ch = line_buffer [lex_col];
  if (('0' <= ch) && (ch <= '7'))
    { lex_col++;
      *rdig = (int) (ch - '0');
      return (1);
    };
  return (0);
}

static int is_hex_digit (int *rdig)
{ char ch = line_buffer [lex_col];
  if (('0' <= ch) && (ch <= '9'))
    { lex_col++;
      *rdig = (int) (ch - '0');
      return (1);
    };
  if (('A' <= ch) && (ch <= 'F'))
    { lex_col++;
      *rdig = (int) (ch - 'A') + 10;
      return (1);
    };
  if (('a' <= ch) && (ch <= 'f'))
    { lex_col++;
      *rdig = (int) (ch - 'a') + 10;
      return (1);
    };
  return (0);
}

static int is_small_letter (char *rch)
{ char ch = line_buffer [lex_col];
  if (('a' <= ch) && (ch <= 'z'))
    { lex_col++;
      *rch = ch;
      return (1);
    };
  return (0);
}

static int is_capital_letter (char *rch)
{ char ch = line_buffer [lex_col];
  if (('A' <= ch) && (ch <= 'Z'))
    { lex_col++;
      *rch = ch;
      return (1);
    };
  return (0);
}

static int is_module_letgit (char *rch)
{ if (is_small_letter (rch)) return (1);
  if (is_digit (rch)) return (1);
  if (!is_char ('_')) return (0);
  *rch = '_';
  return (1);
}

static int is_affix_letund (char *rch)
{ if (is_capital_letter (rch)) return (1);
  if (!is_char ('_')) return (0);
  *rch = '_';
  return (1);
}

static int ahead_rest_real_denoter ()
{ if (line_buffer[lex_col] == '.')
    return (('0' <= line_buffer[lex_col + 1]) && (line_buffer[lex_col + 1] <= '9'));
  if ((line_buffer[lex_col] != 'E') && (line_buffer[lex_col] != 'e')) return (0);
  if (('0' <= line_buffer[lex_col + 1]) && (line_buffer[lex_col + 1] <= '9')) return (1);
  if ((line_buffer[lex_col+1] != '+') && (line_buffer[lex_col + 1] != '-')) return (0);
  return (('0' <= line_buffer[lex_col + 2]) && (line_buffer[lex_col + 2] <= '9'));
}

/*
   Layout and comments
*/
static int is_comment ()
{ if (!is_char ('#')) return (0);
  while (!end_of_line () && (get_next_char () != '#')) ;
  return (1);
}

static void may_be_layout ()
{ while (is_white_space () || is_comment ()) ;
}

static void may_be_embedded_layout ()
{ while (is_embedded_white_space () || is_comment ()) ;
}

/*
   Recognition of fixed symbols
*/
static int my_ahead_symbol (char *sy, int *rcol)
{ char *sptr = sy;
  int col = lex_col;
  for (; *sptr; sptr++, col++)
    if (*sptr != line_buffer[col])
      return (0);
  *rcol = col;
  return (1);
}

int ahead_symbol (char *sy)
{ int col;
  return (my_ahead_symbol (sy, &col));
}

int is_symbol (char *sy)
{ int col;
  if (!my_ahead_symbol (sy, &col)) return (0);
  lex_col = col;
  may_be_layout ();
  return (1);
}

void should_be_symbol (char *sy)
{ if (is_symbol (sy)) return;
  lexer_error ("symbol %s expected", sy);
}

int is_keyword (char *keyword)
{ char nch;
  int col;
  if (!my_ahead_symbol (keyword, &col)) return (0);
  nch = line_buffer [col];
  if (('A' <= nch) && (nch <= 'Z')) return (0);
  lex_col = col;
  may_be_layout ();
  return (1);
}

/*
   Primary token recognizers
*/
#define MAXINTDIV10 214748364
#define MAXINTMOD10 7
int is_number (int *kind, int *inum, real *rnum)
{ char my_buf[MAXBUFLEN];
  int tcol, ix;
  int mloc = 0;
  char dig;
  if (!is_digit (&dig)) return (0);
  tcol = lex_col;
  my_buf[mloc++] = dig;
  while (is_digit (&dig))
    my_buf[mloc++] = dig;
  if (!ahead_rest_real_denoter ())
    { /* We have an integer denoter, convert it to INT */
      int value = 0;
      for (ix = 0; ix < mloc; ix++)
	{ int dval = (int)(my_buf[ix] - '0');
	  if ((value > MAXINTDIV10) ||
	      ((value == MAXINTDIV10) && (dval > MAXINTMOD10)))
	    { lexer_error_at (lex_line, tcol, "Integer overflow");
	      break;
	    };
	  value = value * 10 + dval;
	};

      *inum = value;
      *kind = 0;
    }
  else
    { /* We have the first part of a real denoter, read the rest */
      int digs_after_period = 0;
      real value = 0.0;
      int exponent = 0;
      int exp_sign = 0;
      if (is_char ('.'))
        { /* Parse rest of digits and remember their count */
	  while (is_digit (&dig))
	    { my_buf[mloc++] = dig;
	      digs_after_period++;
	    };
	};

      /* Check for exponent */
      if (is_char ('E') || is_char ('e'))
        { if (is_char ('-')) exp_sign = 1;
	  else if (is_char ('+')) ;
	  while (is_digit (&dig))
	    { int dval = dig - '0';
	      tcol = lex_col;
	      if ((exponent > MAXINTDIV10) ||
	          ((exponent == MAXINTDIV10) && (dval > MAXINTMOD10)))
		{ lexer_error_at (lex_line, tcol, "Exponent overflow");
		  break;
		};
	      exponent = exponent * 10 + dval;
	    };
        };
      if (exp_sign) exponent = -exponent;
      exponent -= digs_after_period;

      /* Convert to real */
      for (ix = 0; ix < mloc; ix++)
	{ real dval = (real)(int)(my_buf[ix] - '0');
	  value = value * 10.0 + dval;
	};

      *rnum = value * pow (10.0, exponent);
      *kind = 1;
    };
  may_be_layout ();
  return (1);
}

/*
   Text constants
*/
void should_be_text_constant (string *text)
{ if (is_text_constant (text)) return;
  lexer_error ("text constant expected");
  *text = string_nil;
}

int is_text_constant (string *text)
{ char my_buf[MAXBUFLEN];
  int mloc = 0;
  int tloc = lex_col;
  int value;
  int cnv;
  if (!is_char ('\"')) return (0);
  while (1)
    { if (end_of_line ())
        { lexer_error_at (lex_line, tloc, "unterminated text constant");
	  break;
	};
      if (is_char ('\"')) break;
      if (is_char ('\\'))
	{ if (is_char ('\\')) my_buf[mloc++] = '\\';
	  else if (is_char ('\"')) my_buf[mloc++] = '\"';
	  else if (is_char ('a')) my_buf[mloc++] = '\a';
	  else if (is_char ('b')) my_buf[mloc++] = '\b';
	  else if (is_char ('f')) my_buf[mloc++] = '\f';
	  else if (is_char ('n')) my_buf[mloc++] = '\n';
	  else if (is_char ('r')) my_buf[mloc++] = '\r';
	  else if (is_char ('t')) my_buf[mloc++] = '\t';
	  else if (is_char ('v')) my_buf[mloc++] = '\v';
	  else if (is_char ('x'))
	    { if (is_hex_digit (&cnv))
		{ value = cnv;
		  if (is_hex_digit (&cnv))
		    value = 16 * value + cnv;
		}
	      else
		{ lexer_error_at (lex_line, lex_col, "hexadecimal digit expected");
		  value = 42;
		};
	      if (!value) lexer_warning_at (lex_line, tloc, "null character in text constant");
	      my_buf[mloc++] = (char)(unsigned char) value;
	    }
	  else if (is_octal_digit (&cnv))
	    { value = cnv;
	      if (is_octal_digit (&cnv))
		{ value = 8 * value + cnv;
		  if (is_octal_digit (&cnv))
		    value = 8 * value + cnv;
		};
	      if (!value) lexer_warning_at (lex_line, tloc, "null character in text constant");
	      my_buf[mloc++] = (char)(unsigned char) value;
	    }
	  else
	    { my_buf[mloc++] = get_next_char ();
	      lexer_error ("Illegal escape sequence");
	    };
	}
      else my_buf[mloc++] = get_next_char ();
    };
  my_buf[mloc] = '\0';
  *text = new_string (my_buf);
  may_be_layout ();
  return (1);
}

/*
   Name parts should start with a small letter
   followed by small letters; the question is whether
   we could allow more to follow
*/
int is_name_part (string *part)
{ char my_buf[MAXBUFLEN];
  int mloc = 0;
  char mch;
  if (!is_small_letter (&mch)) return (0);
  my_buf[mloc++] = mch;
  while (is_small_letter (&mch))
    my_buf[mloc++] = mch;
  my_buf[mloc] = '\0';
  *part = new_string (my_buf);
  may_be_embedded_layout ();
  return (1);
}

/*
   Module names should start with a small letter
   followed by small letters, digits or underscores.
*/
void should_be_module_name (string *mname)
{ if (is_module_name (mname)) return;
  get_erroneous_name (mname);
  lexer_error ("module name expected");
}

int is_module_name (string *mname)
{ char my_buf[MAXBUFLEN];
  int mloc = 0;
  char mch;
  if (!is_small_letter (&mch)) return (0);
  my_buf[mloc++] = mch;
  while (is_module_letgit (&mch))
    my_buf[mloc++] = mch;
  my_buf[mloc] = '\0';
  *mname = new_string (my_buf);
  may_be_layout ();
  return (1);
}

/*
   Name parts should start with a small letter
   followed by small letters; the question is whether
   we could allow more to follow
*/
int is_affix_terminal (string *aname)
{ char my_buf[MAXBUFLEN];
  int mloc = 0;
  char mch;
  if (!is_small_letter (&mch)) return (0);
  my_buf[mloc++] = mch;
  while (is_small_letter (&mch))
    my_buf[mloc++] = mch;
  my_buf[mloc] = '\0';
  *aname = new_string (my_buf);
  may_be_layout ();
  return (1);
}

/*
   Affix names are capital letter sequences
   Affix variables have an optional integer suffix
*/
void should_be_affix_name (string *aname)
{ if (is_affix_name (aname)) return;
  get_erroneous_name (aname);
  lexer_error ("affix name expected");
}

int is_affix_name (string *aname)
{ char my_buf[MAXBUFLEN];
  int mloc = 0;
  char mch;
  if (!is_capital_letter (&mch)) return (0);
  my_buf[mloc++] = mch;
  while (is_affix_letund (&mch))
    my_buf[mloc++] = mch;
  my_buf[mloc] = '\0';
  *aname = new_string (my_buf);
  may_be_layout ();
  return (1);
}

void should_be_affix_variable_name (string *aname)
{ if (is_affix_variable_name (aname)) return;
  get_erroneous_name (aname);
  lexer_error ("affix variable expected");
}

int is_affix_variable_name (string *aname)
{ char my_buf[MAXBUFLEN];
  int mloc = 0;
  char mch;
  if (!is_capital_letter (&mch)) return (0);
  my_buf[mloc++] = mch;
  while (is_affix_letund (&mch))
    my_buf[mloc++] = mch;
  while (is_digit (&mch))
    my_buf[mloc++] = mch;
  my_buf[mloc] = '\0';
  *aname = new_string (my_buf);
  may_be_layout ();
  return (1);
}

/*
   Initialize the lexical scanner with a new file
*/
void init_lexer (FILE *fd, char *fname)
{ if (fd == NULL) dcg_internal_error ("init_lexer");
  lex_fd = fd;

  /* Reset token administration */
  lex_fname = new_string (fname);
  lex_line = 0;
  lex_col = 0;
  strcpy (line_buffer, "\n");
  may_be_layout ();
};

void finish_lexer ()
{ /* dcg_detach (&fname) */
  fclose (lex_fd);
}
