/*
   File: match.c
   Tries to match a list of terms against a type
   This code is for checking guards and resolving calls

   Copyright (C) 2008 Marc Seutter

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

   CVS ID: "$Id: match.c,v 1.14 2012/04/21 14:00:00 marcs Exp $"
*/

/* standard includes */
#include <stdio.h>
#include <string.h>

/* support lib includes */
#include <dcg.h>
#include <dcg_alloc.h>
#include <dcg_error.h>
#include <dcg_string.h>

/* local includes */
#include "cdl_ds.h"
#include "ast_utils.h"
#include "globals.h"
#include "affix_rules.h"
#include "variables.h"
#include "contsens.h"
#include "options.h"
#include "match.h"

/*
   Our backup parsing administration is managed through a continuation stack
   on which member administration alternates with rule goals. Parse_arule is
   responsible for parsing a rule topdown, that is it pushes a rule goal and
   then iterates through all rules which are in a (reflexive transitive) left
   corner relation to try and parse a left prefix of its rule in a bottom up
   fashion. Upon a succesfull partial parse of left prefix, the reduce_arule
   will continue the parse.

   The parameter complete specifies the parse must be complete (end of sentence match):
   0 (no eos required), 1 (total eos required), 2 (local (Enclosed terms) eos required).

   The returned list of guards indicates the structure of the match:
   For a join they will appear in their natural order, for a split, they
   are in reverse. Note also that all markers will have been removed
   from a resolved guard, except for the alternative number.

   Note: all parsing routines return 0 (no parse), 1 (one success), >1 (definitely ambiguous)
*/
static int parse_alternative (int curr_mnr, affix_term_list terms, int iptr, guard_type type,
			      affix_rule lc_rule, int alt_nr, int from_mem, madm_list rstack,
			      affix_term_list tstack, member_list gstack, member_list *res_guards);
static int parse_members (int curr_mnr, affix_term_list terms, int iptr, guard_type type,
			  madm_list rstack, affix_term_list tstack, member_list gstack,
			  member_list *res_guards);
static int reduce_arule (int curr_mnr, affix_term_list terms, int iptr, guard_type type,
			 madm_list rstack, affix_term_list tstack, member_list gstack,
			 member_list *res_guards);
static int parse_arule (int curr_mnr, affix_term_list terms, int iptr, affix_rule arule,
			guard_type type, int complete, madm_list rstack, affix_term_list tstack,
			member_list gstack, member_list *res_guards)
{ /* Create new rule admin and goal */
  affix_rule goal = get_prime_affix_rule (arule);
  int_list lc_nrs = arule -> lc_nrs;
  madm new_top = new_Rule_adm ();
  int stat = 0;
  int lc_nr;

  /* Register the new top of stack */
  new_top -> Rule_adm.complete = complete;
  new_top -> Rule_adm.terms = terms;
  new_top -> Rule_adm.iptr = iptr;
  new_top -> Rule_adm.goal = goal;
  app_madm_list (rstack, new_top);

  /* Add an extra internal consistency check */
  if (lc_nrs == int_list_nil)
    { dcg_wlog ("Goal %s, no left corners available", goal -> aname);
      dcg_internal_error ("parse_arule");
    };

  /* Iterate thru list of left corner nonterminals to get a left corner */
  for (lc_nr = 0; lc_nr < lc_nrs -> size; lc_nr++)
    { affix_rule lc_rule = all_affix_rules -> array[lc_nrs -> array[lc_nr]];
      affix_alternative_list alts = pick_rhs_affix_alts (lc_rule);
      int alt_nr;
      if (alts == affix_alternative_list_nil)
	{ dcg_wlog ("Goal %s, LC %s", goal -> aname, lc_rule -> aname);
	  dcg_internal_error ("parse_arule: 2");
	};
      for (alt_nr = 0; alt_nr < alts -> size; alt_nr++)
	stat += parse_alternative (curr_mnr, terms, iptr, type, lc_rule, alt_nr, 0, rstack,
				   tstack, gstack, res_guards);
    };

  /* Remove what we have pushed */
  del_madm_list (rstack, rstack -> size - 1);
  det_madm (&new_top);
  return (stat);
}

static int parse_alternative (int curr_mnr, affix_term_list terms, int iptr, guard_type type,
			      affix_rule lc_rule, int alt_nr, int from_mem, madm_list rstack,
			      affix_term_list tstack, member_list gstack, member_list *res_guards)
{ /* Create member administration for this alternative and start member */
  madm new_top = new_Member_adm (alt_nr, from_mem);
  int stat;
  new_top -> Member_adm.arule = lc_rule;
  app_madm_list (rstack, new_top);

  /* Call the essential parse_members */
  stat = parse_members (curr_mnr, terms, iptr, type, rstack, tstack, gstack, res_guards);

  /* Remove what we have pushed */
  del_madm_list (rstack, rstack -> size - 1);
  det_madm (&new_top);
  return (stat);
}

static int parse_members (int curr_mnr, affix_term_list terms, int iptr, guard_type type,
			  madm_list rstack, affix_term_list tstack, member_list gstack,
			  member_list *res_guards)
{ int mem_nr, alt_nr, cline, ccol, ix;
  affix_term curr_term, new_tree;
  affix_alternative_list alts;
  affix_term_list guard_parts;
  affix_element_list elems;
  affix_rule arule, vrule;
  affix_element curr_elem;
  madm top_madm, goal_madm; 
  member local_guard;
  int rstat = 0;
 
  /* Consistency check */
  if (rstack -> size < 2) dcg_internal_error ("parse_members: rstack_size");
  top_madm = rstack -> array[rstack -> size - 1];
  goal_madm = rstack -> array[rstack -> size - 2];
  if (top_madm -> tag != TAGMember_adm) dcg_internal_error ("parse_members: top_madm");
  if (goal_madm -> tag != TAGRule_adm) dcg_internal_error ("parse_members: goal_adm");
  alt_nr = top_madm -> Member_adm.alt_nr;
  mem_nr = top_madm -> Member_adm.mem_nr;
  arule = top_madm -> Member_adm.arule;
  alts = pick_rhs_affix_alts (arule);
  elems = alts -> array[alt_nr] -> elems;

  /* Check for a complete partial parse, if so, reduce and continue */
  if (mem_nr == elems -> size)
    { /* Check if we are the right hand side of an abstract type */
      affix_rule orig_arule;
      if (is_abstract_synonym (arule, &orig_arule, &vrule))
	{ /* Term stack does not need a change, only top one gets a type change */
	  /* We remember that we did change the type by putting a cast on the guard stack */
	  affix_term top_term = tstack -> array [tstack -> size - 1];
	  member cast = new_Cast (top_term -> line, top_term -> col, attach_affix_term (top_term));
	  affix_rule old_adef = top_term -> adef;
	  if (top_term -> tag != TAGVar) dcg_internal_error ("parse_members:top_term");
	  top_term -> adef = arule;

	  /* Check for occurrence */
	  if (strict_export && (orig_arule -> mnr != curr_mnr))
	    contsens_warning_by_mnr (curr_mnr, top_term -> line, top_term -> col,
				     "This guard should only be used in the defining module");

          /* Try to reduce the rest */
          app_member_list (gstack, cast);
          del_madm_list (rstack, rstack -> size - 1);
          rstat = reduce_arule (curr_mnr, terms, iptr, type, rstack, tstack, gstack, res_guards);
          app_madm_list (rstack, top_madm);
          delete_member_list (gstack, gstack -> size - 1);

	  /* Restore the changes */
	  top_term -> adef = old_adef;
	}
      else
	{ /* We have a successfull partial parse, construct local guard and new tree */
          variable new_local;
          int tidx, gline, gcol;
	  int can_fail = (type == split) && (alts -> size > 1);

          /* Pop parts from tree stack and push as guard parts */
          guard_parts = new_affix_term_list ();
          for (ix = elems -> size - 1; 0 <= ix; ix--)
	    if (elems -> array[ix] -> tag == TAGAffix_var)
	      { affix_term part = tstack -> array [tstack -> size - 1];
	        del_affix_term_list (tstack, tstack -> size - 1);
	        ins_affix_term_list (guard_parts, 0, attach_affix_term (part));
	      };

          /* Locate position of first element */
          tidx = iptr - elems -> size;
          if (tidx < 0) dcg_internal_error ("parse_members: pos");
          gline = terms -> array[tidx] -> line;
          gcol = terms -> array[tidx] -> col;

          /* Last part popped is first part of alternative, construct local guard */
          new_tree = new_local_affix (gline, gcol, arule);
	  new_local = new_tree -> Var.vdef;
          local_guard = new_Res_guard (gline, gcol, type, new_tree, alt_nr, can_fail, guard_parts);

          /* Push guard and new tree */
          app_member_list (gstack, local_guard);
          app_affix_term_list (tstack, new_tree);

          if (dump_affix_rules)
            { dcg_wlog ("Constructed local %s, arule %s, alt %d",
		        string_from_guard_type (type), arule -> aname, alt_nr);
	      pp_member (dcg_error_file (), local_guard);
	      dcg_wlog ("");
	    };

          /* Try to reduce the rest */
          del_madm_list (rstack, rstack -> size - 1);
          rstat = reduce_arule (curr_mnr, terms, iptr, type, rstack, tstack, gstack, res_guards);
          app_madm_list (rstack, top_madm);

          /* Pop new tree and guard */
          del_affix_term_list (tstack, tstack -> size - 1);
          delete_member_list (gstack, gstack -> size - 1);

          /* Push all parts back on tree stack to undo; then destruct guard */
          for (ix = 0; ix < guard_parts -> size; ix++)
	    app_affix_term_list (tstack, guard_parts -> array[ix]);
          detach_member (&local_guard);
          if (!rstat) detach_variable (&new_local);
	};
      return (rstat);
    };

  /*
     There is still a member to recognize, so we cannot succeed
     if we are already at the end of input
  */
  if (iptr >= terms -> size) return (0);

  /* If the current element to be parsed is an affix terminal, try so */
  curr_term = terms -> array[iptr];
  curr_elem = elems -> array[mem_nr];
  if (curr_elem -> tag == TAGAffix_term)
    { /*
	 Match with the current element at the iptr
	 but only if we are also matching a marker
      */
      if (curr_term -> tag != TAGTerminal)
	rstat = 0;
      else if (streq (curr_elem -> Affix_term.tname, curr_term -> Terminal.marker))
	{ /* We match the terminal */
	  top_madm -> Member_adm.mem_nr++;
	  rstat = parse_members (curr_mnr, terms, iptr + 1, type,
				 rstack, tstack, gstack, res_guards);
	  top_madm -> Member_adm.mem_nr--;
	}
      else rstat = 0;
      return (rstat);
    };

  /*
     Current element must be a nonterminal:
     The primitive ones should match a term or variable
  */
  if (curr_elem -> tag != TAGAffix_var) dcg_internal_error ("parse_members:2");
  vrule = get_prime_affix_rule (curr_elem -> Affix_var.vdef);
  if (curr_term -> adef == AFFIX_ERROR) return (0);
  cline = curr_term -> line;
  ccol = curr_term -> col;
  if (equivalent_affix_rule (vrule, curr_term -> adef))
    { /*
	 There are two possibilities: if the term is a single variable, we
	 should push it as a partially matched resolved guard. If it is an
	 expression and we are checking for a join, we should push it as
	 a resolved arithmetic join.

      */
      if (is_an_affix_variable (curr_term))
	{ app_affix_term_list (tstack, curr_term);
          top_madm -> Member_adm.mem_nr++;
          rstat += parse_members (curr_mnr, terms, iptr + 1, type,
				  rstack, tstack, gstack, res_guards);
          top_madm -> Member_adm.mem_nr--;
          del_affix_term_list (tstack, tstack -> size - 1);
	}
      else
	{ /* Construct a local resolved guard */
	  variable new_local;
	  guard_type arith_type = (type == join)?arith_join:arith_split;
	  int can_fail = (type == split);
	  new_tree = new_local_affix (cline, ccol, vrule);
	  new_local = new_tree -> Var.vdef;
	  guard_parts = new_affix_term_list ();
	  app_affix_term_list (guard_parts, attach_affix_term (curr_term));
	  local_guard = new_Res_guard (cline, ccol, arith_type, new_tree, 0, can_fail, guard_parts);
	  app_member_list (gstack, local_guard);
	  app_affix_term_list (tstack, new_tree);
	  if (dump_affix_rules)
            { dcg_wlog ("Constructed local %s, arule %s, alt %d",
		        string_from_guard_type (arith_type), arule -> aname, alt_nr);
	      pp_member (dcg_error_file (), local_guard);
	      dcg_wlog ("");
	    };
	  top_madm -> Member_adm.mem_nr++;
	  rstat += parse_members (curr_mnr, terms, iptr + 1, type,
				  rstack, tstack, gstack, res_guards);
	  top_madm -> Member_adm.mem_nr--;

	  /*
	     Now (partly) undo what you have done: we may have a memory leak here in the
	     case that the new_local is not needed because of a failing parse
	     Heuristic: use res_stack to determine when it is certain to detach
	  */
          del_affix_term_list (tstack, tstack -> size - 1);
	  delete_member_list (gstack, gstack -> size - 1);
	  detach_member (&local_guard);
          if (!rstat) detach_variable (&new_local);
	};
    };

  /*
     Check for an enclosed list of terms to match the current nonterminal
  */
  if (!equivalent_affix_rule (vrule, curr_term -> adef) &&
      (vrule -> tag != TAGAffix_prim) && (curr_term -> tag == TAGEnclosed))
    { top_madm -> Member_adm.mem_nr++;
      rstat += parse_arule (curr_mnr, curr_term -> Enclosed.terms, 0, vrule, type, 2,
			    rstack, tstack, gstack, res_guards);
      top_madm -> Member_adm.mem_nr--;
    };
  
  /*
     If we are not the first member to match (we do leftcorner parsing)
     and the type of the current term is unknown or not equivalent
     we try and recognize the member topdown by calling parse_rule
  */
  if (!equivalent_affix_rule (vrule, curr_term -> adef) &&
      (vrule -> tag != TAGAffix_prim) && (mem_nr != 0))
    { /*
	 We call parse_rule recursively: when we succeed we recognized this member
	 Note that we do not require this parse to match the whole sentential form
      */
      top_madm -> Member_adm.mem_nr++;
      rstat += parse_arule (curr_mnr, terms, iptr, vrule, type, 0,
			    rstack, tstack, gstack, res_guards);
      top_madm -> Member_adm.mem_nr--;
    };
  return (rstat);
}

static int reduce_arule (int curr_mnr, affix_term_list terms, int iptr, guard_type type,
			 madm_list rstack, affix_term_list tstack, member_list gstack,
			 member_list *res_guards)
{ /* Consistency check */
  madm top_madm = rstack -> array[rstack -> size - 1];
  affix_term top_tree = tstack -> array[tstack -> size - 1];
  affix_rule top_trule = top_tree -> adef;
  affix_rule goal = top_madm -> Rule_adm.goal;
  int_list lc_nrs = goal -> lc_nrs;
  int ix, lc_nr;
  int stat = 0;

  if (top_madm -> tag != TAGRule_adm) dcg_internal_error ("reduce_arule");
  if (top_tree -> tag != TAGVar) dcg_internal_error ("reduce_arule:2");
  if (dump_affix_rules)
    dcg_wlog ("rstack_size: %d, tstacksize: %d, gstacksize: %d",
	      rstack -> size, tstack -> size, gstack -> size);
  
  /*
     If our goal and top tree are the same (equivalent),
     we have a match. If it is also a complete match, we
     have a succesfull parse.
  */
  if (equivalent_affix_rule (goal, top_trule))
    { /* Either we have a complete parse or we have to continue */
      if ((rstack -> size == 1) && (iptr == terms -> size) && (top_madm -> Rule_adm.complete == 1))
	{ /* Our gstack has a complete parsetree */
	  if (*res_guards == member_list_nil)
	    { if (dump_affix_rules) dcg_wlog ("Full parse succeeded");
	      *res_guards = new_member_list ();
	      for (ix = 0; ix < gstack -> size; ix++)
		app_member_list (*res_guards, attach_member (gstack -> array[ix]));
	      return (1);
	    }
	  else return (2);	/* Ambiguous resolution */
	}
      else if ((rstack -> size > 2) && (iptr == terms -> size) &&
	       (top_madm -> Rule_adm.complete == 2))
	{ /* We are certain that we have recognized the current set of terms */
	  madm prev_rule_adm = rstack -> array[rstack -> size - 3];
	  del_madm_list (rstack, rstack -> size - 1);
	  stat += parse_members (curr_mnr, prev_rule_adm -> Rule_adm.terms,
				 prev_rule_adm -> Rule_adm.iptr, type,
				 rstack, tstack, gstack, res_guards);
	  app_madm_list (rstack, top_madm);
	}
      else if ((rstack -> size > 1) || !top_madm -> Rule_adm.complete)
	{ /* Incomplete parse, pop goal and continue */
	  del_madm_list (rstack, rstack -> size - 1);
          stat += parse_members (curr_mnr, terms, iptr, type, rstack, tstack, gstack, res_guards);
	  app_madm_list (rstack, top_madm);
	};
    };

  /*
     There is another possibility: we have only partly recognized the syntactic structure
     but the nonterminal that we just recognized is a leftcorner of our goal, but either
     not our goal or some leftcorner child/grandchild/....
     So we do not pop our goal, but push a member admin on the stack towards the goal
     (Kind of meet in the middle attack).
  */

  /* Iterate thru list of left corner nonterminals to get a left corner */
  for (lc_nr = 0; lc_nr < lc_nrs -> size; lc_nr++)
    { affix_rule lc_rule = all_affix_rules -> array[lc_nrs -> array[lc_nr]];
      affix_alternative_list alts = pick_rhs_affix_alts (lc_rule);
      int alt_nr;
      if (alts == affix_alternative_list_nil)
        { dcg_wlog ("Goal rule %s, LC %s", goal -> aname, lc_rule -> aname);
          dcg_internal_error ("parse_arule");
        };
      for (alt_nr = 0; alt_nr < alts -> size; alt_nr++)
	{ /* Select member list */
	  affix_element_list elems = alts -> array[alt_nr] -> elems;
	  affix_element first_member = elems -> array[0];
	  affix_rule mrule;

	  /* Check that we have a leftcorner towards the goal, if not deplore */
	  if (first_member -> tag != TAGAffix_var) continue;
	  mrule = first_member -> Affix_var.vdef;
	  if (!equivalent_affix_rule (top_trule, mrule)) continue;

	  /* We have determined that our current top tree is really on its way towards the goal */
	  stat += parse_alternative (curr_mnr, terms, iptr, type, lc_rule, alt_nr, 1, 
				     rstack, tstack, gstack, res_guards);
	};
    };
  return (stat);
}

int match_terms (int curr_mnr, affix_term_list terms, affix_rule arule, guard_type type,
		 member_list *res_guards)
{ affix_term_list tstack = new_affix_term_list ();
  member_list gstack = new_member_list ();
  madm_list rstack = new_madm_list ();
  *res_guards = member_list_nil;
  if (dump_affix_rules)
    { dcg_wlog ("Matching term list with affix rule %s, type %s in module %d",
	        arule -> aname, string_from_guard_type (type), curr_mnr);
      pp_affix_term_list (dcg_error_file (), terms);
      dcg_wlog ("");
    };
  return (parse_arule (curr_mnr, terms, 0, arule, type, 1, rstack, tstack, gstack, res_guards));
}
