/*
   File: optimize.c
   Optimizes the intermediate code

   Copyright (C) 2010 Marc Seutter

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

   CVS ID: "$Id: optimize.c,v 1.6 2011/08/13 15:57:30 marcs Exp $"
*/

/* global includes */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

/* libdcg includes */
#include <dcg.h>
#include <dcg_error.h>
#include <dcg_string.h>

/* local includes */
#include "cdl_ds.h"
#include "imc_utils.h"
#include "globals.h"
#include "options.h"
#include "optimize.h"

/*
   Some code to handle locals bitmaps
*/
static char_list make_locals_bitmap (int size)
{ char_list bitmap = init_char_list (size);
  bitmap -> size = size;
  return (bitmap);
}

static void clear_locals_bitmap (char_list bitmap)
{ int ix;
  for (ix = 0; ix < bitmap -> size; ix++)
    bitmap -> array[ix] = 0;
}

static char_list make_cleared_locals_bitmap (int size)
{ char_list locs = make_locals_bitmap (size);
  clear_locals_bitmap (locs);
  return (locs);
}

static char_list copy_locals_bitmap (char_list src)
{ char_list locs = make_locals_bitmap (src -> size);
  int ix;
  for (ix = 0; ix < locs -> size; ix++)
    locs -> array[ix] = src -> array[ix];
  return (locs);
}

static void copy_locals_bitmap_to (char_list src, char_list dst, int *change)
{ int ix;
  if (src -> size != dst -> size) dcg_internal_error ("copy_locals_bitmap_to");
  for (ix = 0; ix < src -> size; ix++)
    if (dst -> array[ix] != src -> array[ix])
      { dst -> array[ix] = src -> array[ix];
	*change = 1;
      };
}

static void add_locals_bitmap (char_list dst, char_list src)
{ int ix;
  if (src -> size != dst -> size) dcg_internal_error ("add_locals_bitmap");
  for (ix = 0; ix < src -> size; ix++)
    if (src -> array[ix]) dst -> array[ix] = 1;
}

static void substract_locals_bitmap (char_list dst, char_list src)
{ int ix;
  if (src -> size != dst -> size) dcg_internal_error ("substract_locals_bitmap");
  for (ix = 0; ix < src -> size; ix++)
    if (src -> array[ix]) dst -> array[ix] = 0;
}

/*
   The following code eliminates attach/detach pairs resulting from most splits
*/
static int attach_removed = 0;
static int detach_removed = 0;

static void try_optimize_attach_detach (instr loc, int lnr)
{ instr curr;
  int lnr2;
  for (curr = loc -> next; curr != instr_nil; curr = curr -> next)
    if (is_a_local_detach (curr, &lnr2) && (lnr == lnr2)) break;
    else if (is_a_local_use (curr, lnr)) return;

  /* Clear attach against detach */
  curr -> dont_code = 1;
  loc -> dont_code = 1;
  detach_removed++;
  attach_removed++;
}

static void optimize_detach_attach (instr loc, int lnr)
{ instr curr;
  int lnr2;
  loc -> dont_code = 1;	 /* Clear attach */
  attach_removed++;
  for (curr = loc; curr != instr_nil; curr = curr -> prev)
    if (is_a_local_detach (curr, &lnr2) && (lnr == lnr2))
      { curr -> dont_code = 1;
	detach_removed++;
	return;
      };
}

static void do_attach_detach_elimination_in_block (block bb, char_list bitmap)
{ address_list args;
  instr curr;
  int lnr;

  /* Eliminate the attach....detach patterns */
  clear_locals_bitmap (bitmap);
  for (curr = bb -> tail; curr != instr_nil; curr = curr -> prev)
    { if (is_a_local_detach (curr, &lnr))
	bitmap -> array[lnr] = 1;
      if (is_a_local_attach (curr, &lnr) && bitmap -> array[lnr])
	{ try_optimize_attach_detach (curr, lnr);
	  bitmap -> array[lnr] = 0;
	};
    };

  /* Eliminate the detach....attach patterns, but be very carefull */
  clear_locals_bitmap (bitmap);
  for (curr = bb -> head; curr != instr_nil; curr = curr -> next)
    { if (is_a_local_detach (curr, &lnr))
	bitmap -> array[lnr] = 1;
      else if (is_a_call (curr, &args)) break;
      else if (is_a_local_update (curr, &lnr))
	bitmap -> array[lnr] = 0;
      else if (is_a_local_attach (curr, &lnr) && bitmap -> array[lnr])
	{ optimize_detach_attach (curr, lnr);
	  bitmap -> array[lnr] = 0;
	}
    };
}

static void do_attach_detach_elimination_in_proc (proc p)
{ int size = p -> pvars -> size;
  block_list bbs = p -> bbs;
  char_list bitmap = make_locals_bitmap (size);
  int ix;
  for (ix = 0; ix < bbs -> size; ix++)
    { block bb = bbs -> array[ix];
      if (bb -> dont_code) continue;
      do_attach_detach_elimination_in_block (bb, bitmap);
    };
  detach_char_list (&bitmap);
}

static void do_attach_detach_elimination ()
{ int ix;
  attach_removed = 0;
  detach_removed = 0;
  dcg_hint ("      eliminating attach/detach pairs...");
  for (ix = 0; ix < all_procedures -> size; ix++)
    { proc p = all_procedures -> array[ix];
      if (p -> tag != TAGAlts) continue;
      if (!p -> called) continue;
      do_attach_detach_elimination_in_proc (p);
    };
  if (attach_removed != detach_removed)
    dcg_internal_error ("do_attach_detach_elimination: unbalanced");
  dcg_hint ("         removed %d attach/detach pairs", attach_removed);
}

/*
   The following code is responsible for live/dead analysis of local variables
*/
static void calc_gen_use_for_bb (instr tail, char_list gen, char_list use)
{ instr curr;
  for (curr = tail; curr != instr_nil; curr = curr -> prev)
    calc_local_gen_use (curr, gen, use);
}

/*
   Initialise the bitsets and calculate Gen, Use for all basic blocks
   For the callf instruction which ends a basic block, also
   calculate the gen and use of the call itself
*/
static void calculate_initial_dataflow_sets (proc p)
{ int size = p -> pvars -> size;
  block_list bbs = p -> bbs;
  int ix;
  
  for (ix = 0; ix < bbs -> size; ix++)
    { block bb = bbs -> array[ix];
      if (bb -> dont_code) continue;
      bb -> in  = make_cleared_locals_bitmap (size);
      bb -> use = make_cleared_locals_bitmap (size);
      bb -> gen = make_cleared_locals_bitmap (size);
      bb -> out = make_cleared_locals_bitmap (size);
      if (bb -> tail -> tag == TAGCallf)
	{ bb -> callf_gen = make_cleared_locals_bitmap (size);
	  bb -> callf_use = make_cleared_locals_bitmap (size);
	  calc_local_gen_use (bb -> tail, bb -> callf_gen, bb -> callf_use);
	  calc_gen_use_for_bb (bb -> tail -> prev, bb -> gen, bb -> use);
	}
      else calc_gen_use_for_bb (bb -> tail, bb -> gen, bb -> use);
    };
}

/*
   Now calculate the out and in sets for every basic block
   and note if one of them changes. We do this in reverse
   order over the basic blocks of the procedure because
   this will usually lead to a faster convergence
*/
static void calculate_out_and_in_sets (proc p, int *change)
{ char_list tmp = make_locals_bitmap (p -> pvars -> size);
  block_list bbs = p -> bbs;
  int ix, iy;
  for (ix = bbs -> size - 1; 0 <= ix; ix--)
    { block bb = bbs -> array[ix];
      if (bb -> dont_code) continue;

      /* First determine the out set as the union of all successor in sets */
      clear_locals_bitmap (tmp);
      if (bb -> true_succ != block_list_nil)
        for (iy = 0; iy < bb -> true_succ -> size; iy++)
	  add_locals_bitmap (tmp, bb -> true_succ -> array[iy] -> in);

      /* All true successors' in sets added, do local calculation for callf instruction */
      if (bb -> tail -> tag == TAGCallf)
	{ substract_locals_bitmap (tmp, bb -> callf_gen);
	  add_locals_bitmap (tmp, bb -> callf_use);
	};
      if (bb -> false_succ != block_nil)
        add_locals_bitmap (tmp, bb -> false_succ -> in);
      copy_locals_bitmap_to (tmp, bb -> out, change);

      /* Calculate in set from out set */
      substract_locals_bitmap (tmp, bb -> gen);
      add_locals_bitmap (tmp, bb -> use);
      copy_locals_bitmap_to (tmp, bb -> in, change);
    };

  detach_char_list (&tmp);
}

static void optimize_dead_assignments (proc p)
{ block_list bbs = p -> bbs;
  int ix;

  for (ix = bbs -> size - 1; 0 <= ix; ix--)
    { /* Pick up block */
      block bb = bbs -> array[ix];
      char_list alive;
      instr tail;
      if (bb -> dont_code) continue;

      /* Prepare to iterate backwards */
      tail = bb -> tail;
      alive = copy_locals_bitmap (bb -> out);
      if (tail -> tag == TAGCallf) tail = tail -> prev;

      while (tail != instr_nil)
	{ int lnr;
	  if (is_a_local_update (tail, &lnr) && !alive -> array[lnr])
	    tail -> dont_code = 1;
	  calc_local_gen_use (tail, char_list_nil, alive);
	  tail = tail -> prev;
	};
      detach_char_list (&alive);
    };
}

static void do_local_live_dead_analysis_in_proc (proc p)
{ int change;
  calculate_initial_dataflow_sets (p);
  do 
    { change = 0;
      calculate_out_and_in_sets (p, &change);
    }
  while (change);    
  optimize_dead_assignments (p);
}

static void do_local_live_dead_analysis ()
{ int ix;
  dcg_hint ("      local live/dead analysis...");
  for (ix = 0; ix < all_procedures -> size; ix++)
    { proc p = all_procedures -> array[ix];
      if (p -> tag != TAGAlts) continue;
      if (!p -> called) continue;
      do_local_live_dead_analysis_in_proc (p);
    };
}

/*
   The following code is responsible for right recursion removal
*/
static void try_remove_tail_recursion_in_block (proc pdef, block bb)
{ instr tail = bb -> tail;
  instr prev_tail = tail;
  proc called_pdef;

  if (tail -> tag != TAGSucceed) return;
  while (1)
    { if (prev_tail == instr_nil) return;
      if (!prev_tail -> dont_code) break;
      prev_tail = prev_tail -> prev;
    }
  if (prev_tail -> tag != TAGCallp) return;

  called_pdef = all_procedures -> array[prev_tail -> Callp.pnr];
  if (pdef == called_pdef)
    { dcg_eprint ("Tail recursion candidate: ");
      dump_proc_name (pdef);
      dcg_eprint ("\n");
    };
}

static void do_tail_recursion_removal_in_proc (proc p)
{ block_list bbs = p -> bbs;
  int ix;
  for (ix = 0; ix < bbs -> size; ix++)
    { block bb = bbs -> array[ix]; ix++;
      if (bb -> dont_code) continue;
      try_remove_tail_recursion_in_block (p, bb);
    }
}

static void do_tail_recursion_removal ()
{ int ix;
  dcg_hint ("      removing tail recursion...");
  for (ix = 0; ix < all_procedures -> size; ix++)
    { proc p = all_procedures -> array[ix];
      if (p -> tag != TAGAlts) continue;
      if (!p -> called) continue;
      do_tail_recursion_removal_in_proc (p);
    };
}

/*
   The following code is responsible for procedure inlining
*/
static void do_procedure_inlining ()
{ int ix;
  dcg_hint ("      inlining procedures...");
  for (ix = 0; ix < all_procedures -> size; ix++)
    { proc p = all_procedures -> array[ix];
      if (p -> tag != TAGAlts) continue;
      if (!p -> called) continue;
      if (p -> called == 1)
	{ dcg_eprint ("        inlining candidate is: ");
	  dump_proc_name (p);
	  dcg_eprint ("\n");
        };
    };
}

/*
   The following code is part of the strength reduction
   It will replace intermediate code patterns of the form

              if Equal_altnr, src, value_1, fail to LAB%2
              Basic block_1, succeeding
              ...
   LAB%2:     if Equal_altnr, src, value_2, fail to LAB%3
	      Basic block_2, succeeding
              ...
              ...
   LAB%n:     if Equal_altnr, src, value_n, fail to LABdefault
	      Basic block_n, succeeding

   LABdefault:

   by:
	      Switch_altnr, src, [(value1, new_LAB%1), ... ,(value_n, new_LAB%n)] fail to LABdefault

   new_LAB%1: Basic block_1, succeeding
   new_LAB%2: Basic block_2, succeeding
              ...
   new_LAB%n: Basic block_n, succeeding
   LABdefault:
*/ 
static int try_pick_case_label (instr tail, address *src, int *altnr, int *fail, int *kind)
{ switch (tail -> tag)
    { case TAGEqual_altnr:
	*src   = tail -> Equal_altnr.src;
	*altnr = tail -> Equal_altnr.altnr;
	*fail  = tail -> Equal_altnr.fail;
	*kind  = 0;
	break;
      case TAGEqual_int:
	if (tail -> Equal_int.cmp -> tag != TAGInum) return (0);
	*src   = tail -> Equal_int.src;
	*altnr = tail -> Equal_int.cmp -> Inum.num;
	*fail  = tail -> Equal_int.fail;
	*kind  = 1;
	break;
      default: return (0);
    };
  return (1);
}

static instr make_new_switch (instr tail, altnr_label_pair_list cases, int fail)
{ if (tail -> tag == TAGEqual_altnr)
    return (new_Switch_altnr (attach_address (tail -> Equal_altnr.src), cases, fail));
  else return (new_Switch_int (attach_address (tail -> Equal_int.src), cases, fail));
}

static int try_collect_case_list (block bb, address src, int kind, altnr_label_pair_list *cases)
{ altnr_label_pair_list ncases = new_altnr_label_pair_list ();
  int altnr, fail, kind2;
  block leader = bb;
  address src2;

  /* Collect the list of cases */
  while (try_pick_case_label (leader -> tail, &src2, &altnr, &fail, &kind2))
    { altnr_label_pair pair;
      block true_succ;
      if (!equal_address (src, src2)) break;
      if (kind != kind2) break;				/* just for safety */

      /* Check that apart from the first, all leaders are just a label and an instruction */
      if (ncases -> size)
        { if (leader -> head -> tag != TAGLabel) break;
          if (leader -> head -> next != leader -> tail) break;
	  if (leader -> pred -> size != 1) break;
	};

      /* Check true successor */
      if (leader -> true_succ -> size != 1) break;	/* messed up? */
      true_succ = leader -> true_succ -> array[0];
      if (true_succ -> pred -> size != 1) break;	/* Some other way to end up here */
      if (true_succ -> pred -> array[0] != leader)
        dcg_internal_error ("try_collect_case_list");

      /* Create pair */
      pair = new_altnr_label_pair (altnr, -1);		/* Label lookup later */
      app_altnr_label_pair_list (ncases, pair);
      leader = leader -> false_succ;
    };

  /* If there are not enough cases, be done with it */
  if (ncases -> size <= 1)
    { detach_altnr_label_pair_list (&ncases);
      return (0);
    };

  /* Ok, we have something to report */
  *cases = ncases;
  return (1);
}

static void process_case_list (block bb, int kind, altnr_label_pair_list cases)
{ block last_leader = block_nil;
  int last_fail_label = -1;
  block leader = bb;
  instr new_sw;
  int ix;
  for (ix = 0; ix < cases -> size; ix++)
    { block true_succ = leader -> true_succ -> array[0];
      int lnr = pick_or_create_new_block_label (true_succ);
      cases -> array[ix] -> lnr = lnr;
      if (ix)
        { true_succ -> pred -> array[0] = bb;
	  app_block_list (bb -> true_succ, true_succ);
          leader -> dont_code = 1;
	};

      last_fail_label = pick_fail_label (leader -> tail);
      last_leader = leader;
      leader = leader -> false_succ;
    };

  new_sw = make_new_switch (bb -> tail, cases, last_fail_label);
  replace_last_instruction (bb, new_sw);
  remove_block_from_list (leader -> pred, last_leader);
  app_block_list (leader -> pred, bb);
  bb -> false_succ = leader;
}

static void try_introduce_switches (block bb)
{ altnr_label_pair_list cases;
  int altnr, fail, kind;
  address src;
  if (!try_pick_case_label (bb -> tail, &src, &altnr, &fail, &kind)) return;
  if (try_collect_case_list (bb, src, kind, &cases))
    process_case_list (bb, kind, cases);
}

static void do_strength_reduction_in_proc (proc p, block_list bbs)
{ int ix = 0;
  while (ix < bbs -> size)
    { if (!bbs -> array[ix] -> dont_code)
        try_introduce_switches (bbs -> array[ix]);
      ix++;
    };
}

static void do_strength_reduction ()
{ int ix;
  dcg_hint ("      strength reduction...");
  for (ix = 0; ix < all_procedures -> size; ix++)
    { proc p = all_procedures -> array[ix];
      if (p -> tag != TAGAlts) continue;
      if (!p -> called) continue;
      do_strength_reduction_in_proc (p, p -> bbs);
    };
}

/* exported code */
void intermediate_code_optimization ()
{ if (!optimize_code) return;
  dcg_warning (0, "   optimizing intermediate code...");
  do_attach_detach_elimination ();
  do_local_live_dead_analysis ();
  /*
  do_tail_recursion_removal ();
  do_procedure_inlining ();
  */
  do_strength_reduction ();
  if (dump_optimizer) dump_intermediate_code ();
}

