/*
   File: variables.c
   Checks the variables

   Copyright (C) 2008 Marc Seutter

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

   CVS ID: "$Id: variables.c,v 1.22 2011/09/02 10:11:23 marcs Exp $"
*/

/* standard includes */
#include <stdio.h>
#include <string.h>

/* support lib includes */
#include <dcg.h>
#include <dcg_alloc.h>
#include <dcg_error.h>
#include <dcg_string.h>

/* local includes */
#include "cdl_ds.h"
#include "ast_utils.h"
#include "globals.h"
#include "options.h"
#include "affix_rules.h"
#include "contsens.h"
#include "procedures.h"
#include "variables.h"

/*
   We sometimes need to create anonymous local affixes
   Keep track of a unique affix rule number
*/
static int unique_affix_nr;
variable new_local_var (affix_rule arule)
{ string vname = dcg_new_fmtd_string ("_%s_%d", arule -> aname, unique_affix_nr);
  variable new_var = new_Local_var (vname);
  unique_affix_nr++;
  new_var -> adef = arule;
  return (new_var);
}

affix_term new_local_affix (int line, int col, affix_rule arule)
{ string vname = dcg_new_fmtd_string ("_%s_%d", arule -> aname, unique_affix_nr);
  affix_term new_tree;
  variable new_var;
  unique_affix_nr++;
  new_var = new_Local_var (vname);
  new_var -> adef = arule;
  new_tree = new_Var (line, col, attach_string (vname));
  new_tree -> adef = arule;
  new_tree -> Var.vdef = new_var;
  return (new_tree);
} 

static void init_variables ()
{ unique_affix_nr = 0;
}

/*
   Checking of all (global) module variables
*/
static void check_module_variables (module mod)
{ variable_list mvars = mod -> mvars;
  int ix;
  for (ix = 0; ix < mvars -> size; ix++)
    { variable mvar = mvars -> array[ix];
      string vname = mvar -> vname;
      affix_rule arule = identify_affix_instance (vname, mod);
      int iy;
      if (arule == affix_rule_nil)
	contsens_error (mod, 0, 0, "can't identify module variable %s", vname);
      mvar -> adef = arule;
      for (iy = ix + 1; iy < mvars -> size; iy++)
	if (streq (mvars -> array[iy] -> vname, vname))
	  contsens_error (mod, 0, 0, "multiple declaration of module variable %s", vname);
      mvar -> nr = all_global_variables -> size;
      app_variable_list (all_global_variables, mvar);
    };
}

void check_global_variables ()
{ int ix;
  dcg_warning (0, "   checking global variables...");
  init_variables ();
  for (ix = 0; ix < all_modules -> size; ix++)
    check_module_variables (all_modules -> array[ix]);
  dcg_panic_if_errors ();
}

static int is_a_module_variable (string vname, module mod, variable *vdef)
{ variable_list mvars = mod -> mvars;
  int ix;
  for (ix = 0; ix < mvars -> size; ix++)
    { variable mvar = mvars -> array[ix];
      if (streq (mvar -> vname, vname))
	{ *vdef = mvar;
	  return (1);
	};
    };
  return (0);
}

/*
   Checking of all arguments in export specifications:
   Arguments in specs may be equal to each other.
*/
static void check_formals (fpar_list fpars, module mod, signature *sig)
{ int ix;
  *sig = init_psig_list (fpars -> size);
  for (ix = 0; ix < fpars -> size; ix++)
    { fpar formal = fpars -> array[ix];
      affix_rule arule = identify_affix_instance (formal -> fname, mod);
      psig ps = new_psig ();
      if (arule == affix_rule_nil)
	contsens_error (mod, formal -> line, formal -> col,
			"can't identify formal parameter %s", formal -> fname);
      ps -> ftyp = arule;
      ps -> fdir = formal -> fdir;
      app_psig_list (*sig, ps);
    };
}

void check_spec_formals (spec pspec, module mod, signature *p1sig, signature *p2sig)
{ check_formals (pspec -> p1fpars, mod, p1sig);
  check_formals (pspec -> p2fpars, mod, p2sig);
}

/*
   Checking of formal procedure arguments
   Note that they may not be equal to each other nor to a module variable
   As a side effect the procedure's variable list gets updated with instances
*/
static void identify_formal (fpar formal, module mod, variable_list pvars, signature sig)
{ /* Lookup affix rule and create local variable */
  affix_rule arule = identify_affix_instance (formal -> fname, mod);
  variable local = new_Local_var (attach_string (formal -> fname));
  psig ps = new_psig ();
  variable vdef;

  /* Identify type definition */
  if (arule == affix_rule_nil)
    contsens_error (mod, formal -> line, formal -> col,
		    "can't identify formal parameter %s", formal -> fname);
  local -> adef = arule;
  local -> nr = pvars -> size;
  app_variable_list (pvars, local);

  /* Check that we are not equal to some module variable */
  if (is_a_module_variable (formal -> fname, mod, &vdef))
    contsens_error (mod, formal -> line, formal -> col,
		    "formal parameter %s equals module variable", formal -> fname);
  else formal -> fdef = local;
  
  /* Add formal type and direction to signature */
  ps -> ftyp = arule;
  ps -> fdir = formal -> fdir;
  app_psig_list (sig, ps);
}

static void identify_formals (fpar_list fpars, module mod, variable_list pvars, signature *sig)
{ int ix;
  *sig = init_psig_list (fpars -> size);
  for (ix = 0; ix < fpars -> size; ix++)
    identify_formal (fpars -> array[ix], mod, pvars, *sig);
}

static void check_double_formal (string fname, fpar formal, module mod)
{ if (!streq (fname, formal -> fname)) return;
  contsens_error (mod, formal -> line, formal -> col, "multiple formal parameter %s", fname);
}

static void check_double_formals (fpar_list p1fpars, fpar_list p2fpars, module mod)
{ int ix, iy;
  for (ix = 0; ix < p1fpars -> size; ix++)
    { string fname = p1fpars -> array[ix] -> fname;
      for (iy = ix + 1; iy < p1fpars -> size; iy++)
	check_double_formal (fname, p1fpars -> array[iy], mod);
      for (iy = 0; iy < p2fpars -> size; iy++)
	check_double_formal (fname, p2fpars -> array[iy], mod);
    };
  for (ix = 0; ix < p2fpars -> size; ix++)
    { string fname = p2fpars -> array[ix] -> fname;
      for (iy = ix + 1; iy < p2fpars -> size; iy++)
	check_double_formal (fname, p2fpars -> array[iy], mod);
    };
}

void check_proc_formals (spec pspec, module mod, variable_list *pvars,
			 signature *p1sig, signature *p2sig)
{ fpar_list p1fpars = pspec -> p1fpars;
  fpar_list p2fpars = pspec -> p2fpars;
  *pvars = init_variable_list ((p1fpars -> size + p2fpars -> size) * 2);	/* Heuristic */
  identify_formals (p1fpars, mod, *pvars, p1sig);
  identify_formals (p2fpars, mod, *pvars, p2sig);
  check_double_formals (p1fpars, p2fpars, mod);
}

/*
   Checking and identification of affix variables
*/
int is_an_affix_variable (affix_term term)
{ switch (term -> tag)
    { case TAGVar:
      case TAGSub:
      case TAGConcr: return (1);
      default: return (0);
    };
}

int is_a_single_affix_variable (affix_term_list terms)
{ if (terms -> size != 1) return (0);
  return (is_an_affix_variable (terms -> array[0]));
}

int is_a_simple_affix_variable (affix_term_list terms)
{ affix_term term;
  if (terms -> size != 1) return (0);
    { term = terms -> array[0];
      while (term -> tag == TAGConcr)
	term = term -> Concr.abs;
      return (term -> tag == TAGVar);
    };
}

static int is_a_local_variable (string vname, variable_list pvars, variable *vdef)
{ int ix;
  for (ix = 0; ix < pvars -> size; ix++)
    { variable pvar = pvars -> array[ix];
      if (streq (vname, pvar -> vname))
	{ *vdef = pvar;
	  return (1);
	};
    };
  return (0);
}

/*
   Simple checking of expressions, affix variables and affix instances

   The result of a check is that every affix formula becomes typed
   with the exception of enclosed expressions for which resolution
   might be needed. In effect it is a kind of precomputation that
   is only done once, leaving the type pointer in the formula node.
   When we return null, we have encountered an error so we may not be resolved
*/
static int make_typed_operator (operator *op, prim_type pt1, prim_type pt2)
{ switch (*op)
    { case plus:
	switch (pt1)
	  { case prim_int:
	      if ((pt2 == prim_int) || (pt2 == prim_none)) { *op = int_plus_int; return (1); }
	      return (0);
	    case prim_real:
	      if ((pt2 == prim_real) || (pt2 == prim_none)) { *op = real_plus_real; return (1); }
	      return (0);
	    case prim_text:
	      if (pt2 == prim_text) { *op = text_plus_text; return (1); }
	    default: return (0);
	  }; 
      case minus:
	switch (pt1)
	  { case prim_int:
	      if ((pt2 == prim_int) || (pt2 == prim_none)) { *op = int_minus_int; return (1); }
	      return (0);
	    case prim_real:
	      if ((pt2 == prim_real) || (pt2 == prim_none)) { *op = real_minus_real; return (1); }
	    default: return (0);
	  };
      case times:
	switch (pt1)
	  { case prim_int:
	      if (pt2 == prim_int) { *op = int_times_int; return (1); }
	      else if (pt2 == prim_text) { *op = int_times_text; return (1); }
	      return (0);
	    case prim_real:
	      if (pt2 == prim_real) { *op = real_times_real; return (1); }
	    default: return (0);
	  };
      case bitwise_not:
	if ((pt1 == prim_int) && (pt2 == prim_none)) return (1);
	return (0);
      default: /* Only allow int op int */
	if ((pt1 == prim_int) && (pt2 == prim_int)) return (1);
	return (0);
    }
}

static affix_rule check_term (affix_term term, variable_list pvars, module mod)
{ prim_type pt1, pt2;
  switch (term -> tag)
    { case TAGEnclosed: 
	term -> adef = check_expr (term -> Enclosed.terms, pvars, mod);
	break;
      case TAGDyop:
	{ affix_rule arg1_rule = check_term (term -> Dyop.arg1, pvars, mod);
	  affix_rule arg2_rule = check_term (term -> Dyop.arg2, pvars, mod);

	  /* Check argument 1 */
	  if (is_integer_type (arg1_rule)) pt1 = prim_int;
	  else if (is_real_type (arg1_rule)) pt1 = prim_real;
	  else if (is_text_type (arg1_rule)) pt1 = prim_text;
	  else if (arg1_rule == AFFIX_ERROR) pt1 = prim_int;
	  else pt1 = prim_none;
	  if (pt1 == prim_none)
	    { contsens_error (mod, term -> line, term -> col,
			      "Dyadic operators can only be applied to numeric or TEXT operands");
	      arg1_rule = AFFIX_ERROR;
	    };

	  /* Check argument 2 */
	  if (is_integer_type (arg2_rule)) pt2 = prim_int;
	  else if (is_real_type (arg2_rule)) pt2 = prim_real;
	  else if (is_text_type (arg2_rule)) pt2 = prim_text;
	  else if (arg2_rule == AFFIX_ERROR) pt2 = prim_int;
	  else pt2 = prim_none;
	  if (pt2 == prim_none)
	    { contsens_error (mod, term -> line, term -> col,
			      "Dyadic operators can only be applied to numeric or TEXT operands");
	      arg2_rule = AFFIX_ERROR;
	    };

	  /* Check for recovery */
	  if ((arg1_rule == AFFIX_ERROR) || (arg2_rule == AFFIX_ERROR))
	    { term -> adef = AFFIX_ERROR; 
	      return (AFFIX_ERROR);
	    };

	  /* Check for compatibility of operator and both operands */
	  if (!make_typed_operator (&term -> Dyop.dop, pt1, pt2))
	    contsens_error (mod, term -> line, term -> col,
			    "can't apply dyadic operator %s to these operands",
			    string_from_operator (term -> Dyop.dop));

	  /* Return the correct (identified) type or AFFIX_ERROR */
	  term -> adef = type_from_operator (term -> Dyop.dop);
	}; break;
      case TAGMonop:
	{ affix_rule arg_rule = check_term (term -> Monop.arg, pvars, mod);
	  if (is_integer_type (arg_rule)) pt1 = prim_int;
          else if (is_real_type (arg_rule)) pt1 = prim_real;
	  else if (arg_rule == AFFIX_ERROR) return (AFFIX_ERROR);
	  else
	    { contsens_error (mod, term -> line, term -> col,
			      "Monadic operators can only be applied to numeric operands");
	      term -> adef = AFFIX_ERROR; 
	      return (AFFIX_ERROR);
	    };

	  /* Check for compatibility of operator and operand */
	  if (!make_typed_operator (&term -> Monop.mop, pt1, prim_none))
	    contsens_error (mod, term -> line, term -> col,
			    "can't apply monadic operator %s to these operands",
			    string_from_operator (term -> Monop.mop));

	  /* Return the correct (identified) type or AFFIX_ERROR */
	  term -> adef = type_from_operator (term -> Monop.mop);
	}; break;
      case TAGTerminal:
	term -> adef = affix_rule_nil;
	break;
      case TAGInt: term -> adef = get_primitive_type (prim_int); break;
      case TAGReal: term -> adef = get_primitive_type (prim_real); break;
      case TAGText: term -> adef = get_primitive_type (prim_text); break;
      default: return (check_affix_variable (term, pvars, mod));
    };

  /* Return nil if you do not know */
  return (term -> adef);
}

affix_rule check_expr (affix_term_list terms, variable_list pvars, module mod)
{ int ix;
  if (terms -> size == 1)
    return (check_term (terms -> array[0], pvars, mod));
  for (ix = 0; ix < terms -> size; ix++)
    (void) check_term (terms -> array[ix], pvars, mod);

  /* For a list of terms we do not know the type yet */
  return (affix_rule_nil);
}

/*
   Checking of variables
*/
static void should_be_integer_affix_formula (affix_term term, variable_list pvars, module mod)
{ affix_rule arule = check_term (term, pvars, mod);
  if (arule == AFFIX_ERROR) return;	/* Already given a warning */
  if (!is_integer_type (arule))
    contsens_error (mod, term -> line, term -> col, "Integer affix formula expected");
}

static affix_rule check_indexed_variable (affix_term term, variable_list pvars, module mod)
{ string aname = term -> Sub.aname;
  variable vdef;
  should_be_integer_affix_formula (term -> Sub.index, pvars, mod);
  if (is_a_module_variable (aname, mod, &vdef))
    { if (vdef -> tag != TAGModule_array)
	contsens_error (mod, term -> line, term -> col,
			"Variable %s can not be subscripted", aname);
      else term -> Sub.vdef = vdef;
    }
  else if (is_a_local_variable (aname, pvars, &vdef))
    contsens_error (mod, term -> line, term -> col,
		    "Local affix %s can not be subscripted", aname);
  else
    { contsens_error (mod, term -> line, term -> col,
		      "Can't identify array %s", aname);
      term -> adef = AFFIX_ERROR;
      return (AFFIX_ERROR);
    };
  term -> adef = vdef -> adef;
  return (vdef -> adef);
}

static affix_rule check_concretized_variable (affix_term term, variable_list pvars, module mod)
{ affix_rule arule = check_affix_variable (term -> Concr.abs, pvars, mod);
  affix_rule orig_arule;
  affix_rule crule;
  if (arule == affix_rule_nil) return (arule);
  if (!is_abstract_synonym (arule, &orig_arule, &crule))
    { contsens_error (mod, term -> line, term -> col, "Affix term can not be concretized");
      term -> adef = AFFIX_ERROR;
      return (AFFIX_ERROR);
    };
  if (strict_export && (orig_arule -> mnr != mod -> mnr))
    contsens_warning (mod, term -> line, term -> col,
		      "Concretization is only allowed in the defining module");
  term -> adef = crule;
  return (crule);
}

static affix_rule check_single_variable (affix_term term, variable_list pvars, module mod)
{ string aname = term -> Var.aname;
  variable vdef;
  if (is_a_module_variable (aname, mod, &vdef))
    { if (vdef -> tag != TAGModule_var)
	{ contsens_error (mod, term -> line, term -> col,
			  "Array %s must be subscripted in this context", aname);
	  return (vdef -> adef);
	};
    }
  else if (is_a_local_variable (aname, pvars, &vdef)) /* done */;
  else
    { affix_rule adef = identify_affix_instance (aname, mod);
      vdef = new_Local_var (aname);
      vdef -> nr = pvars -> size;
      app_variable_list (pvars, vdef);
      if (adef == affix_rule_nil)
        { contsens_error (mod, term -> line, term -> col,
			  "could not identify local affix %s", aname);
	  adef = AFFIX_ERROR;
        };
      vdef -> adef = adef;
    };
  term -> Var.vdef = vdef;
  term -> adef = vdef -> adef;
  return (vdef -> adef);
}

affix_rule check_affix_variable (affix_term term, variable_list pvars, module mod)
{ switch (term -> tag)
    { case TAGVar:   return (check_single_variable (term, pvars, mod));
      case TAGSub:   return (check_indexed_variable (term, pvars, mod));
      case TAGConcr: return (check_concretized_variable (term, pvars, mod));
      default: dcg_bad_tag (term -> tag, "check_affix_variable");
    };
  return (affix_rule_nil);
}

/*
   Local variables must be detached from time to time
*/
variable var_from_affix_term (affix_term term)
{ switch (term -> tag)
    { case TAGConcr: return (var_from_affix_term (term -> Concr.abs));
      case TAGSub: return (term -> Sub.vdef);
      case TAGVar: return (term -> Var.vdef);
      default: dcg_bad_tag (term -> tag, "var_from_affix_term");
    };
  return (variable_nil);
}

int is_a_global_variable (affix_term term)
{ variable var = var_from_affix_term (term);
  return (var -> tag != TAGLocal_var);
}

/*
   Affix term simplification
*/
static void collect_terms (affix_term_list concs, affix_term term)
{ affix_term_list srcs;
  int ix;
  if (term -> tag != TAGConcat)
    { app_affix_term_list (concs, term);
      return;
    };

  /* Copy the constituents */
  srcs = term -> Concat.terms;
  for (ix = 0; ix < srcs -> size; ix++)
    app_affix_term_list (concs, srcs -> array[ix]);

  nonrec_detach_affix_term_list (&srcs);
  dcg_detach ((void **) &term);
}

static void simplify_affix_term (affix_term *term, proc p)
{ affix_term dterm = *term;
  switch (dterm -> tag)
    { case TAGDyop:
	{ simplify_affix_term (&dterm -> Dyop.arg1, p);
	  simplify_affix_term (&dterm -> Dyop.arg2, p);
	  if (dterm -> Dyop.dop == text_plus_text)
	    { affix_term_list concs = new_affix_term_list ();
	      collect_terms (concs, dterm -> Dyop.arg1);
	      collect_terms (concs, dterm -> Dyop.arg2);
	      *term = new_Concat (dterm -> line, dterm -> col, concs);
	      if (p -> concat_size < concs -> size)
		p -> concat_size = concs -> size;
	      dcg_detach ((void **) &dterm);
	    };
	}; break;
      case TAGMonop:
	simplify_affix_term (&dterm -> Monop.arg, p);
	break;
      case TAGEnclosed:	
	{ affix_term_list eterms = dterm -> Enclosed.terms;
	  if (eterms -> size != 1)
	    dcg_internal_error ("simplify_affix_term");
	  simplify_affix_term (&eterms -> array[0], p);
	  *term = eterms -> array[0];
	  nonrec_detach_affix_term_list (&eterms);
	  dcg_detach ((void **) &dterm);
	}; break;
      case TAGConcr: 
	{ simplify_affix_term (&dterm -> Concr.abs, p);
	  *term = dterm -> Concr.abs;
	  dcg_detach ((void **) &dterm);
	};
      case TAGSub:
      case TAGVar:
      case TAGInt:
      case TAGReal:
      case TAGText:	break;
      default: dcg_bad_tag (dterm -> tag, "simplify_affix_term");
    };
}

void simplify_affix_terms (affix_term_list terms, proc p)
{ int ix;
  for (ix = 0; ix < terms -> size; ix++)
    simplify_affix_term (&terms -> array[ix], p);
}
