/*
   File: mrts_mcdl.c
   Defines the basic MCDL runtime system

   Copyright (C) 2008 Marc Seutter

   This library is free software: you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published   
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU Lesser General Public License
   along with this library.  If not, see <http://www.gnu.org/licenses/>.

   CVS ID: "$Id: mrts_mcdl.c,v 1.10 2008/06/28 13:52:07 marcs Exp $"
*/

/* Standard includes */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

/* Local includes */
#include "mrts_mcdl.h"

/*
   Error handler functions
*/
#define MAXERRLEN 1024
void mrts_panic (char *format, ...)
{ char buf[MAXERRLEN + 1];
  va_list arg_ptr;
  va_start (arg_ptr, format);
  vsprintf (buf, format, arg_ptr);
  va_end (arg_ptr);
  fprintf (stderr, "\nMCDL runtime system panic: %s\n", buf);
  exit (4);
};

void mrts_error (char *format, ...)
{ char buf[MAXERRLEN + 1];
  va_list arg_ptr;
  va_start (arg_ptr, format);
  vsprintf (buf, format, arg_ptr);
  va_end (arg_ptr);
  fprintf (stderr, "\nMCDL runtime system error: %s\n", buf);
  exit (4);
};

void mrts_abort (char *module_name, int linenr)
{ fprintf (stderr, "\nAbort in line %d of module %s", linenr, module_name);
  mrts_exit ();
}

void mrts_exit ()
{ fprintf (stderr, "\n");
  exit (0);
}

/*
   Attaching and detaching when MEMCHECK enabled
*/
void mrts_attach_value (value v)
{ if (v.vp == NULL)
    mrts_panic ("Attach to NULL pointer");
  if (v.lv & 7L)
    mrts_panic ("Non aligned attach");
  if (Refcount (v) == -1) return;	/* Constant */
  if (Refcount (v) > 100000)
    mrts_panic ("Absurd refcount");
  Refcount(v)++;
}

int mrts_try_decr_refcount (value v)
{ if (v.vp == NULL) return (0);
  if (v.lv & 7L)
    mrts_panic ("Non aligned detach");
  if (Refcount (v) < 0) return (0);
  if (Refcount (v) == 0)
    mrts_panic ("Detaching a free object");
  Refcount(v)--;
  return (Refcount (v) == 0);
}

/*
   Memory allocation
*/
static void *ck_calloc (size_t nr, size_t nr_bytes)
{ void *vp = calloc (nr, nr_bytes);
  if (vp == NULL)
    mrts_panic ("Not enough memory to allocate %ld * %ld bytes", (long) nr, (long) nr_bytes);
  return (vp);
}

heap_vptr ck_walloc (int nr_words)
{ heap_vptr v;
#ifdef MEMCHECK
  if (nr_words <= 0)
    mrts_panic ("Cannot allocate %d words", nr_words);
#endif
  v = (heap_vptr) calloc ((size_t) nr_words, sizeof (value));
  if (v == NULL)
    mrts_panic ("Not enough memory to allocate %d words", nr_words);
  return (v);
}

/*
   Free list administration
   We allocate in units of 'value' size. For sizes less than a certain
   maximum, we do the memory recovery ourselves. For larger values we
   leave it to malloc/free. To get rid of the usual malloc/free overhead
   we allocate chunks of the size of the requested value.

   Note also that we do not administer the exact amount of allocs and frees.
   When needed they can be added (and should be conditionally added).
*/
#define MAX_NR_WORDS 64
#define CHUNK_SIZE 16
static heap_vptr free_list[MAX_NR_WORDS];

#ifdef MEMCHECK
static void check_free_list_tops ()
{ int ix;
  for (ix = 0; ix < MAX_NR_WORDS; ix++)
    if ((long) free_list[ix] & 7L)
      mrts_panic ("value on top of freelist %d, %p is not aligned\n", ix, free_list[ix]);
}
#endif

heap_vptr mrts_alloc_value (int nr_words)
{ heap_vptr vp;
  int ix;
#ifdef MEMCHECK
  check_free_list_tops ();
  if (nr_words <= 0)
    mrts_panic ("Cannot allocate %d words", nr_words);
#endif
  if (nr_words > MAX_NR_WORDS)
    { /* large structure, do normal malloc */
      vp = ck_walloc (nr_words);
    }
  else
    { /* Check if space on free list */
      if (free_list[nr_words - 1] == NULL)
	{ /* Allocate CHUNK_SIZE values for free list */
	  heap_vptr vp2 = ck_walloc (CHUNK_SIZE * nr_words);
	  for (ix = 0; ix < CHUNK_SIZE; ix++, vp2 += nr_words)
	    { vp2 -> next = free_list[nr_words - 1];
	      free_list[nr_words - 1] = vp2;
	    };
	};

      /* Take structure from free list */
      vp = free_list[nr_words - 1];
      free_list[nr_words - 1] = vp -> next;
    };

  /* Precaution: clear mem, should not be necessary */
  for (ix = 0; ix < nr_words; ix++)
    vp[ix].next = NULL;
  return (vp);
}

void mrts_ucfree_value (value old, int nr_words)
{
#ifdef MEMCHECK
  if (old.vp == NULL)
    mrts_panic ("Freeing the null pointer");
  if (nr_words <= 0)
    mrts_panic ("Cannot free %d words", nr_words);
  if ((unsigned long) old.lv < 0x400000L)
    mrts_panic ("Freeing low core value");
  if (old.lv & 7L)
    mrts_panic ("Freeing unaligned value");
  check_free_list_tops ();
#endif
  if (nr_words > MAX_NR_WORDS)
    { /* Large structure, normal free */
      free ((void *) old.vp);
    }
  else
    { /* Move to head of free list */
      old.vp -> next = free_list[nr_words - 1];
      free_list[nr_words - 1] = old.vp;
    };
}

void mrts_free_value (value old, int nr_words)
{
#ifdef MEMCHECK
  if (old.vp == NULL)
    mrts_panic ("Freeing the null pointer");
  if (nr_words <= 0)
    mrts_panic ("Cannot free %d words", nr_words);
  if ((unsigned long) old.lv < 0x400000L)
    mrts_panic ("Freeing low core value");
  if (old.lv & 7L)
    mrts_panic ("Freeing unaligned value");
  if (Refcount (old) < 0)
    mrts_panic ("Freeing a constant");
  if (Refcount (old) > 0)
    mrts_panic ("Freeing an object with non zero reference count");
  check_free_list_tops ();
#endif
  if (nr_words > MAX_NR_WORDS)
    { /* Large structure, normal free */
      free ((void *) old.vp);
    }
  else
    { /* Move to head of free list */
      old.vp -> next = free_list[nr_words - 1];
      free_list[nr_words - 1] = old.vp;
    };
}

value mrts_create_value (int tag, int nr_words)
{ heap_vptr vp = mrts_alloc_value (nr_words);
  value v;
  v.vp = vp;
  vp -> adm.ref_count = 1;
  vp -> adm.tag = tag;
  return (v);
}

/*
   Define access procedure for arrays
*/
#define INDIRECT_BUCKET_SIZE 512
#define INDIRECT_BUCKET_MASK 0x1ff
#define INDIRECT_BUCKET_SHIFT 9
#define DIRECT_BUCKET_SIZE 512
#define DIRECT_BUCKET_MASK 0x1ff
#define DIRECT_BUCKET_SHIFT 9

value *mrts_index_array (Array *g_array, long idx)
{ /* First check if index is not negative */
  void **dbptr;
  long widx, bidx;
  int idepth;
  if (idx < 0L)
    mrts_error ("Indexing an array with negative index %ld", idx);

  /* split off index into lowest (direct bucket) */
  widx = (idx & DIRECT_BUCKET_MASK); 
  bidx = (idx >> DIRECT_BUCKET_SHIFT);

  /*
     If the bucket index is larger than accomodated by the current depth
     create the intermediate layers to the top of the array structure
  */
  while (bidx >= (1L << (g_array -> depth * INDIRECT_BUCKET_SHIFT)))
    { void **new_id_bucket = (void **) ck_calloc (INDIRECT_BUCKET_SIZE, sizeof (void *));
      new_id_bucket[0] = g_array -> bptr;
      g_array -> bptr = (void *) new_id_bucket;
      g_array -> depth++;
    };

  /*
     Now that we can index at the topmost layer, descend by sub indexing,
     creating intermediate layers in between when necessary
  */
  dbptr = &g_array -> bptr;
  idepth = g_array -> depth;
  while (idepth > 0)
    { void **old_id_bucket;
      int sidx;
      idepth--;
      sidx = (bidx >> (idepth * INDIRECT_BUCKET_SHIFT)) & INDIRECT_BUCKET_MASK;
      if (*dbptr == NULL)
        *dbptr = ck_calloc (INDIRECT_BUCKET_SIZE, sizeof (void *));
      old_id_bucket = (void **) (*dbptr);
      dbptr = &old_id_bucket[sidx];
    };
  
  /*
     dbptr now points to the location where we may hang (or already have hung)
     a pointer to a lowest most bucket, which contains actual CDL values
  */
  if (*dbptr == NULL)
    *dbptr = ck_calloc (DIRECT_BUCKET_SIZE, sizeof (value));
  return (&((value *)(*dbptr))[widx]);
}

void mrts_init_constants (int nr, const_init *table, value *consts)
{ int ix;
  for (ix = 0; ix < nr; ix++)
    { int tag = table[ix].tag;
      int size = (tag == -1)?2:1;
      heap_vptr vp = ck_walloc (size);
      consts[table[ix].const_nr].vp = vp;
      vp -> adm.ref_count = -1;		/* Really constant */
      vp -> adm.tag = tag;
      if (tag == -1)
	vp[1].vtext = table[ix].ctext;
    };
};

/*
   Initialize the basic runtime system
*/
void mrts_init_basic ()
{ int ix;
  for (ix = 0; ix < MAX_NR_WORDS; ix++) free_list[ix] = NULL;
}
