/*
   File: mcdl_predef.c
   Defines the set of predefined operations for MCDL

   Copyright (C) 2008 Marc Seutter

   This library is free software: you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published   
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU Lesser General Public License
   along with this library.  If not, see <http://www.gnu.org/licenses/>.

   CVS ID: "$Id: mrts_predef.c,v 1.15 2009/10/07 14:53:11 marcs Exp $"
*/

/* Standard includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <time.h>
#include <math.h>
#include <errno.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifndef WIN32
#include <unistd.h>
#endif

/* Local includes */
#include "mrts_mcdl.h"
#include "mrts_stdtypes.h"
#include "mrts_predef.h"

/*
   File operations

   PRED open input file (>TEXT, FILE>)
   PRED open output file (>TEXT, FILE>)
   ACTION open standard output (FILE>)
   ACTION open standard error (FILE>)
   ACTION open standard input (FILE>)
   PRED close (>FILE)
   PRED seek (>FILE, >INT)
*/

/*
   PRED open input file (>TEXT, FILE>)
*/
int mrts_open_input_file_TEXT_FILE (value text, value *ret_file)
{ FILE *handle = fopen (Value_TEXT (text), "rb");
  if (handle == NULL) return (0);
  *ret_file = mrts_create_file (handle);
  return (1);
}

/*
   PRED open output file (>TEXT, FILE>)
*/
int mrts_open_output_file_TEXT_FILE (value text, value *ret_file)
{ FILE *handle = fopen (Value_TEXT (text), "wb");
  if (handle == NULL) return (0);
  *ret_file = mrts_create_file (handle);
  return (1);
}

/*
   PRED open standard output (FILE>) to become an action
*/
int mrts_open_standard_output_FILE (value *ret_file)
{ *ret_file = mrts_create_file (stdout);
  return (1);
}

/*
   PRED open standard error (FILE>) to become an action
*/
int mrts_open_standard_error_FILE (value *ret_file)
{ *ret_file = mrts_create_file (stderr);
  return (1);
}

/*
   PRED open standard error (FILE>) to become an action
*/
int mrts_open_standard_input_FILE (value *ret_file)
{ *ret_file = mrts_create_file (stdin);
  return (1);
}

/*
   ACTION close (>FILE) to become a predicate
*/
int mrts_close_FILE (value file)
{ fclose (Value_FILE (file));
  return (1);
}

/*
   PRED seek (>FILE, >INT)
*/
int mrts_seek_FILE_INT (value file, value offset)
{ return (fseek (Value_FILE (file), Value_INT (offset), SEEK_SET) == 0);
}

/*
   Pipe operations

   PRED pipe to command (>TEXT, FILE>)
   PRED pipe from command (>TEXT, FILE>)
   PRED close pipe (>FILE)
*/

/*
   PRED pipe to command (>TEXT, FILE>),
*/
int mrts_pipe_to_command_TEXT_FILE (value text, value *ret_file)
{ FILE *handle;
#ifdef WIN32
  handle = _popen (Value_TEXT (text), "wb");
#else
  handle = popen (Value_TEXT (text), "w");
#endif
  if (handle == NULL) return (0);
  *ret_file = mrts_create_file (handle);
  return (1);
}

/*
   PRED pipe from command (>TEXT, FILE>),
*/
int mrts_pipe_from_command_TEXT_FILE (value text, value *ret_file)
{ FILE *handle;
#ifdef WIN32
  handle = _popen (Value_TEXT (text), "rb");
#else
  handle = popen (Value_TEXT (text), "r");
#endif
  if (handle == NULL) return (0);
  *ret_file = mrts_create_file (handle);
  return (1);
}

/*
   PRED close pipe (>FILE)
*/
int mrts_close_pipe_FILE (value file)
{
#ifdef WIN32
  return (_pclose (Value_FILE (file)) == 0);
#else
  return (pclose (Value_FILE (file)) == 0);
#endif
}

/*
   Primitive write functions
   ACTION write (>TEXT)
   ACTION write (>INT)
   ACTION write (>REAL)
   ACTION write char (>INT)
   ACTION write (>FILE, >TEXT)
   ACTION write (>FILE, >INT)
   ACTION write (>FILE, >REAL)
   ACTION write char (>FILE, >INT)
   ACTION formatted write (>FILE, >FORMAT)
*/

/*
   ACTION write (>TEXT)
*/
int mrts_write_TEXT (value text)
{ fputs (Value_TEXT (text), stdout);
  return (1);
}

/*
   ACTION write (>INT)
*/
int mrts_write_INT (value val)
{ fprintf (stdout, "%ld", Value_INT (val));
  return (1);
}

/*
   ACTION write (>REAL)
*/
int mrts_write_REAL (value val)
{ fprintf (stdout, "%g", Value_REAL (val));
  return (1);
}

/*
   ACTION write char (>INT)
*/
int mrts_write_char_INT (value val)
{ unsigned char c = (unsigned char) (Value_INT (val) & 0xffL);
  fputc (c, stdout);
  return (1);
}

/*
   ACTION write (>FILE, >TEXT)
*/
int mrts_write_FILE_TEXT (value file, value text)
{ fputs (Value_TEXT (text), Value_FILE (file));
  return (1);
}

/*
   ACTION write (>FILE, >INT)
*/
int mrts_write_FILE_INT (value file, value val)
{ fprintf (Value_FILE (file), "%ld", Value_INT (val));
  return (1);
}

/*
   ACTION write char (>FILE, >INT)
*/
int mrts_write_char_FILE_INT (value file, value val)
{ unsigned char c = (unsigned char) (Value_INT (val) & 0xffL);
  fputc (c, Value_FILE (file));
  return (1);
}

/*
   ACTION formatted write (>FILE,>FORMAT)

   FORMAT :: FORMAT INT;
             FORMAT REAL;
             FORMAT TEXT;
             INT;
             REAL;
             TEXT.
*/
static void formatted_write (FILE *file, value format)
{ switch (AlternativeNr (format))
    { case 0:	/* FORMAT INT */
        formatted_write (file, PartValue (format, 1));
	fprintf (file, "%ld", Value_INT (PartValue (format, 2)));
	break;
      case 1:	/* FORMAT REAL */
	formatted_write (file, PartValue (format, 1));
	fprintf (file, "%g", Value_REAL (PartValue (format, 2)));
	break;
      case 2:	/* FORMAT TEXT */
	formatted_write (file, PartValue (format, 1));
	fputs (Value_TEXT (PartValue (format, 2)), file);
	break;
      case 3:	/* INT */
	fprintf (file, "%ld", Value_INT (PartValue (format, 1)));
	break;
      case 4:	/* REAL */
	fprintf (file, "%g", Value_REAL (PartValue (format, 1)));
	break;
      case 5: /* TEXT */
	fputs (Value_TEXT (PartValue (format, 1)), file);
	break;
      default: mrts_panic ("impossible FORMAT alternative");
    };
}

int mrts_formatted_write_FILE_FORMAT (value file, value format)
{ formatted_write (Value_FILE (file), format);
  return (1);
}

/*
   Trace functions

   FUNCTION trace (>TEXT)
   FUNCTION trace (>INT)
   FUNCTION trace (>REAL)
   FUNCTION formatted trace (>FILE, >FORMAT)
*/

/*
   FUNCTION trace (>TEXT)
*/
int mrts_trace_TEXT (value text)
{ fflush (stdout);
  fputs (Value_TEXT (text), stderr);
  fflush (stderr); 
  return (1);
}

/*
   FUNCTION trace (>INT)
*/
int mrts_trace_INT (value val)
{ fflush (stdout);
  fprintf (stderr, "%ld", Value_INT (val));
  fflush (stderr); 
  return (1);
}

/*
   FUNCTION trace (>REAL)
*/
int mrts_trace_REAL (value val)
{ fflush (stdout);
  fprintf (stderr, "%g", Value_REAL (val));
  fflush (stderr); 
  return (1);
}

/*
   FUNCTION formatted trace (>FILE, >FORMAT)
*/
int mrts_formatted_trace_FILE_FORMAT (value file, value format)
{ fflush (stdout);
  formatted_write (stderr, format);
  fflush (stderr); 
  return (1);
}

/*
   Primitive read functions

   PRED read line (TEXT>)
   PRED read line (>FILE,TEXT>)
   PRED read char (TEXT>)
   PRED read char (>FILE,TEXT>)
   PRED read char (INT>)
   PRED read char (>FILE,INT>)
*/
#define MAXLINELEN 4096
static int read_line (FILE *file, value *ret_text)
{ char this_line[MAXLINELEN];
  if (fgets (this_line, MAXLINELEN - 1, file) == NULL)
    return (0);

  /* Terminate just in case */
  this_line[MAXLINELEN - 1] = '\0';
  *ret_text = mrts_create_text (this_line);
  return (1);
}

static int read_char_text (FILE *file, value *ret_text)
{ char str[2];
  int c = fgetc (file);
  if (c == EOF) return (0);

  str[0] = (char) (c & 0xff);
  str[1] = '\0';

  *ret_text = mrts_create_text (str);
  return (1);
}

static int read_char_int (FILE *file, value *ret_int)
{ int c = fgetc (file);
  if (c == EOF) return (0);

  *ret_int = Create_INT ((long) c);
  return (1);
}

/*
   PRED read line (TEXT>)
*/
int mrts_read_line_TEXT (value *ret_text)
{ return (read_line (stdin, ret_text));
}

/*
   PRED read line (>FILE, TEXT>)
*/
int mrts_read_line_FILE_TEXT (value file, value *ret_text)
{ return (read_line (Value_FILE (file), ret_text));
}

/*
   PRED read char (TEXT>)
*/
int mrts_read_char_TEXT (value *ret_text)
{ return (read_char_text (stdin, ret_text));
}

/*
   PRED read char (>FILE, TEXT>)
*/
int mrts_read_char_FILE_TEXT (value file, value *ret_text)
{ return (read_char_text (Value_FILE (file), ret_text));
}

/*
   PRED read char (INT>)
*/
int mrts_read_char_INT (value *ret_int)
{ return (read_char_int (stdin, ret_int));
}

/*
   PRED read char (>FILE, INT>)
*/
int mrts_read_char_FILE_INT (value file, value *ret_int)
{ return (read_char_int (Value_FILE (file), ret_int));
}

/*
   Flushing function
*/
int mrts_flush ()
{ fflush (stdout);
  return (1);
}

int mrts_flush_FILE (value file)
{ fflush (Value_FILE (file));
  return (1);
}

/*
   Text arithmetic functions

   TEST head (>TEXT, TEXT1>)
   TEST tail (>TEXT, TEXT1>)
   TEST split (>TEXT, TEXT1>, TEXT2>)
   FUNCTION extract (>TEXT,>INT1,>INT2,TEXT1>)
   TEST is subtext (>TEXT,>INT>,>TEXT1)
   TEST between (>TEXT,>INT,>TEXT1,>TEXT2)
   TEST not between (>TEXT,>INT,>TEXT1,>TEXT2)
   TEST before (>TEXT,>TEXT1)
   TEST is prefix (>TEXT,>TEXT1,TEXT2>)
   TEST prefix (>TEXT,>INT,TEXT1>,TEXT2>)
   FUNCTION length (>TEXT,INT>)
*/

/*
   TEST head (>TEXT, TEXT1>)
*/
int mrts_head_TEXT_TEXT (value text, value *ret_text)
{ char *txt = Value_TEXT (text);
  char tbuf[2];
  if (*txt == '\0') return (0);
  tbuf[0] = *txt;
  tbuf[1] = '\0';
  *ret_text = mrts_create_text (tbuf);
  return (1);
}

/*
   TEST tail (>TEXT, TEXT1>)
*/
int mrts_tail_TEXT_TEXT (value text, value *ret_text)
{ char *txt = Value_TEXT (text);
  if (*txt == '\0') return (0);
  *ret_text = mrts_create_text (txt + 1);
  return (1);
}

/*
   TEST split (>TEXT, TEXT1>, TEXT2>)
*/
int mrts_split_TEXT_TEXT_TEXT (value text, value *ret_text1, value *ret_text2)
{ char *txt = Value_TEXT (text);
  char tbuf[2];
  if (*txt == '\0') return (0);
  tbuf[0] = *txt;
  tbuf[1] = '\0';
  *ret_text1 = mrts_create_text (tbuf);
  *ret_text2 = mrts_create_text (txt + 1);
  return (1);
}

/*
   FUNCTION extract (>TEXT,>INT1,>INT2,TEXT1>)
   Behaviour is undefined when INT1 and INT2 are out of bounds
*/
int mrts_extract_TEXT_INT_INT_TEXT (value text, value lwb, value upb, value *ret_text)
{ size_t size = (size_t) (Value_INT (upb) - Value_INT (lwb));
  char *vtext;
  if (size <= 0)
    { vtext = (char *) mrts_alloc_value (1);
      vtext[0] = '\0';
    }
  else
    { vtext = (char *) mrts_alloc_value (SizeFromLength (size + 1));
      strncpy (vtext, Value_TEXT (text) + Value_INT (lwb), size);
      vtext[size] = '\0';
    }
  *ret_text = mrts_alloc_text (vtext);
  return (1);
}

/*
   TEST is subtext (>TEXT,>INT>,>TEXT1)
   Note: An empty text may match at the end of first parameter
*/
int mrts_is_subtext_TEXT_INT_TEXT (value text1, value *trans_int, value text2)
{ size_t patlen = (size_t) strlen (Value_TEXT (text2));
  long len = (long) strlen (Value_TEXT (text1));
  long strt_pos = Value_INT (*trans_int);

  if (strt_pos > len) return (0);
  if (strncmp (Value_TEXT (text1) + strt_pos, Value_TEXT (text2), patlen)) return (0);
  *trans_int = Create_INT ((long)(strt_pos + patlen));
  return (1);
}

/*
   TEST between (>TEXT,>INT,>TEXT1,>TEXT2)
*/
int mrts_between_TEXT_INT_TEXT_TEXT (value text, value pos, value lwb, value upb)
{ char *txt = Value_TEXT (text);
  long len = (long) strlen (txt);
  long tpos = Value_INT (pos);
  unsigned char c;
  if (tpos >= len) return (0);
  c = (unsigned char) txt[tpos];
  return (((unsigned char) (*Value_TEXT (lwb)) <= c) &&
	  (c <= (unsigned char) (*Value_TEXT (upb))));
}

/*
   TEST not between (>TEXT,>INT,>TEXT1,>TEXT2)
*/
int mrts_not_between_TEXT_INT_TEXT_TEXT (value text, value pos, value lwb, value upb)
{ char *txt = Value_TEXT (text);
  long len = (long) strlen (txt);
  long tpos = Value_INT (pos);
  unsigned char c;
  if (tpos >= len) return (0);
  c = (unsigned char) txt[tpos];
  return ((c < (unsigned char) (*Value_TEXT (lwb)) ||
	  (unsigned char) (*Value_TEXT (upb)) < c));
}

/*
   TEST before (>TEXT,>TEXT1)
*/
int mrts_before_TEXT_TEXT (value text1, value text2)
{ return (strcmp (Value_TEXT (text1), Value_TEXT (text2)) < 0);
}

/*
   TEST is prefix (>TEXT,>TEXT1,TEXT2>)
*/
int mrts_is_prefix_TEXT_TEXT_TEXT (value prefix, value text, value *ret_text)
{ char *ptxt = Value_TEXT (prefix);
  char *ttxt = Value_TEXT (text);
  long plen = (long) strlen (ptxt);
  long tlen = (long) strlen (ttxt);

  /* Check prefix length against total length */
  if (plen > tlen) return (0);
  if (strncmp (ptxt, ttxt, plen)) return (0);
  *ret_text = mrts_create_text (ttxt + plen);
  return (1);
}

/*
   TEST prefix (>TEXT,>INT,TEXT1>,TEXT2>)
*/
int mrts_prefix_TEXT_INT_TEXT_TEXT (value text, value len, value *ret_text1, value *ret_text2)
{ char *ttxt = Value_TEXT (text);
  long tlen = (long) strlen (ttxt);
  long plen = Value_INT (len);
  char *prefix;

  /* Check prefix length against total length */
  if (plen > tlen) return (0);
  prefix = (char *) mrts_alloc_value (SizeFromLength (plen + 1));
  strncpy (prefix, ttxt, plen);
  prefix[plen] = '\0';
  *ret_text1 = mrts_alloc_text (prefix);
  *ret_text2 = mrts_create_text (ttxt + plen);
  return (1);
}

/*
   FUNCTION length (>TEXT,INT>)
*/
int mrts_length_TEXT_INT (value text, value *ret_len)
{ *ret_len = Create_INT ((long) strlen (Value_TEXT (text)));
  return (1);
}

/*
   Type conversions

   FUNCTION convert (>INT) to (REAL>):             EXTERNAL "mrts_convert_to".
   FUNCTION round (>REAL) to (INT>):               EXTERNAL "mrts_round_to".
   FUNCTION trunc (>REAL) to (INT>):               EXTERNAL "mrts_trunc_to".
   FUNCTION convert (>INT) to (TEXT>):             EXTERNAL "mrts_convert_to".
   FUNCTION convert (>REAL) to (TEXT>):            EXTERNAL "mrts_convert_to".
   TEST convert (>TEXT) to (INT>):                 EXTERNAL "mrts_convert_to".
   TEST convert (>TEXT) to (REAL>):                EXTERNAL "mrts_convert_to".
*/
int mrts_convert_to_INT_REAL (value ival, value *ret_real)
{ double re = (double) (Value_INT (ival));
  *ret_real = Create_REAL (re);
  return (1);
}

int mrts_round_to_REAL_INT (value rval, value *ret_int)
{ double mrval = Value_REAL (rval);
  if (mrval < 0.0) mrval = -(floor (-mrval + 0.5));
  else mrval = floor (mrval + 0.5);
  *ret_int = Create_INT ((long) mrval);
  return (1);
}

int mrts_trunc_to_REAL_INT (value rval, value *ret_int)
{ double mrval = Value_REAL (rval);
  if (mrval < 0.0) mrval = -(floor (-mrval));
  else mrval = floor (mrval);
  *ret_int = Create_INT ((long) mrval);
  return (1);
}

int mrts_convert_to_INT_TEXT (value ival, value *ret_text)
{ char lbuf[32];
  sprintf (lbuf, "%ld", Value_INT (ival));
  *ret_text = mrts_create_text (lbuf);
  return (1);
}

int mrts_convert_to_REAL_TEXT (value rval, value *ret_text)
{ char lbuf[32];
  sprintf (lbuf, "%g", Value_REAL (rval));
  *ret_text = mrts_create_text (lbuf);
  return (1);
}

int mrts_convert_to_TEXT_INT (value text, value *ret_int)
{ char *tptr = Value_TEXT (text); 
  char *rptr;
  long ival;
  errno = 0;
  ival = strtol (tptr, &rptr, 10);

  /* Format check */
  if (rptr == tptr) return (0);
  while (isspace (*rptr)) rptr++;
  if (*rptr) return (0);

  /* Overflow check */
  if ((ival == LONG_MAX) && (errno == ERANGE))
    return (0);

  /* Underflow check */
  if ((ival == LONG_MIN) && (errno == ERANGE))
    return (0);

  /* All ok */
  *ret_int = Create_INT (ival);
  return (1);
}

int mrts_convert_to_TEXT_REAL (value text, value *ret_real)
{ char *tptr = Value_TEXT (text); 
  char *rptr;
  double rval;
  errno = 0;
  rval = strtod (tptr, &rptr);

  /* Format check */
  if (rptr == tptr) return (0);
  while (isspace (*rptr)) rptr++;
  if (*rptr) return (0);

  /* Overflow check */
  if ((rval == HUGE_VAL) && (errno == ERANGE))
    return (0);
  if ((rval == -HUGE_VAL) && (errno == ERANGE))
    return (0);
  
  *ret_real = Create_REAL (rval);
  return (1);
}

/*
   Legacy type conversion

   TEST asciicode (>TEXT,>INT1,INT2>),
   FUNCTION bindec (>INT,TEXT>),	To disappear
   TEST decbin (>TEXT,INT>)		To disappear
*/
int mrts_asciicode_TEXT_INT_INT (value text, value pos, value *ret_int)
{ char *txt = Value_TEXT (text);
  long len = (long) strlen (txt);
  long tpos = Value_INT (pos);
  unsigned char c;
  if (tpos >= len) return (0);
  c = (unsigned char) txt[tpos];
  *ret_int = Create_INT (((long) c) & 0xFFL);
  return (1);
}

int mrts_bindec_INT_TEXT (value ival, value *ret_text)
{ char lbuf[32];
  sprintf (lbuf, "%ld", Value_INT (ival));
  *ret_text = mrts_create_text (lbuf);
  return (1);
}

int mrts_decbin_TEXT_INT (value text, value *ret_int)
{ char *tptr = Value_TEXT (text); 
  char *rptr;
  long ival = strtol (tptr, &rptr, 10);
  if (rptr == tptr) return (0);
  while (isspace (*rptr)) rptr++;
  if (*rptr) return (0);
  *ret_int = Create_INT (ival);
  return (1);
}

/*
   System interface

   TEST command arg (>INT,TEXT>)
   FUNCTION exit (>INT)
   TEST get from environment (>TEXT1,TEXT2>)
   FUNCTION time (INT>)
   PRED execute (>TEXT)
   TEST file date (>TEXT,INT>)
*/

/*
   TEST command arg (>INT, TEXT>)
*/
static int saved_argc;
static value *saved_argv;
static void save_arguments (int argc, char **argv)
{ int ix;
  saved_argc = argc;
  saved_argv = (value *) ck_walloc (argc);
  for (ix = 0; ix < argc; ix++)
    saved_argv[ix] = mrts_create_text (argv[ix]);
}

int mrts_command_arg_INT_TEXT (value nr, value *ret_text)
{ int ix = Value_INT (nr);
  if (ix >= saved_argc) return (0);
  *ret_text = saved_argv[ix];
  Attach_TEXT (*ret_text);
  return (1);
}

/*
   FUNCTION exit (>INT)
*/
int mrts_exit_INT (value rval)
{ exit (Value_INT (rval));
  return (0);
}

/*
   TEST get from environment (>TEXT1, TEXT2>)
*/
int mrts_get_from_environment_TEXT_TEXT (value text, value *ret_text)
{ char *env_value = getenv (Value_TEXT (text));
  if (env_value == NULL) return (0);
  *ret_text = mrts_create_text (env_value);
  return (1);
}

/*
   FUNCTION time (INT>)
*/
int mrts_time_INT (value *ret_int)
{ *ret_int = Create_INT ((long) time (NULL));
  return (1);
}

/*
   PRED execute (>TEXT)
*/
int mrts_execute_TEXT (value text)
{ return (!system (Value_TEXT (text)));
}

/*
   TEST file date (>TEXT,INT>)
*/
int mrts_file_date_TEXT_INT (value text, value *ret_int)
{ 
#ifdef WIN32
  struct _stat filestat;
  if (_stat (Value_TEXT (text), &filestat)) return (0);
#else
  struct stat filestat;
  if (stat (Value_TEXT (text), &filestat)) return (0);
#endif
  *ret_int = Create_INT ((long) (filestat.st_mtime));
  return (1);
}

/*
   Miscellaneous

   FUNCTION hash (>TEXT,>INT1,INT2>)
*/
int mrts_hash_TEXT_INT_INT (value text, value ival, value *ret_int)
{ unsigned char *utxt = (unsigned char *) Value_TEXT (text);
  unsigned long val = 0;

  while (*utxt) val += (unsigned long) (*utxt++);
  *ret_int = Create_INT ((long) val % Value_INT (ival));
  return (1);
}

/*
   Initialize the predefines
*/
void mrts_init_predef (int argc, char **argv)
{ save_arguments (argc, argv);
}
