/*
   File: mrts_stdtypes.h
   Defines the access macros for the standard types INT, REAL, TEXT and FILE

   Copyright (C) 2008 Marc Seutter

   This library is free software: you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published   
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU Lesser General Public License
   along with this library.  If not, see <http://www.gnu.org/licenses/>.

   CVS ID: "$Id: mrts_stdtypes.h,v 1.1 2009/02/18 14:54:21 marcs Exp $"
*/
#ifndef IncMrtsStdtypes
#define IncMrtsStdtypes

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* For compiling with C++ */
#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

/*
   For every implemented type TYPE the following names, either defined
   as a macro or as a function call must be present in the runtime
   system:

   Tag_TYPE: Defines the special tag (only needed for primitive types like TEXT or FILE)
   Attach_TYPE: Defines the attach operation for the type
   Create_TYPE: Defines the creation of a value of the type
   Detach_TYPE: Defines the detach operation for the type
   Value_TYPE:  Gives the actual value (only needed for primitive types)
   Equal_TYPE:  Equality test for the type
   Write_TYPE:  Writes the value to stderr (debugging)

   For some of our primitive types, we would like to cast such a type to a value
   In MSVC71 this is not implemented (It is after all a GCC extension)
   For these case we implement an inline function which does the trick
   Note: enable __inline extensions in MSVC71
*/

/*
   Primitive type INT
   All INT values are represented by a long, cast to value
*/
#ifndef WIN32
#define Create_INT(lv) ((value) (lv))
#else
__inline value Create_INT (long lv) { value v; v.lv = lv; return (v); }
#endif
#define Value_INT(v) ((v).lv)
#define Attach_INT(v)
#define Detach_INT(v)
#define Equal_INT(v1,v2) (Value_INT(v1) == Value_INT(v2))
#define Write_INT(v) fprintf(stderr, "%ld", Value_INT(v))

/*
   Primitive type REAL
   All REAL values are represented by a double, cast to value
*/
#ifndef WIN32
#define Create_REAL(re) ((value) (re))
#else
__inline value Create_REAL (double re) { value v; v.re = re; return (v); }
#endif
#define Value_REAL(v) ((v).re)
#define Attach_REAL(v)
#define Detach_REAL(v)
#define Equal_REAL(v1,v2) (Value_REAL(v1) == Value_REAL(v2))
#define Write_REAL(v) fprintf(stderr, "%g", Value_REAL(v))

/*
   Primitive type TEXT
   All TEXT values are represented by pointer to a heap value,
   of which the first element contains the refcount and tag
   and the second element a pointer to the actual string.
*/
#define Tag_TEXT -1
#define Create_TEXT(tv) mrts_create_text(tv)
#define Value_TEXT(v) ((v).vp[1].vtext)
#define Attach_TEXT(v) Attach(v)
#define Detach_TEXT(v) { if (TryDecrRefcount(v)) mrts_free_text (v); }
#define Equal_TEXT(v1,v2) (!strcmp(Value_TEXT(v1), Value_TEXT(v2)))
#define Write_TEXT(v) fprintf(stderr, "%s", Value_TEXT(v))

/*
   Primitive type FILE
   All FILE values are represented by pointer to a value,
   containing a FILE * pointer
*/
#define Tag_FILE -2
#define Create_FILE(fv) mrts_create_file(fv)
#define Value_FILE(v) ((v).vp[1].vfile)
#define Attach_FILE(v) Attach(v)
#define Detach_FILE(v) { if (TryDecrRefcount(v)) mrts_free_file (v); }
#define Equal_FILE(v) (Value_FILE(v1) == Value_FILE(v2))
#define Write_FILE(v) fprintf(stderr, "%p", Value_FILE(v))

/* Exported routines */
value mrts_alloc_text (char *vtext);
value mrts_create_text (char *tv);
void mrts_free_text (value v);
value mrts_concat_text (value v1, value v2);
value mrts_concat_texts (value *vl, int nr);
value mrts_repl_text (long lv, value v2);
value mrts_create_file (FILE *fv);
void mrts_free_file (value v);

#ifdef __cplusplus
}
#endif /* __cplusplus */
#endif /* IncMrtsStdtypes */
